!
! Module: eimsimbias_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This is a collection of routines used to reduce simulations data.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module eimsimbias_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.

  use dal
  use param  
  use errorHandling

  use array_utils ! for median()
  use constants
  use dal_aux	! for addOrOpenColumn()
!  use histogram_utils	! for makeHistogram().
use l_histogram_utils	! for makeHistogram().
!  use pgplot_aux
use l_pgplot_aux
  use ssc_misc	! for fracVal().

  implicit none 

  private 

  ! Command-line parameters:
  !
  character(256) ::&
    detSrcSetName='',&
    histoSetName='',&
    plot1FileName='',&
    plot2FileName='',&
    plot3FileName=''
  character(20) ::&
    xColName='',&
    pStr=''
  real(single) ::&
    userXHistoBinWidth=0.0
!********** the other bin width.
  logical ::&
    addToExistingHisto=.false.,&
    makePlots=.false.

character(20) :: pgDev = '/cps'

  public eimsimbias

contains 

  !---------------------------------------------------------------------
  subroutine eimsimbias
  !---------------------------------------------------------------------
  ! The source detection procedure measures the values, and the uncertainties, of source parameters such as flux, and X and Y positions. If we choose one of these parameters, denoted by p, then we have a priori a value for p_sim, plus the source detection returns values p_det and p_det_uncert. The purpose of the present task is to make two sets of histograms involving these p values, described as follows.
  !
  ! Histo table 1: all histograms in this table use the values from xColName as their X axis, and have an identical (logarithmic-equal) binning scheme. Histograms are made with the following weights:
  !
  !	Histo name:	Col?	Weight formula:
  ! ------------------------------------------------------
  !	*_DIFF		Y	p_det - p_sim
  !	*_DIFF_ERR	Y	%stdDev field of the previous histogram variable
  !	pUncertHisto		p_det_uncert
  !	*_ERR_RATIO	Y	pUncertHisto / *_DIFF_ERR
!****** how to calculate an error bar for the error ratio?
  ! ------------------------------------------------------
  !
  ! Here * stands for the pStr argument. If the parameter measurements are correct, the values in the first and second columns should not deviate from 0 and 1 respectively.
  !
  ! Histo table 2: a simple unweighted ie frequency histogram is made against the following binned X:
  !
  !	X = ((p_det - p_sim) - <p_det - p_sim>) / p_det_uncert
  !
  ! The histogram X and Y are saved in columns D_*_REL and N respectively. One would expect the resulting histogram to be a gaussian with width 1. For convenience, the task calculates this gaussian and saves it in N_THEOR.
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT) :: histoSet, sourceSet
    type(TableT) :: sourceTab
    integer ::&
      numSrcs, numHistoBins, i, dummyUnderflows, dummyOverflows, row,&
      numNonNull, num
    character(20) ::&
      pDetColName, pDetUncertColName, pSimColName,&
      histo1TabName, yValColName, yStdDevColName, yStdDevErrColName,&
      yAveErrColName, histo2TabName, midColName, xMidColName
    type(ColumnT) ::&
      xMidCol, yValCol, yStdDevCol, yStdDevErrCol, yAveErrCol,&
      midCol, nCol, nTheorCol, nErrCol
    real(single), pointer ::&
      xData(:), binEdges(:), xMid(:), nErrData(:),&
      yVal(:), yStdDev(:), yStdDevErr(:), yAveErr(:), nTheorData(:),&
      midData(:), pDetUncertData(:)
    real(double), pointer :: pDetData(:), pSimData(:)
    real(single) ::&
      meanDiffP, medianDiffP, logMinX, logMaxX, deltaLogX, midLogX,&
      histoXSpan, minHistoVal, maxHistoVal, diffRelBinSize, xLo, xHi
    real(single), allocatable ::&
      binCentres(:), pDiffRel(:), pDiff(:), packedXmid(:), packedYVal(:),&
      packedYStdDev(:), packedYAveErr(:), packedYStdDevErr(:)
    type(weightedHistoSingleType), allocatable ::&
      pDiffHisto(:), pUncertHisto(:)
    type(TableT) :: histo1Tab, histo2Tab
    character(8) :: meanPAttrName, medianPAttrName
    integer(int32), allocatable :: pRelHisto(:)
    integer(int32), pointer :: nData(:)
    character(100) :: xLabel, yLabel
    character(300) :: pgplotDevice
    logical, allocatable :: rowIsNull(:), pointsNotLines(:)
    logical :: histoFileExists
    integer, allocatable :: styles(:), colours(:)
    character(30), allocatable :: texts(:)
 type(ColumnT) :: xCol, pSimCol
 character(20) :: xUnits='', yUnits=''

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    sourceSet = dataSet(detSrcSetName, READ)
    sourceTab = table(sourceSet, 'SRCLIST')
    if (numberOfRows(sourceTab) <= 0)&
call error('noSources', 'No sources in the table.')

    write(errStr, '(a,a,a,a,a)')&
      'Z column name ', trim(xColName), ', coord ',trim(pStr),'; make histogram.'
    call message(AppMsg, Verbose)

    numSrcs = numberOfRows(sourceTab)
    allocate(pDiff(numSrcs))

    pDetColName       = pStr
    pDetUncertColName = trim(pStr)//'_ERR'
    pSimColName  = 'SIM_'//pStr

!    call readColDataToPtr(sourceTab, xColName,          xData) ! in ssclib/dal_aux.
 xCol = column(sourceTab, xColName)
 call readColDataToPtr(xCol, xData) ! in ssclib/dal_aux.
 xUnits = units(xCol)
 call release(xCol)
    call readColDataToPtr(sourceTab, pDetColName,       pDetData) ! in ssclib/dal_aux.
    call readColDataToPtr(sourceTab, pDetUncertColName, pDetUncertData) ! in ssclib/dal_aux.
!    call readColDataToPtr(sourceTab, pSimColName,       pSimData) ! in ssclib/dal_aux.
 pSimCol = column(sourceTab, pSimColName)
 call readColDataToPtr(pSimCol, pSimData) ! in ssclib/dal_aux.
 yUnits = units(pSimCol)
 call release(pSimCol)

    if (any(pDetUncertData <= 0.0)) then
      write(errStr, '(a,a,a)') "Some values of the column ", trim(pDetUncertColName)&
        , " were found to be <= 0"
call error('nonPositiveXErrors')
    end if

    !...................................................................
    ! Get mean and median x offsets:

    pDiff = real((pDetData - pSimData), single)
    meanDiffP = sum(pDiff) / real(numSrcs)
    medianDiffP = median(pDiff) ! in ssclib/array_utils.

    !...................................................................
    ! Histogram of pDiff against log10Z:

    if (minval(xData) <= 0.0) then
      write(errStr, '(a,a,a)') "Some values of the column ", trim(xColName), " were found to be <= 0"
call error('nonPositivexData')
    end if

    logMinX = log10(minval(xData))
    logMaxX = log10(maxval(xData))
    deltaLogX = logMaxX - logMinX
    numHistoBins = ceiling(deltaLogX / userXHistoBinWidth)
    ! userXHistoBinWidth is assumed to be in log space.

    midLogX = (logMinX + logMaxX) / 2.0
    histoXSpan = numHistoBins * userXHistoBinWidth
    minHistoVal = min(midLogX - histoXSpan / 2.0, logMinX)
    maxHistoVal = max(midLogX + histoXSpan / 2.0, logMaxX)

    allocate(binEdges(numHistoBins+1), binCentres(numHistoBins))
    binEdges = fracVal(logMinX, logMaxX&
      , (/(i/real(numHistoBins), i=0,numHistoBins)/)) ! in ssclib/ssc_misc.
    binCentres = (binEdges(1:numHistoBins) + binEdges(2:numHistoBins+1)) / 2.0

    binEdges   = 10.0**binEdges
    binCentres = 10.0**binCentres

    allocate(pDiffHisto(numHistoBins))
    call makeWeightedHistogram(xData, pDiff, binEdges, pDiffHisto&
      , dummyUnderflows, dummyOverflows)
    ! in ssclib/histogram_utils.

    allocate(pUncertHisto(numHistoBins))
    call makeWeightedHistogram(xData, pDetUncertData, binEdges, pUncertHisto&
      , dummyUnderflows, dummyOverflows)
    ! in ssclib/histogram_utils.

    deallocate(binEdges)

!    !** probably ought to multiply by numSrcs/(numSrcs-1)

    histo1TabName      = trim(pStr)//'_VS_S'
!******** pass through a 'xAxisChar' and use this instead of hard-wired 'S' above.
    xMidColName       = trim(xColName)//'_MID'
    yValColName       = trim(pStr)//'_DIFF'
    yStdDevColName    = trim(pStr)//'_DIFF_STDDEV'
    yStdDevErrColName = trim(pStr)//'_DS_ERR'
    yAveErrColName    = trim(pStr)//'_AVE_ERR'

!    uncertColName      = pStr//'_AVE_ERR'
!    binnedErrColName   = pStr//'_ERR_AVE'
!    errRatioColName = trim(pStr)//'_ERR_RATIO'

    inquire(file=histoSetName, exist=histoFileExists)
    if (addToExistingHisto .and. .not. histoFileExists) then
      call warning('noHistoFileToAddTo',"")
      addToExistingHisto = .false.
    end if

    if (addToExistingHisto) then
      histoSet = dataSet(histoSetName, MODIFY)
    else
      histoSet = dataSet(histoSetName, CREATE)
    end if

    if (hasBlock(histoSet, histo1TabName))&
call error('table1AlreadyExists',"")

    histo1Tab = addTable(histoSet, histo1TabName, numHistoBins)
!    xMidCol       = addColumn(histo1Tab, xMidColName, REAL32&
!      , comment=trim(xColName)//' value at mid-bin.')
!    yValCol       = addColumn(histo1Tab, yValColName, REAL32&
!      , comment=trim(pDetColName)//" minus "//trim(pSimColName))
xMidCol       = addColumn(histo1Tab, xMidColName, REAL32&
 , units=trim(xUnits), comment=trim(xColName)//' value at mid-bin.')
yValCol       = addColumn(histo1Tab, yValColName, REAL32&
 , units=trim(yUnits), comment=trim(pDetColName)//" minus "//trim(pSimColName))
    yStdDevCol    = addColumn(histo1Tab, yStdDevColName, REAL32&
      , comment="Scatter in "//trim(yValColName)//" within the bin.")
    yStdDevErrCol = addColumn(histo1Tab, yStdDevErrColName, REAL32&
      , comment=trim(yStdDevColName)//" divided by sqrt(num values in the bin)")
    yAveErrCol    = addColumn(histo1Tab, yAveErrColName, REAL32&
      , comment="Average value within the bin of "//trim(pDetUncertColName))

!    uncertInBinnedValCol = addColumn(histo1Tab, uncertColName&
!      , REAL32)
!    binnedErrCol         = addColumn(histo1Tab, binnedErrColName&
!      , REAL32)
!    chiSquCol            = addColumn(histo1Tab, 'CHI_SQU',      REAL32)
!    chiSquErrCol         = addColumn(histo1Tab, 'CHI_SQU_ERR',  REAL32)
!    errRatioCol = addColumn(histo1Tab, errRatioColName&
!      , REAL32, comment="(Ave uncertainty)/scatter (should == 1).")

    ! Set up null-valued default for error ratio:
    !
    allocate(rowIsNull(numHistoBins))
!, errInErr(numHistoBins))
    rowIsNull = .true. ! default
    do row=0, numHistoBins-1
      call setToNull(      xMidCol, row)
      call setToNull(      yValCol, row)
      call setToNull(   yStdDevCol, row)
      call setToNull(yStdDevErrCol, row)
      call setToNull(   yAveErrCol, row)
    end do

    xMid       => real32Data(      xMidCol)
    yVal       => real32Data(      yValCol)
    yStdDev    => real32Data(   yStdDevCol)
    yStdDevErr => real32Data(yStdDevErrCol)
    yAveErr    => real32Data(   yAveErrCol)

!    uncertInBinnedVal  => real32Data(uncertInBinnedValCol)
!    binnedErr          => real32Data(binnedErrCol)
!    chiSqu             => real32Data(chiSquCol)
!    chiSquErr          => real32Data(chiSquErrCol)
!    errRatio           => real32Data(errRatioCol)
!    errInErr = 0.0 ! default

    where(pDiffHisto%n > 0 .and. pDiffHisto%yStdDev > 0.0)
      xMid       = binCentres
      yVal       = pDiffHisto%yAve
      yStdDev    = pDiffHisto%yStdDev
      yStdDevErr = yStdDev / sqrt(real(pDiffHisto%n, single)) ! sort of a guestimate. There is no rigorous theory behind this expression.
      yAveErr    = pUncertHisto%yAve

!      uncertInBinnedVal  = xDiffHisto%yUncert
!      binnedErr          = xErrHisto%yAve
!      chiSqu             = xChi2Histo%yAve
!      chiSquErr          = xChi2Histo%yUncert
!      errRatio           = pUncertHisto%yAve / pDiffHisto%yStdDev
!      errInErr           = errRatio / sqrt(real(pDiffHisto%n, single)) ! sort of a guestimate. There is no rigorous theory behind this expression.

      rowIsNull = .false.
    end where

    meanPAttrName   = trim(pStr)//'MEAN'
    medianPAttrName = trim(pStr)//'MEDN'

    call setAttribute(histo1Tab, meanPAttrName, meanDiffP)
    call setAttribute(histo1Tab, medianPAttrName, medianDiffP)
    call setAttribute(histo1Tab, 'BIN_SIZE', userXHistoBinWidth)

    if (makePlots) then
      if (all(rowIsNull)) then
        call warning('allRowsNull',"")
      else
        !...............................................................
        ! Make plots 1 and 2.

        numNonNull = count(.not.rowIsNull)
        allocate(packedXMid(numNonNull), packedYVal(numNonNull), packedYStdDev(numNonNull))
        allocate(packedYAveErr(numNonNull), packedYStdDevErr(numNonNull))

        ! Plot 1 displays the difference between the input value of the source parameter p, and the value determined by the source-fitting procedure. This difference is plotted as a histogram against log10 of the X coordinate supplied in the column xColName of the list of detected sources (usually some kind of flux value).

!print *, 'plot 1'

        packedXMid    = pack(xMid,   (.not.rowIsNull))
        packedYVal    = pack(yVal,   (.not.rowIsNull))
        packedYStdDev = pack(yStdDev,(.not.rowIsNull))
        xLo = minval(packedXMid) / 10.0
        xHi = maxval(packedXMid) * 10.0
        write (pgplotDevice, '(a,a)') trim(plot1FileName), pgDev
!        write(xLabel, '(a,a,a)') 'log10(',trim(xColName),')'
!xLabel = trim(xColName)
xLabel = trim(xColName)//' ('//trim(xUnits)//')'
yLabel = trim(yValColName)//' ('//trim(yUnits)//')'

!        call plotInit(pgplotDevice, xLabel, trim(yValColName))
 call plotInit(pgplotDevice, xLabel, yLabel)
        call setXLog
        call plotAutoSizeX(packedXMid)
        call plotAutoSizeY(packedYVal, packedYStdDev)
        call plotAutoSizeY((/0.0/))
        call drawLinePlot((/xLo,xHi/), (/0.0,0.0/), lineStyle=L_DASHED, finished=.false.)
        call drawPointPlot(packedXMid, packedYVal, packedYStdDev)
        ! All in ssclib/pgplot_aux.

        ! Plot 2 uses the same X axis and logarithmic X binning scheme as plot 1. Two Y values are plotted for each histogram bin: (i) the average value of p_uncert and (ii) the scatter (standard deviation) in p within the bin (plus approximate error in this).

! The histogram Y values are calculated from the ratio between these numbers.

!print *, 'plot 2'

!        packedYVal    = pack(errRatio,(.not.rowIsNull))
!        packedYStdDev = pack(errInErr,(.not.rowIsNull))
        packedYAveErr    = pack(yAveErr,   (.not.rowIsNull))
        packedYStdDevErr = pack(yStdDevErr,(.not.rowIsNull))

num = 2
allocate(pointsNotLines(num), styles(num), colours(num), texts(num))
pointsNotLines = (/.true.,.true./)
styles = (/P_DEFAULT, P_CROSS/)
colours = (/1,2/)
texts = (/'Std dev (scatter) within bin'&
         ,'Ave det error within bin    '/)

        write (pgplotDevice, '(a,a)') trim(plot2FileName), pgDev

        call plotInit(pgplotDevice, xLabel, 'Standard deviation compared to average uncertainty.')
        call setXLog
        call plotAutoSizeX(packedXMid)
        call plotAutoSizeY(packedYStdDev, packedYStdDevErr)
        call plotAutoSizeY(packedYAveErr)
!        call plotAutoSizeY((/1.0/))
!        call drawLinePlot((/xLo,xHi/), (/1.0,1.0/), lineStyle=L_DASHED, finished=.false.)
        call drawPointPlot(packedXMid, packedYStdDev, packedYStdDevErr, finished=.false.)
!        call drawPointPlot(packedXMid, packedYAveErr, pointStyle=P_CROSS)
call drawPointPlot(packedXMid, packedYAveErr, pointStyle=P_CROSS, finished=.false.)
call drawLegend(pointsNotLines, styles, colours, texts, position=XY_TOPRIGHT)
call finishPlot
        ! All in ssclib/pgplot_aux.
deallocate(pointsNotLines, styles, colours, texts)
      end if
    end if

!    call release(errRatioCol)
!    call release(chiSquErrCol)
!    call release(chiSquCol)
!    call release(binnedErrCol)
!    call release(uncertInBinnedValCol)
    call release(yAveErrCol)
    call release(yStdDevErrCol)
    call release(yStdDevCol)
    call release(yValCol)
    call release(xMidCol)
    call release(histo1Tab)

    !...................................................................
    ! Frequency histogram of relative position error:

    diffRelBinSize = 0.25

!!!    if (any(pDetUncertData <= 0.0))&
!!!      call warning('',"")
!*** this is already tested above. But I should change the error to a warning, and pack.

    allocate(pDiffRel(numSrcs))
    pDiffRel = (pDiff - meanDiffP) / pDetUncertData

    call calcBinEdges(diffRelBinSize, pDiffRel, binEdges) ! in ssclib/histogram_utils.
    numHistoBins = size(binEdges)-1

    if (numHistoBins < 2)&
call error('tooFewHistoBins',"")

    allocate(pRelHisto(numHistoBins))
    call makeHistogram(pDiffRel, pRelHisto, binEdges)
    ! in ssclib/histogram_utils.

    histo2TabName = trim(pStr)//'REL'
    midColName = 'D_'//trim(pStr)//'_REL'
!    midColComment = 'Mid-bin value of ('&
!      //pDiffColName//'-<'//xDiffColName//'>)/'//xErrColName

    if (hasBlock(histoSet, histo2TabName))&
call error('table2AlreadyExists',"")

    histo2Tab = addTable(histoSet, histo2TabName, numHistoBins)
    midCol    = addColumn(histo2Tab, midColName, REAL32, 'sigmas')
!&
!      , comment=midColComment)
    nCol      = addColumn(histo2Tab, 'N', INTEGER32&
      , comment='Number of entries in the histogram bin.')
    nErrCol   = addColumn(histo2Tab, 'N_ERR', REAL32)
    nTheorCol = addColumn(histo2Tab, 'N_THEOR', REAL32&
      , comment='Theoretical N.')

    midData    => real32Data(   midCol)
    nData      =>  int32Data(     nCol)
    nErrData   => real32Data(  nErrCol)
    nTheorData => real32Data(nTheorCol)

    midData    = (binEdges(1:numHistoBins) + binEdges(2:numHistoBins+1))/2.0
    nData      = pRelHisto
    nErrData   = sqrt(real(nData))
    nTheorData = diffRelBinSize * exp(-real(midData)*real(midData)/2.0)&
      * real(sum(nData)) / RootPi / RootTwo

    call setAttribute(histo2Tab, 'BIN_SIZE', diffRelBinSize)

    if (makePlots) then
      !.................................................................
      ! Make plot 3. The intent of this plot is to display any nett bias in the difference between the input and measured values of p. The nett accuracy of p_uncert can also be assessed from this plot, since the X values are scaled by dividing by p_uncert. If p_uncert is a good measure of the scatter in points, you would expect the width of the plotted distribution to be about 1.

!print *, 'plot 3'

num = 2
allocate(pointsNotLines(num), styles(num), colours(num), texts(num))
pointsNotLines = (/.false.,.true./)
styles = (/L_DEFAULT, P_CROSS/)
colours = (/1,2/)
texts = (/'Expected          '&
         ,'Scatter wrt errors'/)

      write (pgplotDevice, '(a,a)') trim(plot3FileName), pgDev
      write(yLabel, '(a,a)') 'Frequency distribution of ', trim(yValColName)

      call plotInit(pgplotDevice, 'sigmas', yLabel)
      call plotAutoSizeY(real(nData,single), nErrData)
      call plotAutoSizeY(nTheorData)
      call drawLinePlot(midData, nTheorData, finished=.false.)
!      call drawPointPlot(midData, real(nData,single), nErrData, pointStyle=P_CROSS)
call drawPointPlot(midData, real(nData,single), nErrData, pointStyle=P_CROSS, finished=.false.)
call drawLegend(pointsNotLines, styles, colours, texts, position=XY_TOPLEFT)
call finishPlot
        ! All in ssclib/pgplot_aux.
deallocate(pointsNotLines, styles, colours, texts)
    end if

    call release(nTheorCol)
    call release(nCol)
    call release(midCol)
    call release(histo2Tab)

    deallocate(binEdges)

!    call release(pDiffCol)
!    call release(simXCol)
!    call release(xErrCol)
!    call release(xCol)
!    call release(xCol)
!deallocate(simXData)
!deallocate(xErrData)
!deallocate(xData)

deallocate(pSimData)
deallocate(pDetUncertData)
deallocate(pDetData)
deallocate(xData)

    call release(sourceTab)
    call release(sourceSet)
    call release(histoSet)

  end subroutine eimsimbias

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    histoSetName  = stringParameter('histoset')
    detSrcSetName = stringParameter('detsrclistset')
    xColName = stringParameter('xcolumn')
    pStr = stringParameter('pstr')
    userXHistoBinWidth = realParameter('xbinwidth')

    addToExistingHisto = booleanParameter('append')

    makePlots = booleanParameter('withplots')
    if (makePlots) then
      plot1FileName = stringParameter('plot1file')
      plot2FileName = stringParameter('plot2file')
      plot3FileName = stringParameter('plot3file')
    end if

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module eimsimbias_mod
