!
! Module: eimsimcompleteness_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This is a collection of routines used to reduce simulations data.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module eimsimcompleteness_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.

  use dal
  use param  
  use errorHandling

  use power_law_utils

  use dal_aux	! for addOrOpenColumn()
!  use histogram_utils	! for makeHistogram().
use l_histogram_utils
!  use pgplot_aux
use l_pgplot_aux
  use ssc_misc	! for getFreeIoUnit.

  implicit none 

  private 

  ! Command-line parameters:
  !
  character(256) ::&
    srcSpecSetName='',&
    masterHistoSetName='',&
    listOfSimSrcLists='',&
    detSrcSetName='',&
    plot1FileName='',&
    plot2FileName='',&
    plot3FileName='',&
    plot4FileName='',&
    plot5FileName=''
  real(single) ::&
    nullProbCutoff=0.0,&
    histoBinSize=0.0
  logical ::&
    addToExistingHisto=.false.,&
    makePlots=.false.

character(20) :: pgDev = '/cps'

  public eimsimcompleteness

contains 

  !---------------------------------------------------------------------
  subroutine eimsimcompleteness
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    call makeMergedSimSrcHistogram
    call calcCompletenessHistogramsPart1
    call calcCompletenessHistogramsPart2

    if (makePlots) then
      ! Make a line plot of the cumulative total of sim source flux density in cgs deg^-2. x=SIM_FLUX_HI, y=FLUX_DENS_INT; both logarithmic axes.
      !
      call makePlot1

      ! Make a line plot of the inverse-cumulative sky density of simulated sources. This is compared with theory (to make sure that the real logN-logS is as specified) and also with the number of reliable detections. Both axes logarithmic. There are 3 plots as follows:
      !		1:	x=THEORY:FLUX		y=THEORY:DENSITY
      !		2:	x=HISTO:SIM_FLUX_LO	y=HISTO:DENS_SIM_INT
      !		3:	x=HISTO:SIM_FLUX_LO	y=HISTO:DENS_DET_INT
      !
      call makePlot2

      ! Plot 3 again compares the distributions of simulated versus detected sources, but this time in differential form. The sim sources are done as a line plot, the detected as points + y-errors. Both axes logarithmic. The plots are as follows:
      !		1:	x=SIM_FLUX_MID		y=DENS_SIM
      !		2:	x=SIM_FLUX_MID		y=DENS_DET	y_err=DENS_DET_ERR
      !
      call makePlot3

      ! Plot 4 presents the completeness information as a direct ratio. This is a point plot with x=SIM_FLUX_MID, y=COMP_RATIO, y_err=COMP_RATIO_ERR; X axis logarithmic, Y axis linear.
      !
      call makePlot4

      ! Similar to plot 4, but inverse-cumulative totals plotted.
      !
      call makePlot5
    end if

  end subroutine eimsimcompleteness

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    srcSpecSetName = stringParameter('srcspecset')
    histoBinSize = realParameter('histobinsize')
    masterHistoSetName = stringParameter('histoset')
    listOfSimSrcLists = stringParameter('filelist')
    detSrcSetName   = stringParameter('detsrclistset')
    nullProbCutoff = realParameter('probcutoff')
!    addToExistingHisto = booleanParameter('append')

    makePlots = booleanParameter('withplots')
    if (makePlots) then
      plot1FileName = stringParameter('plot1file')
      plot2FileName = stringParameter('plot2file')
      plot3FileName = stringParameter('plot3file')
      plot4FileName = stringParameter('plot4file')
      plot5FileName = stringParameter('plot5file')
    end if

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine makeMergedSimSrcHistogram
  !---------------------------------------------------------------------
  ! This sub makes a histogram table which records the incidence of sim
  ! sources as a function of their rate. The bins are of equal size in
  ! log10(SIM_FLUX). Columns as follows are constructed:
  !
  !	N_SIM		Number of simulated sources in the histogram bin.
  !	NET_FLUX	Sum of SIM_FLUX in the histogram bin.
  !	LOG10_SF_LO	Lower edge of the bin in log10(SIM_FLUX).
  !	LOG10_SF_HI	Upper edge of the bin in log10(SIM_FLUX).
  !
  ! The sub is given the name of a master histogram set and the name of
  ! a text file which contains a 'ls' of all relevant sim source lists.
  ! If the master set exists at entry to the loop over sim source lists,
  ! its bin edges are used as a template to make a histogram of the new
  ! source list (extending the rows if necessary). The new histogram is
  ! added to the old and N_FIELDS is incremented. If the master set
  ! doesn't yet exist, the sub creates it at loop entry and constructs
  ! the histogram according to the passed parameter 'binsize'. The sub
  ! also looks for the SKYAREA keyword in each source list and compares
  ! it with that in the histo file. If they are different, an error dump
  ! ensues.
  !
  ! These columns are not worth plotting themselves, but are calculated
  ! because they form an easy stepping stone to the more informative
  ! results which are calculated later in the program.
  !---------------------------------------------------------------------

    implicit none

    logical :: listFileExists, histoFileExists
    integer ::&
      ioUnit, ioStatus, initialNumHistoBins, numFields, numSimSources,&
      numDetectableSimSources, oldFirstRow, oldLastRow,&
      numHistoRowsToInsertAtStart, numHistoRowsToInsertAtFinis,&
      newFirstRow, newLastRow, dummyUnderflows, dummyOverflows, i
    type(DataSetT) :: masterHistoSet, simSrcSet
    type(TableT) :: masterHistoTab, simSrcTab
    type(ColumnT) :: xLoCol, xHiCol, nCol, YcOL, simFluxCol, simInverseSensyCol
    real(single), pointer ::&
      xLo(:), xHi(:), yData(:), accumBinEdges(:), simFlux(:), simInverseSensyData(:)
    integer(int32), pointer :: nData(:)
    character(256) :: simSrcSetName
    real(single), allocatable ::&
      simInverseSensy(:), packedSimFlux(:), log10SimFlux(:)
    type(weightedHistoSingleType), allocatable :: accumHisto(:), tempHisto(:)
    logical, allocatable :: selectionMask(:)

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose&
      , "Starting the routine which makes the basic histogram of sim source flux.")

    inquire(file=listOfSimSrcLists, exist=listFileExists) ! tests
    if (.not. listFileExists)&
call error('noListFile',"")

    call getFreeIoUnit(ioUnit) ! in ssclib/ssc_misc

    open(ioUnit, iostat=ioStatus, file=listOfSimSrcLists, status='old'&
      , action='read')
    if (ioStatus > 0)&
call error('unableToOpenFile',"")

    ! If the histogram set already exists, obtain the initial bin edges from it;
    ! if not, use the data in the first sim source list to set the initial bin
    ! edges.
    !
    inquire(file=masterHistoSetName, exist=histoFileExists)

    if (addToExistingHisto .and. .not. histoFileExists)&
      call warning('noHistoFileToAddTo',"")

    if (addToExistingHisto .and. histoFileExists) then
      masterHistoSet = dataSet(masterHistoSetName, READ)
      masterHistoTab = table(masterHistoSet, 'HISTO')
      initialNumHistoBins = numberOfRows(masterHistoTab)
      xLoCol = column(masterHistoTab, 'LOG10_SF_LO')
      xHiCol = column(masterHistoTab, 'LOG10_SF_HI')
      nCol   = column(masterHistoTab, 'N_SIM')
      yCol   = column(masterHistoTab, 'NET_FLUX')
      xLo   => real32Data(xLoCol)
      xHi   => real32Data(xHiCol)
      nData => int32Data(   nCol)
      YData => real32Data(  yCol)

      allocate(accumBinEdges(initialNumHistoBins+1))
      allocate(accumHisto(1:initialNumHistoBins))
      accumBinEdges(:initialNumHistoBins)  = xLo
      accumBinEdges(initialNumHistoBins+1) = xHi(initialNumHistoBins-1)
      ! '-1' because these dal pointers start at 0.

      accumHisto%n = nData
      accumHisto%y = yData

      numFields = int32Attribute(masterHistoTab, 'N_SIM_FD')

      call release(yCol)
      call release(nCol)
      call release(xHiCol)
      call release(xLoCol)
      call release(masterHistoTab)
      call release(masterHistoSet)

    else
      ! Attempt to read 1st simSrcSetName from file:
      !
      read(ioUnit, '(a)', iostat=ioStatus) simSrcSetName
      if (ioStatus .ne. 0)&
call error('unableToReadFirstRecord',"")

      simSrcSet = dataSet(simSrcSetName, READ)
      simSrcTab = table(simSrcSet, 'SRCLIST')
      numSimSources = numberOfRows(simSrcTab)

!      maskAreaDeg1st = real64Attribute(simSrcTab, 'SKY_AREA')
!!*** should be templateTab?? Rest of this module also needs a spring clean.
!call error('fixThis','')

      simFluxCol         = column(simSrcTab, 'FLUX')
      simInverseSensyCol = column(simSrcTab, 'INV_SENSY')

      simFlux         => real32Data(simFluxCol)
      simInverseSensyData => real32Data(simInverseSensyCol)
      allocate(simInverseSensy(numSimSources))
      simInverseSensy = 0.0 ! default
      do i = 1, numSimSources
        if (.not.isNull(simInverseSensyCol,i-1))&
          simInverseSensy(i) = simInverseSensyData(i-1)
      end do

      allocate(selectionMask(numSimSources))
      selectionMask = (simInverseSensy>0.0)
      deallocate(simInverseSensy)
      numDetectableSimSources = count(selectionMask)
      if (numDetectableSimSources <= 0)&
call error('noDetectableSourcesInSet',"")

      allocate(packedSimFlux(numDetectableSimSources)&
        , log10SimFlux(numDetectableSimSources))
      packedSimFlux = pack(simFlux, selectionMask)
      if (any(packedSimFlux<=0.0))&
call error('nonPositiveSimFlux',"")

      log10SimFlux = log10(packedSimFlux)
!      deallocate(packedSimFlux, selectionMask)
      deallocate(selectionMask)

      call release(simInverseSensyCol)
      call release(simFluxCol)
      call release(simSrcTab)
      call release(simSrcSet)

      call calcBinEdges(histoBinSize, log10SimFlux, accumBinEdges)
      ! in ssclib/histogram_utils.

      initialNumHistoBins = size(accumBinEdges)-1
      allocate(accumHisto(1:initialNumHistoBins))
!, dummyWeights(numDetectableSimSources))
!      dummyWeights = 1.0
!      call makeWeightedHistogram(log10SimFlux, dummyWeights, accumBinEdges&
      call makeWeightedHistogram(log10SimFlux, packedSimFlux, accumBinEdges&
        , accumHisto, dummyUnderflows, dummyOverflows) ! in ssclib/histogram_utils.
!      ! I have used the weighted routine just to get the xLo and xHi fields of the
!      ! weightedHistoSingleType structure.
!      deallocate(log10SimFlux, dummyWeights)
      deallocate(log10SimFlux, packedSimFlux)

      numFields = 1
    end if ! histo set already exists.

    oldFirstRow = lbound(accumHisto, 1)
    oldLastRow  = ubound(accumHisto, 1)

    do
      read(ioUnit, '(a)', iostat=ioStatus) simSrcSetName
      write(errStr, '(a, i3, a)') 'Field=', numFields, ' simSrcSetName: '//trim(simSrcSetName)
      call message(AppMsg, Verbose)

      if (ioStatus > 0)&
call error('unableToReadRecord',"")

      if (ioStatus < 0)&
    exit ! EOF

      simSrcSet = dataSet(simSrcSetName, READ)
      simSrcTab = table(simSrcSet, 'SRCLIST')
      numSimSources = numberOfRows(simSrcTab)

!      maskAreaDeg = real64Attribute(simSrcTab, 'SKY_AREA')
!if (maskAreaDeg .ne. maskAreaDeg1st) call error('badMaskArea','')
!!*** change to a warning, with percent difference. It is bad practice to trigger errors on comparison of two reals.

      simFluxCol         = column(simSrcTab, 'FLUX')
      simInverseSensyCol = column(simSrcTab, 'INV_SENSY')

      simFlux         => real32Data(simFluxCol)
      simInverseSensyData => real32Data(simInverseSensyCol)
      allocate(simInverseSensy(numSimSources))
      simInverseSensy = 0.0 ! default
      do i = 1, numSimSources
        if (.not.isNull(simInverseSensyCol,i-1))&
          simInverseSensy(i) = simInverseSensyData(i-1)
      end do

      allocate(selectionMask(numSimSources))
      selectionMask = (simInverseSensy>0.0)
      deallocate(simInverseSensy)
      numDetectableSimSources = count(selectionMask)
      if (numDetectableSimSources <= 0)&
call error('noDetectableSourcesInSet',"")

      allocate(packedSimFlux(numDetectableSimSources)&
        , log10SimFlux(numDetectableSimSources))
      packedSimFlux = pack(simFlux, selectionMask)
      if (any(packedSimFlux<=0.0))&
call error('nonPositiveSimFlux',"")

      log10SimFlux = log10(packedSimFlux)
!      deallocate(packedSimFlux, selectionMask)
      deallocate(selectionMask)

      call release(simInverseSensyCol)
      call release(simFluxCol)
      call release(simSrcTab)
      call release(simSrcSet)

      numHistoRowsToInsertAtStart = 0
      if (minval(log10SimFlux) < accumHisto(oldFirstRow)%xLo) then
        numHistoRowsToInsertAtStart = ceiling((accumHisto(oldFirstRow)%xLo&
          - minval(log10SimFlux)) / histoBinSize)
      end if

      numHistoRowsToInsertAtFinis = 0
      if (maxval(log10SimFlux) > accumHisto(oldLastRow)%xHi) then
        numHistoRowsToInsertAtFinis = ceiling((maxval(log10SimFlux)&
          - accumHisto(oldLastRow)%xHi) / histoBinSize)
      end if

      newFirstRow = oldFirstRow - numHistoRowsToInsertAtStart
      newLastRow  = oldLastRow  + numHistoRowsToInsertAtFinis

      if (numHistoRowsToInsertAtStart > 0 .or. numHistoRowsToInsertAtFinis > 0)&
        call resizeBinEdges(numHistoRowsToInsertAtStart, numHistoRowsToInsertAtFinis&
          , accumBinEdges)

      allocate(tempHisto(newFirstRow:newLastRow))
!, dummyWeights(numDetectableSimSources))
!      dummyWeights = 1.0
!      call makeWeightedHistogram(log10SimFlux, dummyWeights, accumBinEdges&
      call makeWeightedHistogram(log10SimFlux, packedSimFlux, accumBinEdges&
        , tempHisto, dummyUnderflows, dummyOverflows) ! in ssclib/histogram_utils.
!      ! I have used the weighted routine just to get the xLo and xHi fields of the
!      ! weightedHistoSingleType structure.
!      deallocate(log10SimFlux, dummyWeights)
      deallocate(log10SimFlux, packedSimFlux)

      tempHisto(oldFirstRow:oldLastRow)%n = tempHisto(oldFirstRow:oldLastRow)%n&
        + accumHisto%n
      tempHisto(oldFirstRow:oldLastRow)%y = tempHisto(oldFirstRow:oldLastRow)%y&
        + accumHisto%y
!**** why not copy over all fields of the structure?

      deallocate(accumHisto)
      allocate(accumHisto(newFirstRow:newLastRow))
      accumHisto = tempHisto
      deallocate(tempHisto)

      oldFirstRow = newFirstRow
      oldLastRow = newLastRow

      numFields = numFields + 1
    end do ! loop over entries in list of sim source lists.

    close(ioUnit)

    masterHistoSet = dataSet(masterHistoSetName, CREATE) ! overwrite it if it exists.
    masterHistoTab = addTable(masterHistoSet, 'HISTO', size(accumHisto))

    yCol   = addColumn(masterHistoTab, 'NET_FLUX',    REAL32)
    nCol   = addColumn(masterHistoTab, 'N_SIM',       INTEGER32)
    xLoCol = addColumn(masterHistoTab, 'LOG10_SF_LO', REAL32)
    xHiCol = addColumn(masterHistoTab, 'LOG10_SF_HI', REAL32)
    yData => real32Data(  yCol)
    nData => int32Data(   nCol)
    xLo   => real32Data(xLoCol)
    xHi   => real32Data(xHiCol)
    yData = accumHisto%y
    nData = accumHisto%n
    xLo   = accumHisto%xLo
    xHi   = accumHisto%xHi

    call setAttribute(masterHistoTab, 'BIN_SIZE', histoBinSize)
    call setAttribute(masterHistoTab, 'N_SIM_FD', numFields&
      , comment='Number of sim fields included in the histogram.')

    call release(xHiCol)
    call release(xLoCol)
    call release(nCol)
    call release(yCol)
    call release(masterHistoTab)
    call release(masterHistoSet)

    deallocate(accumBinEdges)

  end subroutine makeMergedSimSrcHistogram

  !---------------------------------------------------------------------
  subroutine resizeBinEdges(numNewStartRows, numNewFinisRows, binEdges)
  !---------------------------------------------------------------------

    implicit none

    integer, intent(in) :: numNewStartRows, numNewFinisRows
    real(single), pointer :: binEdges(:)

    integer :: newFirstRow, newLastRow, oldFirstRow, oldLastRow, i
    real(single) :: realOldRowDiff
    real(single), allocatable :: tempBinEdges(:)

logical :: testFlag=.false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '  aaa resizeBinEdges'

    if (numNewStartRows < 0)&
      call warning('badNumNewStartRows',"")
    if (numNewFinisRows < 0)&
      call warning('badNumNewFinisRows',"")


    oldFirstRow = lbound(binEdges, 1)
    oldLastRow  = ubound(binEdges, 1)
    realOldRowDiff = real(oldLastRow - oldFirstRow)
if (testFlag) print *, '  oldFirstRow, oldLastRow, realOldRowDiff:'&
  , oldFirstRow, oldLastRow, realOldRowDiff

    newFirstRow = min(oldFirstRow, oldFirstRow-numNewStartRows)
    newLastRow  = max(oldLastRow,  oldLastRow +numNewFinisRows)

if (testFlag) print *, '  newFirstRow, newLastRow:', newFirstRow, newLastRow

    allocate(tempBinEdges(oldFirstRow:oldLastRow))
    tempBinEdges = binEdges
    deallocate(binEdges)

    allocate(binEdges(newFirstRow:newLastRow))
    do i = newFirstRow, oldFirstRow-1
      binEdges(i) = fracVal(tempBinEdges(oldFirstRow), tempBinEdges(oldLastRow)&
        , (i-oldFirstRow)/realOldRowDiff) ! in ssclib/ssc_misc.
    end do
    binEdges(oldFirstRow:oldLastRow) = tempBinEdges
    do i = oldLastRow+1, newLastRow
      binEdges(i) = fracVal(tempBinEdges(oldFirstRow), tempBinEdges(oldLastRow)&
        , (i-oldFirstRow)/realOldRowDiff) ! in ssclib/ssc_misc.
    end do

    deallocate(tempBinEdges)
if (testFlag) print *, '  zzz resizeBinEdges'

  end subroutine resizeBinEdges

  !---------------------------------------------------------------------
  subroutine calcCompletenessHistogramsPart1
  !---------------------------------------------------------------------
  ! This sub adds columns pertaining to detected sources to the histogram
  ! table which gives the incidence of sim sources as a function of their
  ! flux. (The bins are of equal size in log10(SIM_FLUX).) Various other
  ! useful columns are calculated. The final tally of columns should be as
  ! follows:
  !
  ! Pre-existing:
  !
  !	N_SIM		Number of simulated sources in the histogram bin.
  !	NET_FLUX	Sum of SIM_FLUX in the histogram bin.
  !	LOG10_SF_LO	Lower edge of the bin in log10(SIM_FLUX).
  !	LOG10_SF_HI	Upper edge of the bin in log10(SIM_FLUX).
  !
  ! New:
  !
  !	N_SIM_ERR	Square root of N_SIM.
  !	N_SIM_INT	Reverse-cumulative total of N_SIM.
  !	DENS_SIM	Average sky-density (in deg^-2) of (detectable) sim sources.
  !	DENS_SIM_ERR	Error in DENS_SIM.
  !	DENS_SIM_INT	Reverse-cumulative total of DENS_SIM.
  !	FLUX_DENS_INT	Cumulative total of NET_FLUX, divided by the sky area.
  !	FD_INT_THEOR	Value of FLUX_DENS_INT calculated from the starting logN-logS.
!  !	SIM_FLUX_INT	 SIM_FLUX at the lower edge of the bin (10.0**LOG10_SF_LO).
  !	SIM_FLUX_LO	SIM_FLUX at the lower edge of the bin (10.0**LOG10_SF_LO).
  !	SIM_FLUX_MID	SIM_FLUX at the middle of the bin.
  !	SIM_FLUX_HI	SIM_FLUX at the upper edge of the bin (10.0**LOG10_SF_HI).
  !	N_ALL_DET	All detections.
  !	N_ALL_DET_ERR	Square root of N_ALL_DET.
  !	N_ALL_DET_INT	Reverse-cumulative total of N_ALL_DET.
  !	N_TRUE_DET	Reliable detections.
  !	N_TRUE_DET_ERR	Square root of N_TRUE_DET.
  !	N_TRUE_DET_INT	Reverse-cumulative total of N_TRUE_DET.
  !	DENS_DET	Average sky-density (in deg^-2) of reliable detections.
  !	DENS_DET_ERR	Error in DENS_DET.
  !	DENS_DET_INT	Reverse-cumulative total of DENS_DET.
  !---------------------------------------------------------------------

use array_utils

    implicit none

    type(DataSetT) :: histoSet, detSrcSet, srcSpecSet
    type(TableT) :: histoTab, detSrcTab, srcSpecTab, histoTheoryTab
    integer ::&
      numBins, numSimFields, bin, numDetSources, numDetFields, numValidDetSources,&
      numReliableDetSources, numLogNLogSRows
!, row
    type(ColumnT) ::&
      nSimCol, nSimErrCol, nSimIntCol, simHistoXloCol, simHistoXhiCol,&
      netFluxCol, fluxDensIntCol, simFluxLoCol, simFluxMidCol, simFluxHiCol,&
      simSkyDensityCol, simSkyDensityErrCol, fdIntTheorCol,&
      simSkyDensityIntCol, detSimFluxCol, detSimInverseSensyCol, nTotalDetCol,&
      nTotalDetErrCol, nTotalDetIntCol, detNullProbCol, nTrueDetCol,&
      nTrueDetErrCol, nTrueDetIntCol, detSkyDensityCol, detSkyDensityErrCol,&
      detSkyDensityIntCol, histoFluxCol, histoCumDensyCol
    integer(int32), pointer ::&
      nSim(:), nSimInt(:), nTotalDet(:), nTotalDetInt(:)
    real(single), pointer ::&
      nSimErr(:), simHistoXlo(:), simHistoXhi(:),&
      simFluxLo(:), simFluxMid(:), simFluxHi(:), fdIntTheor(:),&
      simSkyDensity(:), simSkyDensityErr(:), netFlux(:), fluxDensInt(:),&
      simSkyDensityInt(:), detSimFlux(:), detSimInverseSensy(:),&
      nTotalDetErr(:), detNullProb(:), nTrueDetErr(:), detSkyDensity(:),&
      detSkyDensityErr(:), detSkyDensityInt(:), modelFluxes(:),&
      modelCumSkyDensity(:), histoFlux(:), histoCumDensy(:),&
      nTrueDet(:), nTrueDetInt(:)
    real(single), allocatable ::&
      binEdges(:), packedDetSimFlux(:), log10DetSimFlux(:)
    real(single) :: maxSimFlux, hiPower
    real(double) :: maskAreaSquareDeg
    logical, allocatable :: selectionMask(:)
    integer(int32), allocatable ::&
      detTotalHisto(:), detTrueHisto(:)

logical :: testFlag=.false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, 'aaa calcCompletenessHistogramsPart1'

    call message(AppMsg, Verbose&
      , "Starting the routine which adds detected-source data to the basic histogram.")

    !...................................................................
    ! Open histo set and get some keywords:

    histoSet = dataSet(masterHistoSetName, MODIFY)
    histoTab = table(histoSet, 'HISTO')
    numBins = numberOfRows(histoTab)

    numSimFields = int32Attribute(histoTab, 'N_SIM_FD')

if (testFlag) print *, 'bbb'

    !...................................................................
    ! N_SIM_ERR, N_SIM_INT:

    nSimCol    = column(         histoTab, 'N_SIM')
    nSimErrCol = addOrOpenColumn(histoTab, 'N_SIM_ERR', REAL32)
    nSimIntCol = addOrOpenColumn(histoTab, 'N_SIM_INT', INTEGER32)
    ! addOrOpenColumn is in ssclib/dal_aux.

    nSim    => int32Data( nSimCol)
    nSimErr => real32Data(nSimErrCol)
    nSimInt => int32Data(nSimIntCol)

    nSimErr = sqrt(real(nSim))
    nSimInt(numBins-1) = nSim(numBins-1)
    do bin = numBins-2, 0, -1
      nSimInt(bin) = nSimInt(bin+1) + nSim(bin)
    end do

!    call release(nSimIntCol)
!    call release(nSimErrCol)
!    call release(nSimCol)

    !...................................................................
!    ! SIM_FLUX_MID, SIM_FLUX_INT:
    ! SIM_FLUX_LO, SIM_FLUX_MID, SIM_FLUX_HI:

if (testFlag) print *, 'ccc'
    simHistoXloCol = column(         histoTab, 'LOG10_SF_LO')
    simHistoXhiCol = column(         histoTab, 'LOG10_SF_HI')
    simFluxLoCol   = addOrOpenColumn(histoTab, 'SIM_FLUX_LO',  REAL32)
    simFluxMidCol  = addOrOpenColumn(histoTab, 'SIM_FLUX_MID', REAL32)
    simFluxHiCol   = addOrOpenColumn(histoTab, 'SIM_FLUX_HI',  REAL32)
    ! addOrOpenColumn is in ssclib/dal_aux.

    simHistoXlo => real32Data(simHistoXloCol)
    simHistoXhi => real32Data(simHistoXhiCol)
    simFluxLo   => real32Data(  simFluxLoCol)
    simFluxMid  => real32Data( simFluxMidCol)
    simFluxHi   => real32Data(  simFluxHiCol)

    simFluxLo  = 10.0**simHistoXlo
    simFluxMid = 10.0**((simHistoXlo + simHistoXhi)/2.0)
    simFluxHi  = 10.0**simHistoXhi

!    call release(simFluxHiCol)
    call release(simFluxMidCol)
    call release(simFluxLoCol)
!    call release(simHistoXhiCol)
!    call release(simHistoXloCol)

    allocate(binEdges(numBins+1))
    binEdges(:numBins) = simHistoXlo
    binEdges(numBins+1) = simHistoXhi(numBins-1)

    maxSimFlux = maxval(simFluxMid)

    !...................................................................
    ! DENS_SIM, DENS_SIM_ERR, DENS_SIM_INT:

if (testFlag) print *, 'ddd'
    detSrcSet = dataSet(detSrcSetName, READ)
    detSrcTab = table(detSrcSet, 'SRCLIST')
    numDetSources = numberOfRows(detSrcTab)
    maskAreaSquareDeg = real64Attribute(detSrcTab, 'SKY_AREA')
    numDetFields      = int32Attribute( detSrcTab, 'N_FIELDS')

    simSkyDensityCol    = addOrOpenColumn(histoTab, 'DENS_SIM', REAL32&
      , 'sources/deg^2')
    simSkyDensityErrCol = addOrOpenColumn(histoTab, 'DENS_SIM_ERR', REAL32&
      , 'sources/deg^2')
    simSkyDensityIntCol = addOrOpenColumn(histoTab, 'DENS_SIM_INT', REAL32&
      , 'sources/deg^2')
    ! addOrOpenColumn is in ssclib/dal_aux.

    simSkyDensity    => real32Data(simSkyDensityCol)
    simSkyDensityErr => real32Data(simSkyDensityErrCol)
    simSkyDensityInt => real32Data(simSkyDensityIntCol)

    simSkyDensity    = real(nSim)    / maskAreaSquareDeg / real(numSimFields)
    simSkyDensityErr =      nSimErr  / maskAreaSquareDeg / real(numSimFields)
    simSkyDensityInt = real(nSimInt) / maskAreaSquareDeg / real(numSimFields)

    call release(simSkyDensityIntCol)
    call release(simSkyDensityErrCol)
    call release(simSkyDensityCol)
!    call release(detSrcTab)
!    call release(detSrcSet)
    call release(nSimIntCol)
    call release(nSimErrCol)
    call release(nSimCol)

    !...................................................................
    ! FLUX_DENS_INT, FD_INT_THEOR:

    netFluxCol     = column(         histoTab, 'NET_FLUX')
    fluxDensIntCol = addOrOpenColumn(histoTab, 'FLUX_DENS_INT', REAL32)
    fdIntTheorCol  = addOrOpenColumn(histoTab, 'FD_INT_THEOR',  REAL32)
    ! addOrOpenColumn is in ssclib/dal_aux.

    netFlux     => real32Data(netFluxCol)
    fluxDensInt => real32Data(fluxDensIntCol)
    fdIntTheor  => real32Data(fdIntTheorCol)

    fluxDensInt(0) = netFlux(0)
    do bin = 1, numBins-1
      fluxDensInt(bin) = fluxDensInt(bin-1) + netFlux(bin)
    end do

    fluxDensInt = fluxDensInt / maskAreaSquareDeg / real(numSimFields)

    srcSpecSet = dataSet(srcSpecSetName, READ)
    srcSpecTab = table(srcSpecSet, 'SRCSPECS')
    call calcCumFluxDensities(srcSpecTab, simFluxHi, fdIntTheor) ! in power_law_utils
    call release(srcSpecTab)
    call release(srcSpecSet)

    call release(fdIntTheorCol)
    call release(fluxDensIntCol)
    call release(netFluxCol)
    call release(simFluxHiCol)

    !...................................................................
    ! Add det list keywords:

!    detSrcSet = dataSet(totalDetSrcSetName, READ)
!    detSrcTab = table(detSrcSet, 'SRCLIST')
!    numDetSrc = numberOfRows(detSrcTab)
!
!    call setAttribute(histoTab, 'DET_SKAR'&
!      , real64Attribute(detSrcTab, 'SKY_AREA'), 'deg^2'&
!      , 'Sky area of det sources.')
!    call setAttribute(histoTab, 'N_DET_FD'&
!      , int32Attribute(detSrcTab, 'N_FIELDS')&
!      , comment='Number of det fields included in the histogram.')
!
!    filterName = stringAttribute(detSrcSet, 'FILTER')
!!    filtrId = filterId(filterName) ! in caloalutils/src/caloaldefs.f90
!    call setAttribute(histoTab, 'FILTER', filterName)
!!    call CAL_setState(filterId=filtrId)

    !...................................................................
    ! N_ALL_DET, N_ALL_DET_ERR, N_ALL_DET_INT:

if (testFlag) print *, 'eee'
    detSimFluxCol         = column(detSrcTab, 'SIM_FLUX')
    detSimInverseSensyCol = column(detSrcTab, 'SIM_INV_SENSY')

    detSimFlux         => real32Data(detSimFluxCol)
    detSimInverseSensy => real32Data(detSimInverseSensyCol)

    allocate(selectionMask(numDetSources))
    selectionMask = (detSimInverseSensy > 0.0)
    numValidDetSources = count(selectionMask)
    allocate(packedDetSimFlux(numValidDetSources), log10DetSimFlux(numValidDetSources))
    packedDetSimFlux = pack(detSimFlux, selectionMask)
if (any(packedDetSimFlux<=0.0)) call error('',"")
    log10DetSimFlux = log10(packedDetSimFlux)
    deallocate(packedDetSimFlux)

    allocate(detTotalHisto(numBins))
    call makeHistogram(log10DetSimFlux, detTotalHisto, binEdges)
    ! in ssclib/histogram_utils.

    deallocate(log10DetSimFlux)

    nTotalDetCol    = addOrOpenColumn(histoTab, 'N_ALL_DET',     INTEGER32)
    nTotalDetErrCol = addOrOpenColumn(histoTab, 'N_ALL_DET_ERR', REAL32)
    nTotalDetIntCol = addOrOpenColumn(histoTab, 'N_ALL_DET_INT', INTEGER32)
    ! addOrOpenColumn is in ssclib/dal_aux.

    nTotalDet    => int32Data( nTotalDetCol)
    nTotalDetErr => real32Data(nTotalDetErrCol)
    nTotalDetInt => int32Data( nTotalDetIntCol)

    nTotalDet    = detTotalHisto
    nTotalDetErr = sqrt(real(nTotalDet))

    nTotalDetInt(numBins-1) = nTotalDet(numBins-1)
    do bin = numBins-2, 0, -1
      nTotalDetInt(bin) = nTotalDetInt(bin+1) + nTotalDet(bin)
    end do

    call release(nTotalDetIntCol)
    call release(nTotalDetErrCol)
    call release(nTotalDetCol)

!    call release(detSimInverseSensyCol)
    call release(detSimFluxCol)
!    call release(detSrcTab)
!    call release(detSrcSet)

    !...................................................................
    ! N_TRUE_DET, N_TRUE_DET_ERR, N_TRUE_DET_INT:
if (testFlag) print *, 'fff'

!    detSrcSet = dataSet(trueDetSetName, READ)
!    detSrcTab = table(detSrcSet, 'SRCLIST')

!    log10SimFlux => real32Data(column(detSrcTab, 'LOG10_SR'))
    detNullProbCol = column(detSrcTab, 'MATCH_PNULL')
    detNullProb => real32Data(detNullProbCol)

    selectionMask = (detSimInverseSensy > 0.0 .and. detNullProb < nullProbCutoff)
    numReliableDetSources = count(selectionMask)

    if (numReliableDetSources <= 0) then
      write(errStr, '(a, f6.4, a, e9.5, a)') "There are no detected sources &
        &with MATCH_PNULL less than the cutoff value ", nullProbCutoff&
        , ". Smallest MATCH_PNULL value found was ", minval(detNullProb), "."
call error('noGoodSources')

    else if (numReliableDetSources == 1) then
      write(errStr, '(a, f6.4, a)') "Only one detected source was found with&
        & MATCH_PNULL less than the cutoff value ", nullProbCutoff, "."
call error('tooFewGoodSources')
    end if

    allocate(packedDetSimFlux(numReliableDetSources), log10DetSimFlux(numReliableDetSources))
    packedDetSimFlux = pack(detSimFlux, selectionMask)
!if (any(packedDetSimFlux<=0.0)) call error('',"")
    log10DetSimFlux = log10(packedDetSimFlux)
    deallocate(packedDetSimFlux)
if (testFlag) print *, 'size(log10DetSimFlux):', size(log10DetSimFlux)

    allocate(detTrueHisto(numBins))
    call makeHistogram(log10DetSimFlux, detTrueHisto, binEdges)
!    call release(detSrcSet)

    call release(detSimInverseSensyCol)
    call release(detNullProbCol)
    call release(detSrcTab)
    call release(detSrcSet)

    nTrueDetCol    = addOrOpenColumn(histoTab, 'N_TRUE_DET', REAL32)
    nTrueDetErrCol = addOrOpenColumn(histoTab, 'N_TRUE_DET_ERR', REAL32)
    nTrueDetIntCol = addOrOpenColumn(histoTab, 'N_TRUE_DET_INT', REAL32)
    ! addOrOpenColumn is in ssclib/dal_aux.

    nTrueDet    => real32Data( nTrueDetCol)
    nTrueDetErr => real32Data(nTrueDetErrCol)
    nTrueDetInt => real32Data( nTrueDetIntCol)

    nTrueDet = (detTrueHisto - detTotalHisto * nullProbCutoff)&
      / (1.0 - nullProbCutoff)

    ! Because the false detections are spread evenly at all values of nullProb, thus the initial numbers of detections with nullProb below the cutoff comprises not the total number R of reliable detections, out of a total A, but R' = R + nullProbCutoff * (A - R). The equation above follows.

    where(nTrueDet < 0.0) nTrueDet = 0.0 ! can happen if detTrueHisto < detTotalHisto * nullProbCutoff.

    nTrueDetErr = sqrt(real(detTotalHisto * nullProbCutoff * nullProbCutoff&
      + detTrueHisto)) / (1.0 - nullProbCutoff)

    nTrueDetInt(numBins-1) = nTrueDet(numBins-1)
    do bin = numBins-2, 0, -1
      nTrueDetInt(bin) = nTrueDetInt(bin+1) + nTrueDet(bin)
    end do

!    call release(nTrueDetIntCol)
!    call release(nTrueDetErrCol)
!    call release(nTrueDetCol)

    !...................................................................
    ! DENS_DET, DENS_DET_ERR, DENS_DET_INT:
if (testFlag) print *, 'ggg'

    detSkyDensityCol    = addOrOpenColumn(histoTab, 'DENS_DET', REAL32&
      , 'sources/deg^2')
    detSkyDensityErrCol = addOrOpenColumn(histoTab, 'DENS_DET_ERR', REAL32&
      , 'sources/deg^2')
    detSkyDensityIntCol = addOrOpenColumn(histoTab, 'DENS_DET_INT', REAL32&
      , 'sources/deg^2')
    ! addOrOpenColumn is in ssclib/dal_aux.

    detSkyDensity    => real32Data(detSkyDensityCol)
    detSkyDensityErr => real32Data(detSkyDensityErrCol)
    detSkyDensityInt => real32Data(detSkyDensityIntCol)

    detSkyDensity    = real(nTrueDet)    / maskAreaSquareDeg / real(numDetFields)
    detSkyDensityErr =      nTrueDetErr  / maskAreaSquareDeg / real(numDetFields)
    detSkyDensityInt = real(nTrueDetInt) / maskAreaSquareDeg / real(numDetFields)

    call release(detSkyDensityIntCol)
    call release(detSkyDensityErrCol)
    call release(detSkyDensityCol)
    call release(nTrueDetIntCol)
    call release(nTrueDetErrCol)
    call release(nTrueDetCol)

    call release(histoTab)

!    !...................................................................
!    ! Open the input logN-logS distribution and get energy limits:
!
!    logNLogSSet = dataSet(logNLogSSetName, READ)
!    logNLogSTab = table(logNLogSSet, 'LOGNLOGS')
!    EeVloFlux = real64Attribute(logNLogSTab, 'E_MIN')
!    EeVhiFlux = real64Attribute(logNLogSTab, 'E_MAX')
!    photonSpectralIndex = real64Attribute(logNLogSTab, 'SPECINDX')
!
!    !...................................................................
!    ! FLUX_MID, FLUX_INT:
!
!    allocate(amplitudeMid(numBins), amplitudeInt(numBins))
!!    filtrId = filterId(filterName) ! in caloalutils/src/caloaldefs.f90
!
!    EeVloRate = 200.0d0		! Appropriate for the RATE values,
!    EeVhiRate = 12000.0d0	! which are those in band 8.
!
!    call setAttribute(histoTab, 'ERATEMIN', EeVloRate, 'eV'&
!      , 'Valid for the RATE values.')
!    call setAttribute(histoTab, 'ERATEMAX', EeVhiRate, 'eV'&
!      , 'Valid for the RATE values.')
!
!    amplitudeMid = rateToFluxDensity0(photonSpectralIndex&
!     , EeVloRate, EeVhiRate, simFluxMid)!, filtrId)
!
!    amplitudeInt = rateToFluxDensity0(photonSpectralIndex&
!     , EeVloRate, EeVhiRate, simFluxInt)!, filtrId)
!
!    call setAttribute(histoTab, 'EFLUXMIN', EeVloFlux, 'eV'&
!      , 'Valid for the FLUX and FLUX_INT values.')
!    call setAttribute(histoTab, 'EFLUXMAX', EeVhiFlux, 'eV'&
!      , 'Valid for the FLUX and FLUX_INT values.')
!
!!    fluxSoftMid => addOrOpenColumnDataReal32(histoTab, 'FLUX_MID'&
!!      , 'erg/cm^2/s', 'Valid for the E bounds and spectral index given.')
!fluxSoftMidCol = addOrOpenColumn(histoTab, 'FLUX_MID', REAL32&
!  , 'erg/cm^2/s', 'Valid for the E bounds and spectral index given.')
!    ! addOrOpenColumn is in ssclib/dal_aux.
!fluxSoftMid => real32Data(fluxSoftMidCol)
!
!!    fluxSoftInt => addOrOpenColumnDataReal32(histoTab, 'FLUX_INT'&
!!      , 'erg/cm^2/s', 'Valid for the E bounds and spectral index given.')
!fluxSoftIntCol = addOrOpenColumn(histoTab, 'FLUX_INT', REAL32&
!  , 'erg/cm^2/s', 'Valid for the E bounds and spectral index given.')
!    ! addOrOpenColumn is in ssclib/dal_aux.
!fluxSoftInt => real32Data(fluxSoftIntCol)
!
!    fluxSoftMid = fluxDensity0ToFlux(photonSpectralIndex&
!      , EeVloFlux, EeVhiFlux, amplitudeMid)
!
!    fluxSoftInt = fluxDensity0ToFlux(photonSpectralIndex&
!      , EeVloFlux, EeVhiFlux, amplitudeInt)
!
!    deallocate(amplitudeMid, amplitudeInt)

    !...................................................................
    ! Add an extension containing the input (theoretical) logN-logS
    ! distribution:

if (testFlag) print *, 'hhh'

    srcSpecSet = dataSet(srcSpecSetName, READ)
    srcSpecTab = table(srcSpecSet, 'SRCSPECS')
    call getLogNlogS(srcSpecTab, modelFluxes, modelCumSkyDensity&
      , hiPower) ! in power_law_utils

    numLogNLogSRows = size(modelFluxes)

    if (hasBlock(histoSet, 'THEORY')) call deleteBlock(histoSet, 'THEORY')
    histoTheoryTab = addtable(histoSet, 'THEORY', numLogNLogSRows+1)

    call setAttributes(attributable(histoTheoryTab)&
      , attributable(srcSpecTab))

    histoFluxCol     = addColumn(histoTheoryTab, 'FLUX',    REAL32, 'erg/cm^2/s')
    histoCumDensyCol = addColumn(histoTheoryTab, 'DENSITY', REAL32, 'sources/deg^2'&
      , comment='Sky density of sources brighter than the given flux.')

    histoFlux     => real32Data(histoFluxCol)
    histoCumDensy => real32Data(histoCumDensyCol)

    histoFlux(0:numLogNLogSRows-1) = modelFluxes
    histoFlux(numLogNLogSRows) = 10.0 * maxSimFlux ! the factor of 10
    ! is arbitrary - we just want it to be off the RH edge of any reasonable plot.

    histoCumDensy(0:numLogNLogSRows-1) = modelCumSkyDensity
    histoCumDensy(numLogNLogSRows) = modelCumSkyDensity(numLogNLogSRows-1)&
      * (histoFlux(numLogNLogSRows)&
      / modelFluxes(numLogNLogSRows-1))**hiPower

    deallocate(modelFluxes, modelCumSkyDensity)

    call release(histoCumDensyCol)
    call release(histoFluxCol)
    call release(histoTheoryTab)
    call release(srcSpecTab)
    call release(srcSpecSet)

    call release(histoSet)

if (testFlag) print *, 'zzz calcCompletenessHistogramsPart1'

  end subroutine calcCompletenessHistogramsPart1

  !---------------------------------------------------------------------
  subroutine calcCompletenessHistogramsPart2
  !---------------------------------------------------------------------
  ! function=completeness2
  !
  ! This sub adds columns pertaining to detected sources to the histogram table giving the incidence of sim sources as a function of their rate. (The bins are of equal size in log10(SIM_FLUX).) Various other useful columns are calculated. The final tally of columns should be as follows:
  !
  ! Pre-existing:
  !
  !	N_SIM		Number of simulated sources in the histogram bin.
  !	NET_FLUX	Sum of SIM_FLUX in the histogram bin.
  !	LOG10_SF_LO	Lower edge of the bin in log10(SIM_FLUX).
  !	LOG10_SF_HI	Upper edge of the bin in log10(SIM_FLUX).
  !	N_SIM_ERR	Square root of N_SIM.
  !	N_SIM_INT	Reverse-cumulative total of N_SIM.
  !	DENS_SIM	Average sky-density (in deg^-2) of sim sources.
  !	DENS_SIM_ERR	Error in DENS_SIM.
  !	DENS_SIM_INT	Reverse-cumulative total of DENS_SIM.
  !	FLUX_DENS_INT	Cumulative total of NET_FLUX, divided by the sky area.
!  !	SIM_FLUX_INT	 SIM_FLUX at the lower edge of the bin. Use this for the x coordinate when plotting against integrated quantities.
  !	SIM_FLUX_LO	SIM_FLUX at the lower edge of the bin. Use this for the x coordinate when plotting against reverse-cumulative quantities.
  !	SIM_FLUX_MID	SIM_FLUX at the middle of the bin. Use this for the x coordinate when plotting against differential quantities.
  !	SIM_FLUX_HI	SIM_FLUX at the upper edge of the bin. Use this for the x coordinate when plotting against forward-cumulative quantities.
  !	N_ALL_DET	All detections.
  !	N_ALL_DET_ERR	Square root of N_ALL_DET.
  !	N_ALL_DET_INT	Reverse-cumulative total of N_ALL_DET.
  !	N_TRUE_DET	Reliable detections.
  !	N_TRUE_DET_ERR	Square root of N_TRUE_DET.
  !	N_TRUE_DET_INT	Reverse-cumulative total of N_TRUE_DET.
  !	DENS_DET	Average sky-density (in deg^-2) of reliable detections.
  !	DENS_DET_ERR	Error in DENS_DET.
  !	DENS_DET_INT	Reverse-cumulative total of DENS_DET.
  !
  ! New:
  !
  !	COMP_RATIO	DENS_DET/DENS_SIM.
  !	COMP_RATIO_ERR	Error in COMP_RATIO.
  !	COMP_RATIO_INT	DENS_DET_INT/DENS_SIM_INT.

    implicit none

    type(DataSetT) :: histoSet
    type(TableT) :: histoTab
    integer :: numBins, bin
    type(ColumnT) ::&
      simSkyDensityCol, detSkyDensityCol, simSkyDensityIntCol,&
      detSkyDensityIntCol, simSkyDensityErrCol, detSkyDensityErrCol,&
      completenessRatioCol, completenessRatioErrCol, completenessRatioIntCol
    real(single), pointer ::&
      simSkyDensity(:), simSkyDensityErr(:), simSkyDensityInt(:),&
      completenessRatio(:), completenessRatioErr(:), completenessRatioInt(:),&
      detSkyDensity(:), detSkyDensityErr(:), detSkyDensityInt(:)

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose&
      , "Starting the routine which calculates the completeness histogram.")

    histoSet = dataSet(masterHistoSetName, MODIFY)
    histoTab = table(histoSet, 'HISTO', MODIFY)
    numBins = numberOfRows(histoTab)

    simSkyDensityCol    = column(histoTab, 'DENS_SIM')
    detSkyDensityCol    = column(histoTab, 'DENS_DET')
    simSkyDensityIntCol = column(histoTab, 'DENS_SIM_INT')
    detSkyDensityIntCol = column(histoTab, 'DENS_DET_INT')
    simSkyDensityErrCol = column(histoTab, 'DENS_SIM_ERR')
    detSkyDensityErrCol = column(histoTab, 'DENS_DET_ERR')

    simSkyDensity    => real32Data(simSkyDensityCol)
    detSkyDensity    => real32Data(detSkyDensityCol)
    simSkyDensityInt => real32Data(simSkyDensityIntCol)
    detSkyDensityInt => real32Data(detSkyDensityIntCol)
    simSkyDensityErr => real32Data(simSkyDensityErrCol)
    detSkyDensityErr => real32Data(detSkyDensityErrCol)

    completenessRatioCol    = addOrOpenColumn(histoTab, 'COMP_RATIO'&
      , REAL32)
    completenessRatioIntCol = addOrOpenColumn(histoTab, 'COMP_RATIO_INT'&
      , REAL32)
    completenessRatioErrCol = addOrOpenColumn(histoTab, 'COMP_RATIO_ERR'&
      , REAL32)
    ! addOrOpenColumn is in ssclib/dal_aux.

    do bin = 0, numBins - 1
      if (simSkyDensity(bin) <= 0.0)&
        call setToNull(completenessRatioCol, bin)
    end do
    completenessRatio => real32Data(completenessRatioCol)
    where(simSkyDensity > 0.0)&
      completenessRatio = min(1.0, max(0.0&
        , real(detSkyDensity) / simSkyDensity))

    do bin = 0, numBins - 1
      if (simSkyDensityInt(bin) <= 0.0)&
        call setToNull(completenessRatioIntCol, bin)
    end do
    completenessRatioInt => real32Data(completenessRatioIntCol)
    where(simSkyDensityInt > 0.0)&
      completenessRatioInt&
        = min(1.0, max(0.0, real(detSkyDensityInt) / simSkyDensityInt))

    completenessRatioErr => real32Data(completenessRatioErrCol)
!    where(simSkyDensityErr > 0.0 .and. detSkyDensityErr > 0.0)
where(simSkyDensity > 0.0 .and. detSkyDensity > 0.0)
      completenessRatioErr = completenessRatio&
      *sqrt(detSkyDensityErr * detSkyDensityErr / detSkyDensity / detSkyDensity&
      +     simSkyDensityErr * simSkyDensityErr / simSkyDensity / simSkyDensity)
    elsewhere
      completenessRatioErr = 0.0
    end where

    call release(completenessRatioErrCol)
    call release(completenessRatioIntCol)
    call release(completenessRatioCol)
    call release(detSkyDensityErrCol)
    call release(simSkyDensityErrCol)
    call release(detSkyDensityIntCol)
    call release(simSkyDensityIntCol)
    call release(detSkyDensityCol)
    call release(simSkyDensityCol)

    call release(histoTab)
    call release(histoSet)

  end subroutine calcCompletenessHistogramsPart2

  !---------------------------------------------------------------------
  subroutine makePlot1
  !---------------------------------------------------------------------
  ! Make a line plot of the cumulative total of sim source flux density in cgs deg^-2. x=SIM_FLUX_HI, y=FLUX_DENS_INT (actual) and y=FD_INT_THEOR (from logN-logS); both logarithmic axes. This is useful for checking against faintbackcalc.

    implicit none

    type(DataSetT) :: histoSet
    type(TableT) :: histoTab
    real(single), pointer :: xVals(:), yVals(:), yValsTheor(:)
    character(300) :: pgplotDevice
    logical, allocatable :: pointsNotLines(:)
    integer, allocatable :: styles(:), colours(:)
    character(6), allocatable :: texts(:)
    integer :: num

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, "Making plot 1.")

    histoSet = dataSet(masterHistoSetName, READ)
    histoTab = table(histoSet, 'HISTO')

    call readColDataToPtr(histoTab, 'SIM_FLUX_HI',   xVals)
    call readColDataToPtr(histoTab, 'FLUX_DENS_INT', yVals)
    call readColDataToPtr(histoTab, 'FD_INT_THEOR',  yValsTheor)
    ! In ssclib/dal_aux.

num = 2
allocate(pointsNotLines(num), styles(num), colours(num), texts(num))
pointsNotLines = (/.false.,.false./)
styles = (/L_DEFAULT, L_DEFAULT/)
colours = (/1,2/)
texts = (/'Theory','Actual'/)

    write (pgplotDevice, '(a,a)') trim(plot1FileName), pgDev

!    call plotInit(pgplotDevice, 'FLUX (cgs)', 'Cumulative flux density (cgs/deg^2)')
 call plotInit(pgplotDevice, 'FLUX (cgs)', 'Cumulative flux density (cgs deg\u-2\d)', 'Performance of the simulation.')
    call setXLog
    call setYLog
    call drawLinePlot(xVals, yValsTheor, finished=.false.)
!    call drawLinePlot(xVals, yVals)
call drawLinePlot(xVals, yVals, finished=.false.)
call drawLegend(pointsNotLines, styles, colours, texts, position=XY_TOPLEFT)
call finishPlot
        ! All in ssclib/pgplot_aux.
deallocate(pointsNotLines, styles, colours, texts)

    deallocate(yValsTheor, yVals, xVals)

    call release(histoTab)
    call release(histoSet)

  end subroutine makePlot1

  !---------------------------------------------------------------------
  subroutine makePlot2
  !---------------------------------------------------------------------
  ! Make a line plot of the inverse-cumulative sky density of simulated sources. This is compared with theory (to make sure that the real logN-logS is as specified) and also with the number of reliable detections. Both axes logarithmic. There are 3 plots as follows:
  !	1:	x=THEORY:FLUX		y=THEORY:DENSITY
  !	2:	x=HISTO:SIM_FLUX_LO	y=HISTO:DENS_SIM_INT
  !	3:	x=HISTO:SIM_FLUX_LO	y=HISTO:DENS_DET_INT

    implicit none

    type(DataSetT) :: histoSet
    type(TableT) :: theoryTab, histoTab
    real(single), pointer ::&
      simDensTheory(:), fluxTheory(:), detDensActual(:), simDensActual(:), fluxActual(:)
    character(300) :: pgplotDevice
    logical, allocatable :: pointsNotLines(:)
    integer, allocatable :: styles(:), colours(:)
    character(20), allocatable :: texts(:)
    integer :: num

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, "Making plot 2.")

    histoSet = dataSet(masterHistoSetName, READ)
    theoryTab = table(histoSet, 'THEORY')

    call readColDataToPtr(theoryTab, 'FLUX',    fluxTheory)
    call readColDataToPtr(theoryTab, 'DENSITY', simDensTheory)

    call release(theoryTab)
    histoTab = table(histoSet, 'HISTO')

    call readColDataToPtr(histoTab,  'SIM_FLUX_LO', fluxActual)
    call readColDataToPtr(histoTab,  'DENS_SIM_INT', simDensActual)
    call readColDataToPtr(histoTab,  'DENS_DET_INT', detDensActual)
    ! In ssclib/dal_aux.

num = 3
allocate(pointsNotLines(num), styles(num), colours(num), texts(num))
pointsNotLines = (/.false.,.false.,.false./)
styles = (/L_DEFAULT, L_DEFAULT, L_DEFAULT/)
colours = (/1,2,3/)
texts = (/'Simulation: theory'&
         ,'Simulation: actual'&
         ,'Good detections   '/)

    write (pgplotDevice, '(a,a)') trim(plot2FileName), pgDev

!    call plotInit(pgplotDevice, 'FLUX (cgs)', 'cumulative source density (deg^-2)')
 call plotInit(pgplotDevice, 'FLUX (cgs)', 'Cumulative source density (deg\u-2\d)')
    call setXLog
    call setYLog
    call plotAutoSizeX(fluxActual)
    call plotAutoSizeY(simDensActual)
    call drawLinePlot(fluxTheory, simDensTheory, finished=.false.)
    call drawLinePlot(fluxActual, simDensActual, finished=.false.)
!    call drawLinePlot(fluxActual, detDensActual)
call drawLinePlot(fluxActual, detDensActual, finished=.false.)
call drawLegend(pointsNotLines, styles, colours, texts, position=XY_TOPRIGHT)
call finishPlot
        ! All in ssclib/pgplot_aux.
deallocate(pointsNotLines, styles, colours, texts)

    deallocate(detDensActual, simDensActual, fluxActual)
    deallocate(simDensTheory, fluxTheory)

    call release(histoTab)
    call release(histoSet)

  end subroutine makePlot2

  !---------------------------------------------------------------------
  subroutine makePlot3
  !---------------------------------------------------------------------
  ! Plot 3 again compares the distributions of simulated versus detected sources, but this time in differential form. The sim sources are done as a line plot, the detected as points + y-errors. Both axes logarithmic. The plots are as follows:
  !	1:	x=SIM_FLUX_MID		y=DENS_SIM
  !	2:	x=SIM_FLUX_MID		y=DENS_DET	y_err=DENS_DET_ERR

    implicit none

    type(DataSetT) :: histoSet
    type(TableT) :: histoTab
    real(single), pointer :: xVals(:), simYVals(:), detYVals(:), detYErrs(:)
    character(300) :: pgplotDevice
    real(single) :: yLo, yHi
!integer :: i
    logical, allocatable :: pointsNotLines(:)
    integer, allocatable :: styles(:), colours(:)
    character(20), allocatable :: texts(:)
    integer :: num

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, "Making plot 3.")

    histoSet = dataSet(masterHistoSetName, READ)
    histoTab = table(histoSet, 'HISTO')

    call readColDataToPtr(histoTab, 'SIM_FLUX_MID', xVals)
    call readColDataToPtr(histoTab, 'DENS_SIM',     simYVals)
    call readColDataToPtr(histoTab, 'DENS_DET',     detYVals)
    call readColDataToPtr(histoTab, 'DENS_DET_ERR', detYErrs)
    ! In ssclib/dal_aux.

    yLo = minval(simYVals, (simYVals>0.0))
    yHi = maxval(simYVals, (simYVals>0.0))

num = 2
allocate(pointsNotLines(num), styles(num), colours(num), texts(num))
pointsNotLines = (/.false.,.true./)
styles = (/L_DEFAULT, P_DEFAULT/)
colours = (/1,2/)
texts = (/'Simulated sources'&
         ,'Good detections  '/)

    write (pgplotDevice, '(a,a)') trim(plot3FileName), pgDev

!    call plotInit(pgplotDevice, 'FLUX (cgs)', 'Source sky density (cgs^-1 deg^-2)')
 call plotInit(pgplotDevice, 'FLUX (cgs)', 'Source sky density (cgs\u-1\d deg\u-2\d)')
    call setXLog
    call setYLog
    call plotAutoSizeY(detYVals, detYErrs)
    call plotAutoSizeY((/yLo,yHi/))
    call drawLinePlot(xVals, simYVals, finished=.false.)
!    call drawPointPlot(xVals, detYVals, detYErrs)
call drawPointPlot(xVals, detYVals, detYErrs, finished=.false.)
call drawLegend(pointsNotLines, styles, colours, texts, position=XY_TOPRIGHT)
call finishPlot
    ! All in ssclib/pgplot_aux.
deallocate(pointsNotLines, styles, colours, texts)

    deallocate(detYErrs, detYVals, simYVals, xVals)

    call release(histoTab)
    call release(histoSet)

  end subroutine makePlot3

  !---------------------------------------------------------------------
  subroutine makePlot4
  !---------------------------------------------------------------------
  ! Plot 4 presents the completeness information as a direct ratio. This is a point plot with x=SIM_FLUX_MID, y=COMP_RATIO, y_err=COMP_RATIO_ERR; X axis logarithmic, Y axis linear.

    implicit none

    type(DataSetT) :: histoSet
    type(TableT) :: histoTab
    integer :: numRows, row, numNonNullRows
!, i
    type(ColumnT) :: compCol
    logical, allocatable :: rowIsNull(:)
    real(single), pointer :: xVals(:), yVals(:), yErrs(:)
    real(single), allocatable :: packedXVals(:), packedYVals(:), packedYErrs(:)
    character(300) :: pgplotDevice

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, "Making plot 4.")

    histoSet = dataSet(masterHistoSetName, READ)
    histoTab = table(histoSet, 'HISTO')
    numRows = numberOfRows(histoTab)

    compCol = column(histoTab, 'COMP_RATIO')
    allocate(rowIsNull(numRows))
    rowIsNull = .false. ! default
    do row = 0, numRows-1
      if (isNull(compCol, row)) rowIsNull(row+1) = .true.
    end do
    call release(compCol)

    numNonNullRows = count(.not.rowIsNull)
    if (numNonNullRows<=0) then
      call warning('noNonNullRows',"")
    else
      call readColDataToPtr(histoTab, 'SIM_FLUX_MID',   xVals)
      call readColDataToPtr(histoTab, 'COMP_RATIO',     yVals)
      call readColDataToPtr(histoTab, 'COMP_RATIO_ERR', yErrs)
      ! In ssclib/dal_aux.

      allocate(packedXVals(numNonNullRows), packedYVals(numNonNullRows))
      allocate(packedYErrs(numNonNullRows))
      packedXVals = pack(xVals, .not.rowIsNull)
      packedYVals = pack(yVals, .not.rowIsNull)
      packedYErrs = pack(yErrs, .not.rowIsNull)

      write (pgplotDevice, '(a,a)') trim(plot4FileName), pgDev

      call plotInit(pgplotDevice, 'FLUX (cgs)', 'Fraction of reliable detections.')
      call setXLog
      call drawPointPlot(packedXVals, packedYVals, packedYErrs, pointStyle=P_CROSS)
      ! All in ssclib/pgplot_aux.

      deallocate(yErrs, yVals, xVals)
    end if

    call release(histoTab)
    call release(histoSet)

  end subroutine makePlot4

  !---------------------------------------------------------------------
  subroutine makePlot5
  !---------------------------------------------------------------------
  ! Similar to plot 4, but inverse-cumulative totals plotted.

    implicit none

    type(DataSetT) :: histoSet
    type(TableT) :: histoTab
    integer :: numRows, row, rowFirstNull
    type(ColumnT) :: compCol
!    logical :: anyNullFound
    real(single), pointer :: xVals(:), yVals(:)
    real(single), allocatable :: packedXVals(:), packedYVals(:)
    character(300) :: pgplotDevice

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, "Making plot 5.")

    histoSet = dataSet(masterHistoSetName, READ)
    histoTab = table(histoSet, 'HISTO')
    numRows = numberOfRows(histoTab)

    compCol = column(histoTab, 'COMP_RATIO_INT')
!    anyNullFound = .false. ! default
    rowFirstNull = numRows ! default
    do row = 0, numRows-1
      if (isNull(compCol, row)) then
!        anyNullFound = .true.
        rowFirstNull = row
    exit
      end if
    end do
    call release(compCol)

    if (rowFirstNull<=0) then
      call warning('noNonNullRows',"")
    else
      call readColDataToPtr(histoTab, 'SIM_FLUX_LO',    xVals)
      call readColDataToPtr(histoTab, 'COMP_RATIO_INT', yVals)
      ! In ssclib/dal_aux.

      allocate(packedXVals(rowFirstNull), packedYVals(rowFirstNull))
      packedXVals = xVals(:rowFirstNull-1)
      packedYVals = yVals(:rowFirstNull-1)

      write (pgplotDevice, '(a,a)') trim(plot5FileName), pgDev

      call plotInit(pgplotDevice, 'FLUX (cgs)', 'Fraction of reliable detections > FLUX.')
      call setXLog
      call drawLinePlot(packedXVals, packedYVals)
      ! All in ssclib/pgplot_aux.

      deallocate(yVals, xVals)
    end if

    call release(histoTab)
    call release(histoSet)

  end subroutine makePlot5

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module eimsimcompleteness_mod
