!+
! Module: eratetoflux_mod
!
! Authors:  Ian Stewart - JBO
!          
! Description: Converts the RATE values of an EPIC source list into fluxes. The output source list is in 'emldetect format', in that each source is represented by 1+n_inst*(n_bands+1) rows, but only those columns are retained which eimsim requires.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module eratetoflux_mod

  use dal
  use param  
  use errorHandling
  use caloaldefs	! for instId

  use temp_ssclib	! for packColumn

  use array_utils
  use dal_aux
  use ssc_misc

  implicit none 

  private 

  type srcRateMatrixType
    real(single) :: rate, rateErr, fluxDensity, fluxDensityErr
    logical :: isNull
  end type srcRateMatrixType
  type srcDataType
    integer :: id
    real(single) :: fluxDensity, fluxDensityErr, flux, fluxErr, like
    logical :: isNull
  end type srcDataType
  type instFilterType
    integer(int32) :: instId
    character(10) :: filter
  end type instFilterType

  ! Command-line parameters:
  !
  character(256) ::&
    srcSpecSetName='',&
    inSrcListSetName='',&
    outSrcListSetName=''
  character(20) ::&
!!!    rateColName='',&
!!!    rateErrColName='',&
!!!    fluxColName='',&
!!!    fluxErrColName='',&
    srcIdColName=''
  logical ::&
    withUserBandList=.false.
  character(1), allocatable ::&
    userBands(:)
  type(instFilterType), allocatable ::&
    userInstsFilters(:)

  public eratetoflux

contains 

  !---------------------------------------------------------------------
  subroutine eratetoflux
  !---------------------------------------------------------------------
  ! This subroutine reads in RATE values (in counts/sec) from a list of
  ! detected sources and, using the appropriate conversion factors,
  ! converts the RATE values to the same flux scale as the original list
  ! of simulated sources, in preparation for comparing the detected with
  ! the starting fluxes.
  !
  ! The problem with converting rate to flux is that very generally there
  ! are several rate values associated with each source, representing the
  ! rate measured by different instruments in different bands. These must
  ! somehow be converted to a single flux value per source. The present
  ! program accomplishes this for each source by applying the ECF conv-
  ! ersion factor appropriate to the instrument, filter and energy band
  ! to convert all valid rate values to flux densities at 1 keV; an
  ! optimally-weighted average of these flux densities is then calculated
  ! for each source; finally, the densities are converted to absolute
  ! flux values.
  !
  ! Note that we should NOT use any RATE values from rows for which either
  ! ID_INST or ID_BAND equals zero. These are 'summary' rows and their
  ! values of RATE are calculated from the non-summary rows by algorithms
  ! which may vary from script to script and which are neither predictable
  ! nor accessible to eimsim. Only the non-summary row values of RATE have
  ! an interpretation which may reasonably be expected to be invariant
  ! between scripts, namely the predicted nett count rate of the source
  ! for the specified band and instrument.
  !
  ! The actual algorithm is as follows:
  !
  !	(i) For each instrument, for each (non-summary) energy band,
  !     use the ECF value (read from the FLUX_SCALES table of the source-
  !     specification dataset) to convert the RATE value in the detected
  !     source list to flux S.
  !
  !	(ii) The FLUX_SCALES table also contains flux values in each band
  !     for a source which has a flux density of 1 erg cm^-2 s^-1 keV^-1
  !     at 1 keV. Use these to further convert the flux values (and
  !     associated uncertainties) into flux densities at 1 keV.
  !
  !	(iii) For each source, perform a weighted average of all the flux-
  !     density values. The weights used are the reciprocals of the
  !     variance (error^2) in the input values.
  !
  !	(iv) Perform the final conversion (using the flux value in the
  !     FLUX_SCALES table header) from flux density to flux.
  !
  ! A decision must be made as to which instruments and bands to select
  ! for the weighted sum. There are three sources of information which
  ! define the set of instruments and the set of bands:
  !	1: The FLUX_SCALES table of the specifation data set;
  !	2: The rows of the source list;
  !	3: (Optionally in the case of bands) user-supplied lists.
  ! For each source, a particular RATE value is only included in the
  ! calculation if both the instrument and band associated with that row
  ! of the source list are found, firstly, in the FLUX_SCALES table, and
  ! secondly, in the user-supplied lists. A third condition is that the
  ! filter specified by the user for that instrument is also found in the
  ! FLUX_SCALES table.
  !
  ! Note that the program also copies the summary-row DET_ML values to
  ! the output source list.
  !---------------------------------------------------------------------

    implicit none

    character(1), pointer :: allowedBands(:), srcListBands(:)
    integer(int32), pointer :: allowedInstIds(:), srcListInstIds(:)
    character(10), pointer :: allowedFilters(:), instSasNames(:)
    character(2), pointer :: instAbbreviations(:)
    type(srcDataType), pointer :: srcData(:)
    type(srcRateMatrixType), pointer :: srcRateMatrix(:,:,:)
    integer :: numSources, numInsts, numBands, i, instI, ii, bi
    logical, allocatable ::&
      instrumsPresent(:), bandsPresent(:), instBandIsPresent(:,:), isValid(:,:)
    character(2), allocatable :: srcListInstAbbrevs(:)
    character(10), allocatable :: srcListFilters(:)
    real(single), allocatable :: ecfs(:,:), fluxDensitiesToFluxes(:), weights(:,:)
    real(single) :: fluxDensityToFlux

logical :: testFlag = .false.

integer :: srcNum

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, 'aaa detRateToFluxRandom'

    call readParameters

    call readFluxScalesTable(allowedBands, allowedInstIds, allowedFilters&
      , instSasNames, instAbbreviations)

    call loadDetListStuff(instSasNames, srcData, srcRateMatrix, srcListBands, srcListInstIds)

    numSources = size(srcData)
    if (size(srcRateMatrix,1) .ne. numSources)&
call error('bug0',"")

if (testFlag) then
do srcNum = 1, numSources
  print *, 'src', srcNum, 'summary DET_ML=', srcData(srcNum)%like
end do
end if

    numInsts = size(srcListInstIds)
    if (size(srcRateMatrix,2) .ne. numInsts)&
call error('bug1',"")

    numBands = size(srcListBands)
    if (size(srcRateMatrix,3) .ne. numBands)&
call error('bug2',"")

    ! Now flag just those instruments which we are going to use the rates from:
    !
    allocate(instrumsPresent(numInsts), srcListFilters(numInsts))
    allocate(srcListInstAbbrevs(numInsts))
    instrumsPresent = .false. ! default
    srcListFilters = '' ! default
    do i = 1, numInsts
      if (.not.any(allowedInstIds==srcListInstIds(i)))&
    cycle

      if (.not.any(userInstsFilters%instId==srcListInstIds(i))) then
        call warning('instNotInAllowedList',"")
    cycle
      end if

      instI = smallestTrueIndex(userInstsFilters%instId==srcListInstIds(i))
      if (.not.any(allowedFilters==userInstsFilters(instI)%filter)) then
        call warning('filterNotInAllowedList',"")
   cycle
      end if

      instrumsPresent(i) = .true. ! phew, made it!
      srcListFilters(i) = userInstsFilters(instI)%filter

      instI = smallestTrueIndex(allowedInstIds==srcListInstIds(i))
      srcListInstAbbrevs(i) = instAbbreviations(instI)
    end do

    ! Do the same for the bands:
    !
    allocate(bandsPresent(numBands))
    bandsPresent = .false. ! default
    do i = 1, numBands
      if (.not.any(allowedBands==srcListBands(i)))&
    cycle

      if (withUserBandList) then
        if (.not.any(userBands==srcListBands(i))) then
          call warning('bandNotInAllowedList',"")
    cycle
        end if
      end if

      bandsPresent(i) = .true. ! phew, made it!
    end do

    ! Now fetch the ECFs and flux factors from the FLUX_SCALES table.
    ! We just want those for the source bands, insts and filters.
    !
    allocate(ecfs(numInsts, numBands), fluxDensitiesToFluxes(numBands))
    call readConversionFactors(srcListInstIds, srcListFilters, srcListInstAbbrevs&
      , instrumsPresent, srcListBands, bandsPresent, ecfs, fluxDensitiesToFluxes&
      , fluxDensityToFlux)

    ! Convert the RATE values (and uncertainties) to flux densities:
    !
    do ii = 1, numInsts
    if (.not.instrumsPresent(ii)) cycle
      do bi = 1, numBands
      if (.not.bandsPresent(bi)) cycle

        where(.not. srcRateMatrix(:, ii, bi)%isNull)
          srcRateMatrix(:, ii, bi)%fluxDensity&
          = srcRateMatrix(:, ii, bi)%rate * 1.0e-11 / ecfs(ii, bi)&
          / fluxDensitiesToFluxes(bi)
          srcRateMatrix(:, ii, bi)%fluxDensityErr&
          = srcRateMatrix(:, ii, bi)%rateErr * 1.0e-11 / ecfs(ii, bi)&
          / fluxDensitiesToFluxes(bi)
        end where
      end do
    end do

    ! Recast the inst/band present bool vectors into a matrix form, which will
    ! be found more convenient when calculating the weighted average:
    !
    allocate(instBandIsPresent(numInsts, numBands))
    instBandIsPresent = .false.
    do ii = 1, numInsts
    if (.not.instrumsPresent(ii)) cycle
      do bi = 1, numBands
      if (.not.bandsPresent(bi)) cycle
        instBandIsPresent(ii,bi) = .true.
      end do
    end do

    ! Calculate the optimum-weighted average (note that the default value for
    ! srcData(i)%isNull is TRUE):
    !
    allocate(weights(numInsts, numBands), isValid(numInsts, numBands))
    do i = 1, numSources
!    if (all(srcRateMatrix(i,:,:)%isNull .and. .not. instBandIsPresent)) cycle
      isValid = (.not. srcRateMatrix(i,:,:)%isNull .and. instBandIsPresent&
        .and. srcRateMatrix(i,:,:)%fluxDensityErr > 0.0)
!      if (all(srcRateMatrix(i,:,:)%isNull .and. .not. instBandIsPresent)) then
      if (.not.any(isValid)) then
        write(errStr, '(a,a,a1,i3.3,a)') 'All relevant RATE values for source with '&
          ,trim(srcIdColName),'=',srcData(i)%id,' are nulls, or have RATE_ERR <= 0. &
          &Source omitted from output.'
        call warning("allRateValuesNull")
    cycle
      end if

      srcData(i)%isNull = .false.
      weights = 0.0
!      where(.not. srcRateMatrix(i,:,:)%isNull .and. instBandIsPresent)&
      where(isValid)&
        weights = 1.0 / srcRateMatrix(i,:,:)%fluxDensityErr

      weights = weights * weights
if (sum(weights)<=0.0) call error('bug3',"")
!      srcData(i)%fluxDensity = sum(srcRateMatrix(i,:,:)%fluxDensity * weights)&
      srcData(i)%fluxDensity = sum(srcRateMatrix(i,:,:)%fluxDensity * weights, isValid)&
        / sum(weights)
      srcData(i)%fluxDensityErr = sqrt(1.0 / sum(weights))
    end do
!********* check these formulae for optimum weighted average.

    where(.not.srcData%isNull)
      srcData%flux    = srcData%fluxDensity    * fluxDensityToFlux
      srcData%fluxErr = srcData%fluxDensityErr * fluxDensityToFlux
    end where

    call saveFluxValues(srcData)

    deallocate(srcData, srcRateMatrix, srcListBands, srcListInstIds)
    deallocate(allowedBands, allowedInstIds, allowedFilters, instSasNames, instAbbreviations)

if (testFlag) print *, 'zzz detRateToFluxRandom'

  end subroutine eratetoflux

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    integer :: i, numFilters, numBandsInList!, instId, bandId
    character(10) :: instStr

    !-------------------------------------------------------------------
    ! Start

    srcSpecSetName = stringParameter('srcspecset')
    inSrcListSetName = stringParameter('insrclistset')
    outSrcListSetName = stringParameter('outsrclistset')
    srcIdColName = stringParameter('idcol')

    numFilters = parameterCount('filters')
    if (numFilters .ne. parameterCount('instrums'))&
call error('wrongNumberInstrums',"There must be 1 per filter.")

    allocate(userInstsFilters(numFilters))

    userInstsFilters%filter = '' ! default
    do i = 1, numFilters
      instStr = ucase(stringParameter('instrums',i-1))
      select case(instStr)
      case('PN', 'EPN')
        userInstsFilters(i)%instId = instId('EPN') ! in caloalutils/src/caloaldefs.f90
      case('M1', 'MOS1', 'EMOS1')
        userInstsFilters(i)%instId = instId('EMOS1') ! in caloalutils/src/caloaldefs.f90
      case('M2', 'MOS2', 'EMOS2')
        userInstsFilters(i)%instId = instId('EMOS2') ! in caloalutils/src/caloaldefs.f90
      case default
call error('badInstrum',"The value of the --instrums parameter was unrecognized.")
      end select

      if (any(userInstsFilters(1:i-1)%instId == userInstsFilters(i)%instId))&
call error('duplicatedInstrum',"Each instrument may appear only once in the --instrums list.")

      userInstsFilters(i)%filter = ucase(stringParameter('filters',i-1))
    end do

    withUserBandList = booleanParameter('withbands')
    if (withUserBandList) then
      numBandsInList = parameterCount('bands')

      allocate(userBands(numBandsInList))
      do i = 1, numBandsInList
        write(userBands(i), '(i1)') intParameter('bands',i-1)
!****** eventually make the parameter a string parameter.
        if (any(userBands(1:i-1)==userBands(i)))&
call error('duplicatedBand',"Each band may appear only once in the --bands list.")

      end do
    end if

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine readFluxScalesTable(allowedBands, allowedInstIds&
  , allowedFilters, instSasNames, instAbbreviations)
  !---------------------------------------------------------------------
  ! Note that the calling program should deallocate the pointer arguments after use.

    implicit none

    character(1),   pointer :: allowedBands(:)
    integer(int32), pointer :: allowedInstIds(:)
    character(10),  pointer :: allowedFilters(:), instSasNames(:)
    character(2),   pointer :: instAbbreviations(:)

    type(DataSetT) :: set
    type(TableT) :: tab
    integer :: numAllowedInsts, i, numAllowedFilters, numAllowedBands
integer(int16), allocatable :: allowedBandIntegers(:)

    !-------------------------------------------------------------------
    ! Start

    set = dataSet(srcSpecSetName, READ)
    tab = table(set, 'INSTRUMENTS')
    numAllowedInsts = numberOfRows(tab)
    allocate(allowedInstIds(numAllowedInsts), instSasNames(numAllowedInsts))
    allocate(instAbbreviations(numAllowedInsts))
    call readColDataToFixed(tab, 'NAME', instSasNames) ! this sub is in ssclib/dal_aux.
    call readColDataToFixed(tab, 'ABBREV', instAbbreviations)
    call release(tab)

    do i = 1, numAllowedInsts
      allowedInstIds(i) = instId(instSasNames(i)) ! in caloalutils/src/caloaldefs.f90
    end do

    tab = table(set, 'FILTERS')
    numAllowedFilters = numberOfRows(tab)
    allocate(allowedFilters(numAllowedFilters))
    call readColDataToFixed(tab, 'NAME', allowedFilters) ! this sub is in ssclib/dal_aux.
    call release(tab)

    tab = table(set, 'FLUX_SCALES')
    numAllowedBands = numberOfRows(tab)
    allocate(allowedBands(numAllowedBands))
!!!    call readColDataToFixed(tab, 'ID_BAND', allowedBands) ! this sub is in ssclib/dal_aux.
allocate(allowedBandIntegers(numAllowedBands))
call readColDataToFixed(tab, 'ID_BAND', allowedBandIntegers) ! this sub is in ssclib/dal_aux.
do i = 1, numAllowedbands
  write(allowedBands(i), '(i1)') allowedBandIntegers(i)
end do
    call release(tab)
    call release(set)

  end subroutine readFluxScalesTable

  !---------------------------------------------------------------------
  subroutine loadDetListStuff(instSasNames, srcData, srcRateMatrix&
  , srcListBands, srcListInstIds)
  !---------------------------------------------------------------------
  ! The function of this subroutine is to read the source list and return
  ! a list of values for each source. For each source, the routine should
  ! return the following quantities:
  !	- An ID integer, unique to that source, to be obtained from the
  !	  column designated by the srcIdColName string variable.
  !	- A boolean flag to indicate if a source corresponding to the ID
  !	  integer exists in the source list.
  !	- The summary-row value of DET_ML.
  !	- A 2d array of numbers (returned in srcRateMatrix(i,:,:) for
  !	  each source i) relating to various RATE values for that source.
  !	  The second index of srcRateMatrix indexes instrument; the third,
  !	  energy band. Numbers of instruments and bands must be determined
  !	  from the source list. Instrument/band combinations from all non-
  !	  summary rows are counted. Three quantities are associated with
  !	  each element of the srcRateMatrix(i,:,:) array:
  !		- RATE
  !		- RATE_ERR
  !		- A boolean 'isNull' to indicate whether the either of the
  !		  first two are nulls.
  !
  ! In the 'Potsdam-style' source lists for which the present software
  ! is designed, each source is associated with several rows in the table.
  ! Each row is associated with a particular combination of instrument and
  ! energy band. Note that the present subroutine assumes the following:
  !	(i) That each source is associated with the same set of
  !	    instrument/band rows. Actually the first source defines the set:
  !	    all subsequent sources must at least have rows corresponding to
  !	    that set; excess rows are ignored.
  !	(ii) That this single set of instrument/band combinations is such
  !	     that, if a given (non-summary) instrument has a given band,
  !	     then all the (non-summary) instruments also have that band.
  !	     Thus all the entries of the 2D array srcRateMatrix(i,:,:)
  !	     are associated with a row which exists, and no (non-summary)
  !	     rows are missed out for any source.
  !	(iii) That no source may have more than 1 row of a given inst/band
  !	      combination.
  !	(iv) That for each source, there is a single 'summary' row for
  !	     which both ID_INST and ID_BAND equal zero.
  !
  ! Note! The calling program should deallocate the pointer arguments after use.
  !---------------------------------------------------------------------

    implicit none

    character(10),           intent(in) :: instSasNames(:)
    type(srcDataType),       pointer    :: srcData(:)
    type(srcRateMatrixType), pointer    :: srcRateMatrix(:,:,:)
    character(1),            pointer    :: srcListBands(:)
    integer(int32),          pointer    :: srcListInstIds(:)

    type(DataSetT) :: set
    type(TableT) :: tab
    integer ::&
      numRows, minSrcNum, maxSrcNum, srcNum, numNonSummaryRows, numInsts,&
      numBands, ii, bi, idInst, idBand, numRowsThisCombo, row
    type(ColumnT) ::&
      rateCol, rateErrCol, srcIdCol, idInstCol, idBandCol, detMlCol
    real(single), pointer :: rateData(:), rateErrData(:), detMlData(:)
    integer(int32), pointer :: idBandData(:), idInstData(:), srcIdData(:), uniqueInsts(:), uniqueBands(:)
    logical, allocatable :: mask0(:), mask1(:)
    integer(int32), allocatable :: packedInsts(:), packedBands(:)

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '  aaa loadDetListStuff'

    set = dataSet(inSrcListSetName, READ)
    tab = table(set, 'SRCLIST')
    numRows = numberOfRows(tab)
if (numRows <= 0) call error('noSources','')

    srcIdCol   = column(tab, srcIdColName)
    rateCol    = column(tab, 'RATE')
    rateErrCol = column(tab, 'RATE_ERR')
    idInstCol  = column(tab, 'ID_INST')
    idBandCol  = column(tab, 'ID_BAND')
    detMlCol   = column(tab, 'DET_ML')

    srcIdData   => int32Data( srcIdCol)
    rateData    => real32Data(rateCol)
    rateErrData => real32Data(rateErrCol)
    idInstData  => int32Data( idInstCol)
    idBandData  => int32Data( idBandCol)
    detMlData   => real32Data(detMlCol)

    minSrcNum = minval(srcIdData)
    maxSrcNum = maxval(srcIdData)

    allocate(mask0(numRows), mask1(numRows))

    ! The first thing to do is to find out how many (non-summary) instruments
    ! and bands there are. We also at the same time check assumptions
    ! (ii) and (iii) described in the header to the present subroutine.
    !
    srcNum = minSrcNum
    mask0 = (srcIdData == srcNum)
    mask1 = (mask0 .and. .not.(idInstData==0 .or. idBandData==0))
    numNonSummaryRows = count(mask1)
    if (numNonSummaryRows < 1)&
call error('noNonSummaryRowsSrc1',"")

    allocate(packedInsts(numNonSummaryRows), packedBands(numNonSummaryRows))
    packedInsts = pack(idInstData, mask1)
    packedBands = pack(idBandData, mask1)
    uniqueInsts => findUniqueValues(packedInsts) ! in ssclib/array_utils.
    uniqueBands => findUniqueValues(packedBands)
    numInsts = size(uniqueInsts)
    numBands = size(uniqueBands)

    do ii = 1, numInsts
      idInst = uniqueInsts(ii)
      do bi = 1, numBands
        idBand = uniqueBands(bi)

        numRowsThisCombo = count(mask1 .and. idInstData==idInst .and. idBandData==idBand)
        if (numRowsThisCombo < 1) then
          write(errStr, '(a,i4.4,a,i1,a,i1,a)') 'Source ', srcNum, ' ID_INST '&
            , idInst, ' ID_BAND ', idBand, ': no rows found.'
call error('noRowForAllowedInstBandCombo')
        end if
        if (numRowsThisCombo > 1) then
          write(errStr, '(a,i4.4,a,i1,a,i1,a,i4.4,a)') 'Source ', srcNum, ' ID_INST '&
            , idInst, ' ID_BAND ', idBand, ': ',numRowsThisCombo ,' rows found. There should only be 1.'
call error('tooManyRowsThisInstBandCombo')
        end if
      end do
    end do

    ! Allocate the pointer arguments, and fill in some values:
    !
    allocate(srcData(maxSrcNum))
    call initializeSrcDataType(srcData)
    allocate(srcRateMatrix(maxSrcNum,numInsts,numBands))
    call initializeRateMatrixType(srcRateMatrix)
    allocate(srcListInstIds(numInsts), srcListBands(numBands))

    ! Some care is needed with the instrument IDs. There are three ways
    ! to identify an instrument: (i) by its 'official SAS' name, eg 'EPN';
    ! (ii) by its 'official SAS' instId, which is returned from the name
    ! by the caloalutils function instId(); (iii) by the 'Potsdam' integer
    ! code, as used for the ID_INST column values in all source lists
    ! output by eboxdetect and emldetect. Within the present subroutine,
    ! wherever a variable has an 'instId' component, the 2nd definition
    ! is meant; whereas an 'idInst' component indicates the Potsdam
    ! convention. At present we wish to convert from the Potsdam to 2nd
    ! definition. This is accomplished by means of the instSasNames()
    ! vector, which was read from the INSTRUMENTS table of the source
    ! specification dataset. The elements of this are numbered according
    ! to the Potsdam scheme.
    !
    do ii = 1, numInsts
      idInst = uniqueInsts(ii) ! Potsdam scheme
      if (idInst<1 .or. idInst>size(instSasNames))&
call error('badIdInst',"")

      srcListInstIds(ii) = instId(instSasNames(idInst)) ! in caloalutils/src/caloaldefs.f90
    end do

    do bi = 1, numBands
      write(srcListBands(bi), '(i1)') uniqueBands(bi)
    end do

    ! Now read the main data, checking assumptions (i) and (iv) along the way:
    !
    do srcNum = minSrcNum, maxSrcNum
      srcData(srcNum)%id = srcNum ! this is redundant at present
      ! but it is a good facility to keep nevertheless.

      mask0 = (srcIdData == srcNum)
    if (.not. any(mask0)) cycle

      mask1 = (mask0 .and. idInstData==0 .and. idBandData==0) !**** should probably store the integers used for summaries in the appropriate tables of the source spec set. Then rather have the following:
!!!      mask1 = (mask0 .and. idInstData==summaryInst .and. idBandData==summaryBand)
if (count(mask1) .ne. 1) call error('badNumSummaryRows',"")
      row = smallestTrueIndex(mask1) ! in ssclib/array_utils.
      srcData(srcNum)%like = detMlData(row-1)

      do ii = 1, numInsts
        idInst = uniqueInsts(ii)
        do bi = 1, numBands
          idBand = uniqueBands(bi)

          mask1 = (mask0 .and. idInstData==idInst .and. idBandData==idBand)
if (count(mask1) .ne. 1) call error('badNumberRowsThisInstBand',"")

          row = smallestTrueIndex(mask1) ! in ssclib/array_utils.
          srcRateMatrix(srcNum,ii,bi)%rate    = rateData(   row-1)
          srcRateMatrix(srcNum,ii,bi)%rateErr = rateErrData(row-1)

          if (isNull(rateCol, row-1))&
        cycle ! because the default state of srcRateMatrix(srcNum,ii,bi)%isNull should be TRUE

          srcRateMatrix(srcNum,ii,bi)%isNull = .false.
        end do
      end do
    end do

if (testFlag) then
    do srcNum = 1, maxSrcNum
      do idInst = 1, numInsts
        do idBand = 1, numBands
          print *, srcNum, idInst, idBand, srcRateMatrix(srcNum,idInst,idBand)%isNull
        end do
      end do
    end do
end if

    deallocate(uniqueInsts, uniqueBands)

    call release(detMlCol)
    call release(idBandCol)
    call release(idInstCol)
    call release(rateErrCol)
    call release(rateCol)
    call release(srcIdCol)
    call release(tab)
    call release(set)

if (testFlag) print *, '  zzz loadDetListStuff'
  end subroutine loadDetListStuff

  !---------------------------------------------------------------------
  subroutine readConversionFactors(srcListInstIds, srcListFilters&
  , srcListInstAbbrevs, instrumsPresent, srcListBands, bandsPresent&
  , ecfs, fluxDensitiesToFluxes, fluxDensityToFlux)
  !---------------------------------------------------------------------
  ! Note that the range of instruments, filters and bands in the
  ! FLUX_SCALES table has already been read and used to set the flags in
  ! instrumsPresent and bandsPresent. Thus there should be no danger that
  ! this subroutine will attempt to read an ECF for an inst/filt/band combo
  ! not in the table.

    implicit none

    integer(int32), intent(in)  :: srcListInstIds(:)
    character(10),  intent(in)  :: srcListFilters(size(srcListInstIds))
    character(2),   intent(in)  :: srcListInstAbbrevs(size(srcListInstIds))
    logical,        intent(in)  :: instrumsPresent(size(srcListInstIds))
    character(1),   intent(in)  :: srcListBands(:)
    logical,        intent(in)  :: bandsPresent(size(srcListBands))
    real(single),   intent(out) :: ecfs(size(srcListInstIds), size(srcListBands)),&
                                   fluxDensitiesToFluxes(size(srcListBands)),&
                                   fluxDensityToFlux

    integer :: numInsts, numBands, numRows, i, row, ii, bi
    type(DataSetT) :: set
    type(TableT) :: tab
    type(ColumnT) :: idBandCol, fluxCol, ecfCol
    integer(int16), pointer :: idBandData(:)
    real(single), pointer :: fluxData(:), ecfData(:)
    character(1), allocatable :: bandChars(:)
    logical, allocatable :: mask(:)
    character(1) :: band
    character(17) :: ecfColName

    !-------------------------------------------------------------------
    ! Start

    numInsts = size(srcListInstIds)
    numBands = size(srcListBands)

    set = dataSet(srcSpecSetName, READ)
    tab = table(set, 'FLUX_SCALES')
    numRows = numberOfRows(tab)

    fluxDensityToFlux = real32Attribute(tab, 'FLUX')

    idBandCol = column(tab, 'ID_BAND')
    fluxCol   = column(tab, 'FLUX')
    idBandData => int16Data(idBandCol)
    fluxData   => real32Data(fluxCol)

!*** for now: Convert int16 band ids to chars:
allocate(bandChars(numRows))
do i = 1, numRows
  write(bandChars(i), '(i1)') idBandData(i-1)
end do

    allocate(mask(numRows))

    ! Load flux densities:
    !
    fluxDensitiesToFluxes = 0.0 ! default
    do bi = 1, numBands
    if (.not.bandsPresent(bi)) cycle
      band = srcListBands(bi)
      mask = (bandChars==band)
      if (.not.any(mask))&
call error('bug99',"")

      row = smallestTrueIndex(mask) - 1 ! in ssclib/array_utils.
      fluxDensitiesToFluxes(bi) = fluxData(row)
    end do

    call release(fluxCol)

    ! Now the ECFs:
    !
    ecfs = 0.0 ! default
    do ii = 1, numInsts
    if (.not.instrumsPresent(ii)) cycle
      ! Construct the column name:
      !
      write(ecfColName, '(a4,a2,a1,a10)') 'ECF_', srcListInstAbbrevs(ii), '_', srcListFilters(ii)

      if (.not.hasColumn(tab, ecfColName))&
call error('bug98',"")

      ecfCol = column(tab, ecfColName)
      ecfData => real32Data(ecfCol)

      do bi = 1, numBands
      if (.not.bandsPresent(bi)) cycle

        band = srcListBands(bi)
        mask = (bandChars==band)
        row = smallestTrueIndex(mask) - 1 ! in ssclib/array_utils.
!*** set up an array of these values? Silly to do it again for each row.
        ecfs(ii,bi) = ecfData(row)
      end do
      call release(ecfCol)
    end do

    call release(tab)
    call release(set)

  end subroutine readConversionFactors

  !---------------------------------------------------------------------
  subroutine saveFluxValues(srcData)
  !---------------------------------------------------------------------

    implicit none

    type(srcDataType), intent(in) :: srcData(:)

    integer, parameter :: numOutCols = 4 !12

    type(DataSetT) :: inSet, outSet
    type(TableT) :: inTab, outTab
    integer ::&
      numInSources, numOutSources, maxSrcNum, srcNum, inRow, outRow, colNum
    type(ColumnT) ::&
      idInstCol, idBandCol, fluxCol, fluxErrCol, inCol, srcIdCol, nullProbCol
    real(single), pointer :: fluxData(:), fluxErrData(:), nullProbData(:)
! real(single), pointer :: fluxData(:), fluxErrData(:)
! real(double), pointer :: nullProbData(:)
    integer(int32), pointer ::&
      idBandData(:), idInstData(:), srcIdData(:)
    logical, allocatable :: tempMask(:), packMask(:)
    integer, allocatable :: unpackedIndices(:), indices(:)
    logical :: indicesAreSorted
    character(20) :: outColNames(numOutCols)

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

    outColNames( 1) = trim(srcIdColName)
    outColNames( 2) = 'RA'
    outColNames( 3) = 'DEC'
    outColNames( 4) = 'RADEC_ERR'
!    outColNames( 5) = 'STREAM_N'
!    outColNames( 6) = 'FIELD_N'
!    outColNames( 7) = 'DET_ML'
!    outColNames( 8) = 'LIKE'
!    outColNames( 9) = 'BOX_SIZE'
!    outColNames(10) = 'BKG_BOX_CTS'
!    outColNames(11) = 'SRC_BOX_CTS'
!    outColNames(12) = 'SRC_BOX+CTS_ERR'

    maxSrcNum = size(srcData)
    numOutSources = count(.not.srcData%isNull)

    outSet = dataSet(outSrcListSetName, CREATE)
    outTab = addTable(outSet, 'SRCLIST', numOutSources)
    call setAttribute(outTab, 'ID_COL', trim(srcIdColName), comment="Source ID col name in the input list.")

    ! First copy over all the other summary-row values of the columns 
    ! in the input list. We're going to do this by first making a mask,
    ! then using this to 'pack' each column.
    !
    inSet = dataSet(inSrcListSetName, READ)
    inTab = table(inSet, 'SRCLIST')
    numInSources = numberOfRows(inTab)
    srcIdCol   = column(inTab, srcIdColName)
    idInstCol  = column(inTab, 'ID_INST')
    idBandCol  = column(inTab, 'ID_BAND')
    srcIdData   => int32Data(  srcIdCol)
    idInstData  => int32Data( idInstCol)
    idBandData  => int32Data( idBandCol)

    allocate(tempMask(numInSources), packMask(numInSources))
    allocate(unpackedIndices(numInSources))
    packMask = .false.
    unpackedIndices = 0
    outRow = 0
    do srcNum = 1, maxSrcNum
    if (srcData(srcNum)%isNull) cycle
      outRow = outRow + 1

      tempMask = (srcIdData == srcNum)
      if (.not. any(tempMask))&
call error('bug0',"")

      if (any(tempMask .and. idInstData==0)) then
        inRow = smallestTrueIndex((tempMask .and. idInstData==0)) ! in ssclib/array_utils.
      else
        call warning('noInstSummaryRow',"")
        if (any(tempMask .and. idInstData==0)) then
          inRow = smallestTrueIndex((tempMask .and. idBandData==0)) ! in ssclib/array_utils.
        else
          call warning('noBandSummaryRow',"")
          inRow = smallestTrueIndex(tempMask) ! in ssclib/array_utils.
        end if
      end if
      packMask(inRow) = .true.
      unpackedIndices(inRow) = outRow
    end do

    call release(idBandCol)
    call release(idInstCol)
    call release(srcIdCol)

    if (count(packMask) .ne. numOutSources)&
call error('bug1',"")

    ! Since the SRC_NUM numbers don't necessarily occur in increasing order, set up an index array:
    !
    allocate(indices(numOutSources))
    indices = pack(unpackedIndices, packMask)
    if (all(indices(2:numOutSources) > indices(1:numOutSources-1))) then
      indicesAreSorted = .true.
    else
      indicesAreSorted = .false.
    end if

    ! Now copy the column values:
    !
    if (indicesAreSorted) then
      do colNum = 0, numberOfColumns(inTab)-1
        inCol = column(inTab, colNum)
      if (.not. any(name(inCol)==outColNames)) cycle ! in ssclib/ssc_misc.

        if (name(inCol) .eq. srcIdColName) then
          call packColumn(inCol, outTab, packMask, 'SRC_NUM') ! in ssclib/dal_aux.
        else
          call packColumn(inCol, outTab, packMask) ! in ssclib/dal_aux.
        end if

        call release(inCol)
      end do

    else
      do colNum = 0, numberOfColumns(inTab)-1
        inCol = column(inTab, colNum)
      if (.not. any(name(inCol)==outColNames)) cycle ! in ssclib/ssc_misc.

        if (name(inCol) .eq. srcIdColName) then
          call packColumn(inCol, outTab, packMask, 'SRC_NUM', indices) ! in ssclib/dal_aux.
        else
          call packColumn(inCol, outTab, packMask, indices=indices) ! in ssclib/dal_aux.
        end if

        call release(inCol)
      end do
    end if

    call release(inTab)
    call release(inSet)

    ! Add the new columns:
    !
    fluxCol     = addOrOpenColumn(outTab, 'FLUX',     REAL32)
    fluxErrCol  = addOrOpenColumn(outTab, 'FLUX_ERR', REAL32)
    nullProbCol = addOrOpenColumn(outTab, 'DETEC_PNULL', REAL32)
    fluxData     => real32Data(fluxCol)
    fluxErrData  => real32Data(fluxErrCol)
    nullProbData => real32Data(nullProbCol)

    outRow = 0
    do srcNum = 1, maxSrcNum
    if (srcData(srcNum)%isNull) cycle

!      if (srcData(srcNum)%isNull) then
!        print *, 'src', srcNum, 'is null.'
!    cycle
!      end if

!      print *, 'src', srcNum, 'summary DET_ML=', srcData(srcNum)%like

      outRow = outRow + 1
      fluxData(    outRow-1) = srcData(srcNum)%flux
      fluxErrData( outRow-1) = srcData(srcNum)%fluxErr
      nullProbData(outRow-1) = exp(-srcData(srcNum)%like)
!      print *, 'src', srcNum, 'summary DET_ML=', srcData(srcNum)%like, 'row', outRow, 'DETEC_PNULL=', nullProbData(outRow-1)
    end do

    call release(nullProbCol)
    call release(fluxErrCol)
    call release(fluxCol)

    call release(outTab)
    call release(outSet)

  end subroutine saveFluxValues

  !---------------------------------------------------------------------
  subroutine initializeRateMatrixType(arr)
  !---------------------------------------------------------------------

    implicit none

    type(srcRateMatrixType), intent(inout) :: arr(:,:,:)

    !-------------------------------------------------------------------
    ! Start

    arr%rate           = 0.0
    arr%rateErr        = 0.0
    arr%fluxDensity    = 0.0
    arr%fluxDensityErr = 0.0
    arr%isNull         = .true.

  end subroutine initializeRateMatrixType

  !---------------------------------------------------------------------
  subroutine initializeSrcDataType(srcData)
  !---------------------------------------------------------------------

    implicit none

    type(srcDataType), intent(inout) :: srcData(:)

    !-------------------------------------------------------------------
    ! Start

    srcData%id             = 0
    srcData%fluxDensity    = 0.0
    srcData%fluxDensityErr = 0.0
    srcData%flux           = 0.0
    srcData%fluxErr        = 0.0
    srcData%like           = 0.0
    srcData%isNull         = .true.

  end subroutine initializeSrcDataType



  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module eratetoflux_mod
