!+
! Module: faintbackcalc_mod
!
! Authors:  Ian Stewart - JBO
!          
! Description: Calculates the average background flux due to undetectable sources.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module faintbackcalc_mod

  use dal
  use param  
  use errorHandling

  use array_utils
  use dal_aux		! for isNullArray_ims
  use save_image
  use wcs_aux
!use temp_ssclib

  use power_law_utils	! for getLogNlogS, calcAccumFluxes, calcUndetectableFlux
!  use flux_transforms	! for fluxDensity0ToFlux

  implicit none 

  private 

  ! Command-line parameters:
  !
  character(20) ::&
    inputStyle=''
  character(256) ::&
    templateImageSetName='',&
    reciprocalSensyMapName='',&
    srcSpecSetName='',&
    faintSrcBkgMapName='',&
    tempImageName=''
  real(single) ::&
    srcFluxCutoff=0.0


  public faintbackcalc

contains 

  !---------------------------------------------------------------------
  subroutine faintbackcalc
  !---------------------------------------------------------------------
!******* hasn't been tested yet.

  ! This task takes as inputs a map of the estimated detection sensitivity S(x,y), plus the logN-logS template for the simulated sources, and calculates at each (x,y) the expected amount of total flux arising from undetected sources.
  !
  ! Note that the sensitivity map is actually required to be input in reciprocal form. This obviates the need to represent infinite values by nulls in areas of zero exposure. Testing of image nulls is very slow at the moment and thus should be avoided.

    implicit none

    type(DataSetT) :: srcSpecSet
    type(TableT) :: srcSpecTab
    real(single), pointer ::&
      modelFluxes(:), modelCumSkyDensity(:), reciprocalSensyMap(:,:)
    real(single) :: hiPower, pixelAreaDeg, faintSourceFluxDensity
    type(WcsT) :: wcs
    integer(int32) :: xiSize, yiSize, numModelKnees, xi, yi
    real(single), allocatable ::&
      sensyMap(:,:), modelPowers(:), offsets(:), multipliers(:),&
      faintSourceFluxMap(:,:), kVals(:)
    logical, allocatable :: nulls(:,:)

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, 'aaa faintbackcalc'

    call readParameters

    ! Read the logN-logS distribution model from the template file:
    !
    srcSpecSet = dataSet(srcSpecSetName, READ)
    srcSpecTab = table(srcSpecSet, 'SRCSPECS')
if (testFlag) print *, 'bbb faintbackcalc'
    call getLogNlogS(srcSpecTab, modelFluxes, modelCumSkyDensity&
      , hiPower) ! in power_law_utils
if (testFlag) print *, 'ccc faintbackcalc'
    call release(srcSpecTab)
    call release(srcSpecSet)

    numModelKnees = size(modelFluxes)
if (numModelKnees<1) call error('',"")
    allocate(modelPowers(numModelKnees), offsets(numModelKnees))
    allocate(multipliers(numModelKnees), kVals(numModelKnees))
if (testFlag) print *, 'ddd faintbackcalc'
    call calcLogNLogSPowers(modelFluxes, modelCumSkyDensity, hiPower&
      , modelPowers) ! in power_law_utils
if (testFlag) print *, modelPowers
if (testFlag) print *, 'eee faintbackcalc'
    kVals = -modelCumSkyDensity * (1.0 + modelPowers)&
      / (modelFluxes**(1.0+modelPowers))
if (testFlag) print *, kVals
if (testFlag) print *, 'fff faintbackcalc'
    call calcAccumFluxCoeffs(modelFluxes, kVals, modelPowers&
      , offsets, multipliers)
if (testFlag) print *, 'ggg faintbackcalc'

    select case(inputStyle)
    case('sensmap')
      ! Read or calculate the sensitivity map:
      !
!******** should read a keyword to confirm that the map contains reciprocal sensitivities.
      call readArrayData(reciprocalSensyMapName, reciprocalSensyMap)
      ! in ssclib/dal_aux.
      xiSize = size(reciprocalSensyMap, 1)
      yiSize = size(reciprocalSensyMap, 2)

      call readWcs(reciprocalSensyMapName, wcs) ! in ssclib/wcs_aux.

      allocate(sensyMap(xiSize, yiSize))
      sensyMap = 0.0
      where(reciprocalSensyMap > 0.0)&
        sensyMap = 1.0 / reciprocalSensyMap

      allocate(nulls(xiSize, yiSize))
      nulls = .true.
      where(reciprocalSensyMap <= 0.0 .and. sensyMap <= modelFluxes(1))&
        nulls = .false.
      deallocate(reciprocalSensyMap)

      allocate(faintSourceFluxMap(xiSize, yiSize))
      faintSourceFluxMap = 0.0
      do xi = 1, xiSize
        do yi = 1, yiSize
        if (nulls(xi, yi)) cycle

          faintSourceFluxMap(xi, yi) = calcCumFluxDensity(sensyMap(xi, yi), modelFluxes&
            , modelPowers, offsets, multipliers)
        end do
      end do

      deallocate(nulls, sensyMap)

      select case(wcs%y%worldUnit)
      case('deg')
        pixelAreaDeg = wcsPixelAreaWorld(wcs) ! in ssclb/wcs_aux.
      case default
call error('notSupported',"At present, WCS units (CUNITn keywords) must be 'deg'.")
      end select

      faintSourceFluxMap = faintSourceFluxMap * pixelAreaDeg

      call saveImage(faintSourceFluxMap, reciprocalSensyMapName, faintSrcBkgMapName&
        , tempImageName) ! in ssclib/save_image.

    case('single')
      call readWcs(templateImageSetName, wcs) ! in ssclib/wcs_aux.

      faintSourceFluxDensity = calcCumFluxDensity(srcFluxCutoff, modelFluxes&
        , modelPowers, offsets, multipliers)

      select case(wcs%y%worldUnit)
      case('deg')
        pixelAreaDeg = wcsPixelAreaWorld(wcs) ! in ssclb/wcs_aux.
      case default
call error('notSupported',"At present, WCS units (CUNITn keywords) must be 'deg'.")
      end select

      faintSourceFluxDensity = faintSourceFluxDensity * pixelAreaDeg

print *, 'Average flux per image pixel from faint sources is ', faintSourceFluxDensity

    case default
call error('',"")
    end select

    deallocate(modelFluxes, modelCumSkyDensity)
if (testFlag) print *, 'zzz faintbackcalc'

  end subroutine faintbackcalc

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    srcSpecSetName = stringParameter('srcspecset')
    inputStyle = stringParameter('inputstyle')

    select case(inputStyle)
    case('sensmap')
      reciprocalSensyMapName = stringParameter('invsensmapset')
      faintSrcBkgMapName = stringParameter('bkgfluxmapset')
      tempImageName = stringParameter('tempset')
    case('single')
      templateImageSetName = stringParameter('templateset')
      srcFluxCutoff = realParameter('fluxcutoff')
    case default
call error('',"")
    end select

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module faintbackcalc_mod
