!+
! Module: fluxlinearize_mod
!
! Authors:  Ian Stewart - JBO
!          
! Description: Transforms fluxes so that their expected values are evenly-distributed.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module fluxlinearize_mod

  use dal
  use param  
  use errorHandling

  use power_law_utils

  use dal_aux

  implicit none 

  private 

  ! Command-line parameters:
  character(256) ::&
    srcSpecSetName='',&
    srcListSetTabName=''
  character(20) ::&
    fluxColName='',&
    fluxErrColName='',&
    randColName='',&
    randErrColName=''

  public fluxlinearize

contains 

  !---------------------------------------------------------------------
  subroutine fluxlinearize
  !---------------------------------------------------------------------
  ! This task is designed to be part of a suite of tasks which generate simulated x-ray images, perform source detection on these images, then evaluate the results. At the start of this process one has a list of simulated sources, each characterized by a position and a flux; the detection procedure produces a list of detected sources, each with a measured value of position and flux plus associated uncertainties. Before any other analysis can be done it is necessary to find, for each detected source, that simulated source which is most likely to be the one detected. The detection likelihood can then be deduced, and completeness and sensitivity limits calculated.
  !
  ! The process of comparing and matching between simulated and detected sources is performed as follows. Imagine that the 3 numbers (position x, y and flux s) which specify each source are coordinates in an abstract 3-dimensional space. The uncertainties in position and flux thus define an ellipsoidal 'one-sigma' volume around each detected source. From detected source i, a metric r_i to the jth simulated source can be constructed as follows:
  !
  !	         (sim_x_j - det_x_i)^2   (sim_y_j - det_y_i)^2   (sim_s_j - det_s_i)^2
  !	r_ij^2 = --------------------- + --------------------- + ---------------------.
  !	             x_err_i^2               y_err_i^2               s_err_i^2
  !
  ! Thus r_ij is a measure of the 'number of sigmas' and will eg equal 1 in the case that the simulated source lies exactly on the ellipsoid. The simulated source j which is associated with the smallest value of r_ij is chosen as the matching source for detection i; the detection likelihood p_ij is also easily calculated as
  !
  !	p_ij = 1-exp(-F_ij)
  !
  ! where F_ij is the integrated number density of simulated sources inside the ellipsoid with semiaxes (r_ij*x_err_i, r_ij*y_err_i, r_ij*s_err_i).
  !
  ! The drawback with this is that sources are not distributed uniformly in all three directions: their density increases markedly with decreasing flux s. Some transform from s to s' is desirable, such that sources are uniformly or evenly distributed wrt s'. In fact the easiest choice for s' is the original, evenly-distributed random numbers which were subsequently transformed to flux values having the specified logN-logS distribution. These numbers have been retained for the simulated sources; it remains just to calculate them for the detections. That is the function of the present program.
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT) :: srcSpecSet, set
    type(TableT) :: srcSpecTab, tab
    real(single), pointer ::&
      modelFluxes(:),&
      modelCumSkyDensity(:)
    real(single) :: hiPower
    character(256) :: srcListSetName
    character(20) :: srcListTabName
    integer :: numSources, foreignStatus
    type(ColumnT) :: fluxCol, fluxErrCol, randCol, randErrCol
    real(single), pointer :: fluxData(:), fluxErrData(:), randData(:), randErrData(:)
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, 'aaa detRateToFluxRandom'

    call readParameters

    ! Read the logN-logS distribution model from the template file:
    !
    srcSpecSet = dataSet(srcSpecSetName, READ)
    srcSpecTab = table(srcSpecSet, 'SRCSPECS')
if (testFlag) print *, 'bbb'
    call getLogNlogS(srcSpecTab, modelFluxes, modelCumSkyDensity&
      , hiPower) ! in power_law_utils
if (testFlag) print *, 'ccc'

!    numModelPoints = size(modelFluxes)
    call release(srcSpecTab)
    call release(srcSpecSet)

    ! Now open the source table and read the columns of interest, plus
    ! the band edges and ecfs and other inst/band dependent factors. Note
    ! that the source list is expected to have only 1 row per source.
    ! Source lists output from the xmm tasks emldetect or srcmatch will
    ! need to be subject to some manipulation before they are suitable
    ! for input to the present task.
    !
    call splitSetTabName(srcListSetTabName, srcListSetName, srcListTabName)
    ! in ssclib/dal_aux.
    set = dataSet(srcListSetName, MODIFY)
    tab = table(set, srcListTabName)
    numSources = numberOfRows(tab)
if (numSources <= 0) call error('noSources','')

    fluxCol    = column(tab, fluxColName)
    fluxErrCol = column(tab, fluxErrColName)
    randCol    = addOrOpenColumn(tab, randColName, REAL32) ! in ssclib/dal_aux.
    randErrCol = addOrOpenColumn(tab, randErrColName, REAL32) ! in ssclib/dal_aux.
    fluxData    => real32Data(fluxCol)
    fluxErrData => real32Data(fluxErrCol)
    randData    => real32Data(randCol)
    randErrData => real32Data(randErrCol)

    randData    = 0.0
    randErrData = 0.0

    call invertLogNlogSrandoms(modelFluxes, modelCumSkyDensity&
      , fluxData, hiPower, randData, foreignStatus)

    if (foreignStatus < 0) then
print *, 'Status return from invertLogNlogSrandoms was', foreignStatus
call error('invertLogNlogSrandoms',"The function &
&invertLogNlogSrandoms returned an error status.")
    end if

    call ampErrToFluxRandErr(modelFluxes, modelCumSkyDensity&
      , fluxData, fluxErrData, hiPower, randErrData, foreignStatus)

    if (foreignStatus < 0) then
print *, 'Status return from ampErrToFluxRandErr was', foreignStatus
call error('ampErrToFluxRandErr',"The function &
&ampErrToFluxRandErr returned an error status.")
    end if

    call release(randErrCol)
    call release(randCol)
    call release(fluxErrCol)
    call release(fluxCol)
    call release(tab)
    call release(set)

    deallocate(modelFluxes, modelCumSkyDensity)

  end subroutine fluxlinearize

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    srcSpecSetName = stringParameter('srcspecset')
    srcListSetTabName = stringParameter('srclisttab')
    randColName = stringParameter('randcol')
    randErrColName = stringParameter('randerrcol')
    fluxColName = stringParameter('fluxcol')
    fluxErrColName = stringParameter('fluxerrcol')

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module fluxlinearize_mod
