!+
! Module: freqhistogen_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Calculates a probability density histogram and a cumulative probability histogram.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module freqhistogen_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.
  use dal
  use param  
  use errorHandling

  ! ssclib modules:
  use dal_aux		! for readColDataToFixed
  use histogram_utils
  use interpolations
  use ssc_misc		! for fracVal

  implicit none 

  ! Command-line parameters:
  character(256) ::&
    inDataSetTabName='',&
    inTheorySetTabName='',&
    outHistoSetTabName=''
  character(20) ::&
    inDataXColName='',&
    inTheoryXColName='',&
    inTheoryDiffProbColName='',&
    inTheoryIntProbColName='',&
    histoXMidColName='',&
    histoXIntColName='',&
    histoDataDiffProbColName='',&
    histoDataIntProbColName='',&
    histoTheoryDiffProbColName='',&
    histoTheoryIntProbColName='',&
    xRangeStyle='',&
    xBinningStyle=''
  logical ::&
    doReverseCumulative=.false.,&
    doLogXScale=.false.,&
    getHistoXColNamesFromUser=.false.,&
!    getHistoDataColNamesFromUser=.false.,&
!    getHistoTheoryColNamesFromUser=.false.,&
    lookForTheorySet=.false.,&
    newHistoSet=.false.
  integer ::&
    userNumHistoBins = 0
  real(single) ::&
    userXMin=0.0,&
    userXMax=0.0,&
    userHistoBinSize=0.0

  private 

  public freqhistogen

contains 

  !---------------------------------------------------------------------
  subroutine freqhistogen
  !---------------------------------------------------------------------

    implicit none

    character(256) :: inDataSetName, inTheorySetName, outHistoSetName
    character(20) :: inDataTabName, inTheoryTabName, outHistoTabName
    type(DataSetT) :: set
    type(TableT) :: tab
    integer ::&
      numPoints, i, numTheoryPoints, underflows, overflows, row, numHistoBins
    real(single), allocatable ::&
      inDataX(:), histoBinEdges(:), histoBinWidths(:), histoBinCentres(:),&
      inTheoryX(:), inTheoryDiffProb(:), inTheoryIntProb(:), theoryDiffProb(:),&
      theoryIntProb(:), dataDiffProb(:), dataIntProb(:)
    real(single) :: minXVal, maxXVal, histoXOvershoot
    logical, allocatable :: theoryDiffNulls(:), theoryIntNulls(:)
    integer, allocatable :: diffHisto(:), intHisto(:)
    type(ColumnT) :: diffProbCol, intProbCol
    real(single), pointer :: diffProbData(:), intProbData(:)
!    logical :: histoFileExists, overwriteXData

logical :: testFlag=.false.

    !-------------------------------------------------------------------
    ! Start:

if (testFlag) print *, 'aaa'

    call readParameters

    if (.not. getHistoXColNamesFromUser) then
      histoXMidColName = trim(inDataXColName)//'_MID'
      histoXIntColName = trim(inDataXColName)//'_INT'
    end if

    call splitSetTabName(outHistoSetTabName, outHistoSetName, outHistoTabName)
    ! in ssclib/dal_aux.

    ! Read the data values:
    !
    call splitSetTabName(inDataSetTabName, inDataSetName, inDataTabName) ! in ssclib/dal_aux.
    set = dataSet(inDataSetName, READ)
    tab = table(set, inDataTabName)
    numPoints = numberOfRows(tab)
    allocate(inDataX(numPoints))
    call readColDataToFixed(tab, inDataXColName, inDataX) ! in ssclib/dal_aux.
    call release(tab)
    call release(set)

    ! Calculate the histogram bin edges:
    !
    if (newHistoSet) then
      select case(xRangeStyle)
      case ('user')
        minXVal = userXMin
        maxXVal = userXMax
      case ('auto')
        minXVal = minval(inDataX)
        maxXVal = maxval(inDataX)
      case default
call error('badXRangeStyle',"")
      end select

      if (minXVal >= maxXVal)&
call error('noDataSpread',"")

      if (doLogXScale) then
        if (minXVal <= 0.0)&
call error('negativeMinimum',"Can't do log X scale")

        minXVal = log10(minXVal)
        maxXVal = log10(maxXVal)
      end if

      select case(xBinningStyle)
      case ('nbins')
        numHistoBins = userNumHistoBins
      case ('binsize') ! Note that in this case the x range may be increased.
        numHistoBins = ceiling((maxXVal - minXVal) / userHistoBinSize)
        histoXOvershoot = userHistoBinSize * numHistoBins - (maxXVal - minXVal)
        maxXVal = maxXVal + histoXOvershoot / 2.0
        minXVal = minXVal - histoXOvershoot / 2.0
      case default
call error('badXBinningStyle',"")
      end select

      allocate(histoBinEdges(numHistoBins+1), histoBinWidths(numHistoBins))
      allocate(histoBinCentres(numHistoBins))

      histoBinEdges = fracVal(minXVal, maxXVal, (/(i/real(numHistoBins), i=0,numHistoBins)/))
      ! in ssclib/ssc_misc.

      histoBinCentres = (histoBinEdges(2:numHistoBins+1) + histoBinEdges(1:numHistoBins)) / 2.0

      if (doLogXScale) then
!        minXVal = minval(inDataX)
!        maxXVal = maxval(inDataX)
        histoBinEdges   = 10.0**histoBinEdges
        histoBinCentres = 10.0**histoBinCentres
      end if

    else ! get bin widths etc from an existing histogram:
      set = dataSet(outHistoSetName, READ)
      tab = table(set, outHistoTabName)
      numHistoBins = numberOfRows(tab)

      allocate(histoBinEdges(numHistoBins+1), histoBinWidths(numHistoBins))
      allocate(histoBinCentres(numHistoBins))
      call readColDataToFixed(tab, histoXMidColName, histoBinCentres)
      doReverseCumulative = booleanAttribute(tab, 'REVERSE')
      if (doReverseCumulative) then
        call readColDataToFixed(tab, histoXIntColName, histoBinEdges(1:numHistoBins))
        histoBinEdges(numHistoBins+1) = real32Attribute(tab, 'LASTEDGE')
      else
        histoBinEdges(             1) = real32Attribute(tab, 'FRSTEDGE')
        call readColDataToFixed(tab, histoXIntColName, histoBinEdges(2:numHistoBins+1))
      end if

      call release(tab)
      call release(set)
    end if

    histoBinWidths = histoBinEdges(2:numHistoBins+1) - histoBinEdges(1:numHistoBins)

    if (lookForTheorySet) then
      ! Interpolate the theory values:
      !
      call splitSetTabName(inTheorySetTabName, inTheorySetName, inTheoryTabName)
      ! in ssclib/dal_aux.
      set = dataSet(inTheorySetName, READ)
      tab = table(set, inTheoryTabName)
      numTheoryPoints = numberOfRows(tab)
      allocate(inTheoryX(numTheoryPoints), inTheoryDiffProb(numTheoryPoints))
      allocate(inTheoryIntProb(numTheoryPoints))
      call readColDataToFixed(tab, inTheoryXColName,        inTheoryX)
      call readColDataToFixed(tab, inTheoryDiffProbColName, inTheoryDiffProb)
      call readColDataToFixed(tab, inTheoryIntProbColName,  inTheoryIntProb)
      ! readColDataToFixed is in ssclib/dal_aux.
      call release(tab)
      call release(set)

      allocate(theoryDiffProb(numHistoBins), theoryDiffNulls(numHistoBins))
      call interpolate(inTheoryDiffProb, theoryDiffProb, histoBinCentres&
        , inTheoryX, INTERP_LINEAR, theoryDiffNulls)
      allocate(theoryIntProb(numHistoBins), theoryIntNulls(numHistoBins))
      if (doReverseCumulative) then
        call interpolate(inTheoryIntProb, theoryIntProb, histoBinEdges(1:numHistoBins)&
          , inTheoryX, INTERP_LINEAR, theoryIntNulls)
      else
        call interpolate(inTheoryIntProb, theoryIntProb, histoBinEdges(2:numHistoBins+1)&
          , inTheoryX, INTERP_LINEAR, theoryIntNulls)
      end if
    end if ! lookForTheorySet

    ! Make the histogram:
    !
    allocate(diffHisto(numHistoBins), intHisto(numHistoBins))
    call makeHistogram(inDataX, diffHisto, histoBinEdges, underflows, overflows)
    ! in ssclib/histogram_utils.

    allocate(dataDiffProb(numHistoBins))
    dataDiffProb = diffHisto / real(numPoints) / histoBinWidths

    ! Make the cumulative histogram:
    !
    allocate(dataIntProb(numHistoBins))
    if (doReverseCumulative) then
      intHisto(numHistoBins) = overflows + diffHisto(numHistoBins)
      do i = numHistoBins-1, 1, -1
        intHisto(i) = intHisto(i+1) + diffHisto(i)
      end do
    else
      intHisto(1) = underflows + diffHisto(1)
      do i = 2, numHistoBins
        intHisto(i) = intHisto(i-1) + diffHisto(i)
      end do
    end if
    dataIntProb = intHisto / real(numPoints)

    ! Save the histograms:
    !
    if (newHistoSet) then
      set = dataSet(outHistoSetName, CREATE)
      tab = addTable(set, outHistoTabName, numHistoBins)
      call setAttribute(tab, 'REVERSE', doReverseCumulative)

      call writeColData(tab, histoXMidColName, histoBinCentres)
      if (doReverseCumulative) then
        call writeColData(tab, histoXIntColName, histoBinEdges(1:numHistoBins)) ! lower edges
        call setAttribute(tab, 'LASTEDGE', histoBinEdges(numHistoBins+1))
      else
        call setAttribute(tab, 'FRSTEDGE', histoBinEdges(1))
        call writeColData(tab, histoXIntColName, histoBinEdges(2:numHistoBins+1)) ! upper edges
      end if

    else
      set = dataSet(outHistoSetName, MODIFY)
      tab = table(set, outHistoTabName)
    end if

    call writeColData(tab, histoDataDiffProbColName, dataDiffProb)
    call writeColData(tab, histoDataIntProbColName,  dataIntProb)
    ! writeColData is in ssclib/dal_aux.

    if (lookForTheorySet) then
      diffProbCol = addOrOpenColumn(tab, histoTheoryDiffProbColName, REAL32)
      ! addOrOpenColumn is in ssclib/dal_aux.
      diffProbData => real32Data(diffProbCol)
      diffProbData = theoryDiffProb
      do row = 0, numHistoBins-1
        if (theoryDiffNulls(row+1)) call setToNull(diffProbCol, row)
      end do
      call release(diffProbCol)

      intProbCol = addOrOpenColumn(tab, histoTheoryIntProbColName, REAL32)
      ! addOrOpenColumn is in ssclib/dal_aux.
      intProbData => real32Data(intProbCol)
      intProbData = theoryIntProb
      do row = 0, numHistoBins-1
        if (theoryIntNulls(row+1)) call setToNull(intProbCol, row)
      end do
      call release(intProbCol)
    end if

    call release(tab)
    call release(set)

if (testFlag) print *, 'zzz'

  end subroutine freqhistogen

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start:

    inDataSetTabName           = stringParameter('indatatab')
    inDataXColName             = stringParameter('dataxcol')
    outHistoSetTabName         = stringParameter('histotab')

    getHistoXColNamesFromUser  = booleanParameter('withhistoxcols')
    if (getHistoXColNamesFromUser) then
    histoXMidColName           = stringParameter('histoxmidcol')
    histoXIntColName           = stringParameter('histoxintcol')
    end if
!    getHistoDataColNamesFromUser  = booleanParameter('withhistodatacols')
!    if (getHistoDataColNamesFromUser) then
      histoDataDiffProbColName   = stringParameter('histodatadiffcol')
      histoDataIntProbColName    = stringParameter('histodataintcol')
!    end if
    lookForTheorySet  = booleanParameter('withtheory')
    if (lookForTheorySet) then
      inTheorySetTabName         = stringParameter('intheorytab')
      inTheoryXColName           = stringParameter('theoryxcol')
      inTheoryDiffProbColName    = stringParameter('theorydiffcol')
      inTheoryIntProbColName     = stringParameter('theoryintcol')
!      getHistoTheoryColNamesFromUser  = booleanParameter('withhistotheorycols')
!      if (getHistoTheoryColNamesFromUser) then
        histoTheoryDiffProbColName = stringParameter('histotheorydiffcol')
        histoTheoryIntProbColName  = stringParameter('histotheoryintcol')
!      end if
    end if

    newHistoSet  = booleanParameter('newhistoset')
    if (newHistoSet) then
      doReverseCumulative = booleanParameter('doreverse')
      doLogXScale         = booleanParameter('dologx')

      xRangeStyle = stringParameter('xrangestyle') ! user, auto, legal or actual
      select case(xRangeStyle)
      case ('user')
        userXMin = realParameter('xmin')
        userXMax = realParameter('xmax')
      case ('auto') !!!, 'legal', 'actual')
        ! do nothing
      case default
call error('badXRangeStyle',"")
      end select

      xBinningStyle = stringParameter('xbinningstyle') ! npixels or pixelsize
      select case(xBinningStyle)
      case ('nbins')
        userNumHistoBins = intParameter('nbins')
      case ('binsize')
        userHistoBinSize = realParameter('binsize')
      case default
call error('badXBinningStyle',"")
      end select
    end if

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module freqhistogen_mod
