!+
! Module: mosaicprep_mod
!
! Authors:  Ian Stewart - JBO
!          
! Description: Contains several tasks to help make simulated x-ray images.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module mosaicprep_mod

  use dal
  use param  
  use errorHandling

  use constants
  use coordinate
  use dal_aux
  use wcs_aux

use temp_ssclib	! for the cone subroutines.

  implicit none 

  private 

  ! Command-line parameters:
  character(256), allocatable ::&
    expMapNames(:)
  character(256) ::&
    sourceSpecSet=''
  real(single) ::&
    padSizeArcsec=0.0

  public mosaicprep

contains 

  !---------------------------------------------------------------------
  subroutine mosaicprep
  !---------------------------------------------------------------------
  ! This subroutine takes a list of exposure maps and calculates the
  ! direction and angle of the narrowest conical locus (or, circle on
  ! the celestial sphere) which includes all non-zero parts of each
  ! exposure map. The padSizeArcsec is then added to this angle and the
  ! resulting pair of values (ie the direction and the angle) are written
  ! to the specified template file. The intention is that the routine which
  ! constructs the list of sim sources should read the cone details from
  ! the template set (which should also contain the logN-logS details for
  ! the source brightness distribution) and restrict the spatial
  ! distribution of sources to within the cone.
  !
  ! In practice (at least at first writing) I think I will attempt a
  ! rather simpler problem. I'm going to represent the non-zero exposure
  ! of a single exposure map by a cone of 15 arcmin (900 arcsec) radius,
  ! centred on the optic axis. Such a cone ought to include all the non-
  ! zero exposure, and this will make construction of the 'union' cone
  ! for all maps much easier.
  !
  ! The intent of this procedure is to allow simulations with overlapping
  ! fields. There isn't much point in doing this for a list of exposure
  ! maps if they don't substantially overlap. However at present there is
  ! no protection against calling makeLocationTemplate with a list of
  ! widely-scattered exposure maps.

    implicit none

    integer(int8), parameter :: X=1, Y=2, Z=3
    integer :: numFiles, fileNum
    type(WcsT) :: wcs
    real(double) ::&
      centralRay(X:Z), coneAngle, centreRa, centreDec, coneAngleArcSec,&
      coneAreaDeg
    real(double), allocatable :: rays(:,:), coneAngles(:)
    type(DataSetT) :: set
    type(TableT) :: tab

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    numFiles = size(expMapNames)
    if (numFiles <= 0)&
call error('noInfiles','eimsim_utils: no exposure maps were supplied.')

    allocate(rays(numFiles, X:Z), coneAngles(numFiles))

    do fileNum = 1, numFiles
      call readWcs(expMapNames(fileNum), wcs) ! in ssclib/wcs_aux
      rays(fileNum, X:Z) = skyToCartesian(wcs%x%refWorld * PiOn180&
        , wcs%y%refWorld * PiOn180) ! in ssclib/coordinate
      coneAngles(fileNum) = calcApproxConeAngle(expMapNames(fileNum), wcs)
    end do

    select case(numFiles)
    case(1)
      centralRay = rays(1,X:Z)
      coneAngle = 0.0d0

    case(2)
      call calcSmallestConePair(rays(1,X:Z), rays(2,X:Z), centralRay, coneAngle)
      ! at present this is in temp_ssclib.

    case(3)
      call calcSmallestConeTriplet(rays, centralRay, coneAngle)
      ! at present this is in temp_ssclib.

    case default
      call calcSmallestCone(rays, centralRay, coneAngle)
      ! at present this is in temp_ssclib.

!call test_doConePlot(rays, centralRay, coneAngle)
    end select

!    coneAngleArcSec = (coneAngle * 3600.0d0 / PiOn180) + 900.0 + padSizeArcsec
!    ! The 900 is the approximate radius in arcsec of EPIC fields.
!*** Uh-uh. I want this routine to be more general now, ie not tied to xmm specs.
! The radius of the largest map is added to the cone radius which just encloses all the central rays. It is rough and ready - not an elegant, minimal solution - but it should ensure that _all_ pixels of _all_ exposure maps fall within the resulting cone.
coneAngleArcSec = ((coneAngle + maxval(coneAngles))&
 * 3600.0d0 / PiOn180) + padSizeArcsec

    coneAngle = coneAngleArcSec / 3600.0 * PiOn180
    coneAreaDeg = (TwoPi * (1.0d0 - cos(coneAngle))) / PiOn180 / PiOn180

!print *, 'coneAngleArcSec, coneAngle, coneArea, coneAreaDeg:'&
!, coneAngleArcSec, coneAngle, TwoPi * (1.0d0 - cos(coneAngle)), coneAreaDeg

    call cartesianToSky(centralRay, centreRa, centreDec) ! in ssclib/coordinate
    set = dataSet(sourceSpecSet, MODIFY)
    tab = table(set, 'SRCSPECS')
    call setAttribute(tab, 'CONE_RA',  centreRa  / PiOn180, 'deg')
    call setAttribute(tab, 'CONE_DEC', centreDec / PiOn180, 'deg')
    call setAttribute(tab, 'CONE_RAD', coneAngleArcSec, 'arcsec')
    call setAttribute(tab, 'SKY_AREA', coneAreaDeg, 'deg^2')

    call release(set)

  end subroutine mosaicprep

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    integer :: numMaps, i

    !-------------------------------------------------------------------
    ! Start

    numMaps = parameterCount('expmapsets')
    allocate(expMapNames(numMaps))
    do i = 1, numMaps
      expMapNames(i) = stringParameter('expmapsets',i-1)
    end do

    sourceSpecSet = stringParameter('srcspecset')
    padSizeArcsec = realParameter('padsize')

  end subroutine readParameters

  !---------------------------------------------------------------------
  function calcApproxConeAngle(expMapName, wcs) result(approxConeAngle)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: expMapName
    type(WcsT),   intent(in) :: wcs
    real(double)             :: approxConeAngle

    real(double) :: xPixelSizeWorld, yPixelSizeWorld, rotAngle
    real(single), pointer :: expMap(:,:)
    integer :: xiSize, yiSize, xi, yi, locOfMax(2)
    real(double), allocatable :: x2(:), y2(:), rad2(:,:)

    !-------------------------------------------------------------------
    ! Start

    call convertNewToOld(wcs, xPixelSizeWorld, yPixelSizeWorld, rotAngle)
    ! in ssclib/wcs_aux.

    call readArrayData(expMapName, expMap) ! in ssclib/dal_aux.
if (.not.any(expMap>0.0)) call error('noTrueExpMapPixels',"")

    xiSize = size(expMap, 1)
    yiSize = size(expMap, 2)
    allocate(x2(xiSize), y2(yiSize), rad2(xiSize, yiSize))
    x2 = (/(xi-wcs%x%refPixels, xi=1,xiSize)/)
    x2 = x2 * xPixelSizeWorld
    x2 = x2 * x2
    y2 = (/(yi-wcs%y%refPixels, yi=1,yiSize)/)
    y2 = y2 * yPixelSizeWorld
    y2 = y2 * y2
    rad2 = spread(x2, 2, yiSize) + spread(y2, 1, xiSize)
    locOfMax = maxloc(rad2, (expMap>0.0))
    approxConeAngle = sqrt(rad2(locOfMax(1),locOfMax(2))) * PiOn180

    deallocate(expMap)

  end function calcApproxConeAngle

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module mosaicprep_mod
