!+
! Module: eimsim_mod
!
! Authors:  Ian Stewart - JBO
!          
! Description: Contains several tasks to help make simulated x-ray images.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module eimsim_mod

  use dal
  use cal
  use param  
  use errorHandling
  use dss	! for RangeT

  use cal_aux
  use constants
  use coordinate
  use dal_aux
use dump
  use geometric_types
  use regridders
  use save_image
  use ssc_misc
  use wcs_aux

  use eimsim_utils

  implicit none 

  private 

  ! Command-line parameters:
  character(256) ::&
    expMapSetName='',&
    sourceListSetName='',&
    paddedMaskSetName='',&
    rateImageSetName='',&
    tempImageName=''
  character(10) ::&
    obsId=''
  real(double) ::&
    ecf=0.0d0,&
    normalizedFlux=0.0d0,&
    psfEnergyFraction=0.0d0

!*******
type(RangeT) :: rateEeVRanges(1)

  integer(int32) ::&
    instrumentId
  type(SpacecraftAttitudeType) ::&
    scAttitude, instrumAttitude
  real(double) ::&
    expStartTimeStamp


  public eimsim

contains 

  !---------------------------------------------------------------------
  subroutine eimsim
  !---------------------------------------------------------------------
  ! This makes a simulated sky map from the source list, convolving each
  ! source position with the PSF appropriate to that position in the focal
  ! plane. The resulting sky map values are in counts/pixel/sec.
  !
  ! The sky map is created in the projection plane implicit in the wcs
  ! keywords of the exposure map. Sky map pixels are left set to zero
  ! where the exposure map is zero.
  !
  ! The exposure map also carries information about the instrument.
  !
  ! The medium-accuracy PSF from the cal is used for bright sources (ie
  ! those which have flux densities in erg cm^-2 s^-1 eV^-1 greater than
  ! psfCutoffFluxDensity) and a simpler, quicker PSF model is used for
  ! all fainter sources. The cal PSF requires instrument type and
  ! instrument pointing (obtained from the exposure map) as well as
  ! energy and energy fraction (supplied by the arguments psfEeV and
  ! psfEnergyFraction).
  !
  ! The flux densities p (in erg cm^-2 s^-1 eV^-1) for each source (used
  ! for comparison with psfCutoffFluxDensity) are calculated at E=psfEeV
  ! using the equation
  !
  !	p(E) = p(1) * E**gamma
!*** wrong, look at fluxTransforms stuff
  !
  ! where p(1) is the flux density at 1 keV, obtained from the source table.
  !.....................................................................

    implicit none

    type(DataSetT) :: sourceSet, imageSet
    type(TableT) :: sourceTab
!, templateTab
    type(ArrayT) :: imageArray
    integer(int32) :: numSources, src
    type(ColumnT) ::&
      raDegCol, decDegCol, fluxDensity0Col, flagCol
    real(single), pointer :: fluxDensity0(:), patch(:,:)
    real(double), pointer :: raDeg(:), decDeg(:)
    integer(int16), pointer :: flag(:)
    integer(int16) :: xSize, ySize, xi, yi
    real(single), allocatable :: image(:,:), rate(:)
    type(WcsT) :: wcs
    real(double) ::&
      numerator, denominator, psfEeV, sigmaTan,&
!photonSpectralIndex,
      lo, hi, xPixelSizeTan
    real(double) :: xPixelSizeWorld, yPixelSizeWorld, rotAngle
    real(double), allocatable ::&
      srcTanX(:), srcTanY(:), srcThetaArcsec(:), srcPhi(:)
    real(single), pointer :: srcXpixels(:), srcYpixels(:)
    real(single) :: sigmaPixels
!, convertToFlux, sigmaXpixels, sigmaYpixels
    real(single), parameter :: sigmaArcsec = 5.0 ! for coarse (gaussian) PSF.
    integer :: numRanges, rangeNum
    real(single), pointer :: expMapImage(:,:)
    character(8) :: wcsType
    logical(bool), pointer :: paddedMask(:,:)
    character(3) :: iStr
    logical :: thetaGtFovRadius
    integer(int16) ::&
      xStart, xFinis, yStart, yFinis,&
      gaussianPatchHalfWidthXi, gaussianPatchHalfWidthYi,&
      gaussianPatchWidthXi, gaussianPatchWidthYi
    real(single), pointer :: gaussianPatch(:,:)
    type(AttributeT) :: fluxCutoffAttr
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

    call readParameters

if (testFlag) print *, 'aaa eimsim'

    !...................................................................
    ! Load the information from the exposure map:

    call readArrayData(expMapSetName, expMapImage) ! in ssclib/dal_aux.

    xSize = size(expMapImage, 1)
    ySize = size(expMapImage, 2)

    call readArrayData(paddedMaskSetName, paddedMask) ! in ssclib/dal_aux.
    if (size(paddedMask,1) .ne. xSize .or. size(paddedMask,2) .ne. ySize)&
call error('','')

    call readWcs(expMapSetName, wcs) ! in ssclib/wcs_aux.
    call easifyWcsType(wcs%x%ctype, wcs%y%ctype, wcsType) ! in ssclib/wcs_aux.

    call getCalInfo(expMapSetName, instrumentId, expStartTimeStamp&
      , scAttitude, instrumAttitude) ! sets instrument in cal. In ssclib/cal_aux.

if (testFlag) print *, 'bbb'

    !...................................................................
    ! Set up rest of cal (inst already set in call to getCalInfo())
    ! and calculate some miscellaneous quantities:

    call CAL_setState(accuracyLevel=ACCURACY_MEDIUM)

    sigmaTan = sigmaArcsec * PiOn180 / 3600.0d0
!print *, 'sigmaTan=', sigmaTan

!    call tanToPixels(wcs, sigmaTan, sigmaTan, sigmaXpixels, sigmaYpixels)
!    sigmaXpixels = abs(sigmaXpixels)
!    sigmaYpixels = abs(sigmaYpixels)

call convertNewToOld(wcs, xPixelSizeWorld, yPixelSizeWorld, rotAngle)
! in ssclib/wcs_aux.
xPixelSizeTan = abs(xPixelSizeWorld) * PiOn180
sigmaPixels = sigmaTan / xPixelSizeTan

!print *, 'sigmaPixels=', sigmaPixels

    ! Calculate the average PSF by finding the mean of the energy ranges:

    if (any(rateEeVRanges%lower%type == UNDEFINED&
    .or.    rateEeVRanges%upper%type == UNDEFINED))&
call error('',"")

    if (any(rateEeVRanges%lower%value > rateEeVRanges%upper%value))&
call error('',"")

    numerator   = 0.0
    denominator = 0.0
    do rangeNum = 1, size(rateEeVRanges)
      lo = rateEeVRanges(rangeNum)%lower%value
      hi = rateEeVRanges(rangeNum)%upper%value
      numerator   = numerator   + hi * hi - lo * lo
      denominator = denominator + hi      - lo
    end do
if (denominator <= 0.0) call error('bug0',"Please contact the code developer.")
    psfEeV = numerator / denominator / 2.0
if (testFlag) print *, 'ccc'

    !...................................................................
    ! Open source set and load the columns:

    sourceSet = dataSet(sourceListSetName, MODIFY) ! 'modify' because need
    ! to set flag col values.

    sourceTab = table(sourceSet, 'SRCLIST')
    numSources = numberOfRows(sourceTab)
if (numSources <= 0) call error('noSources','')
    fluxCutoffAttr = attribute(sourceTab, 'SDY0_CUT')

    raDegCol        = column(sourceTab, 'RA')
    decDegCol       = column(sourceTab, 'DEC')
    fluxDensity0Col = column(sourceTab, 'S_DENS_0')
    flagCol         = column(sourceTab, 'FLAG_'//obsId)

    raDeg        => real64Data(raDegCol)
    decDeg       => real64Data(decDegCol)
    fluxDensity0 => real32Data(fluxDensity0Col) ! flux density at 1 keV.
    flag         => int16Data(flagCol)
if (testFlag) print *, 'ddd'

    !...................................................................
    ! Initialize flag column:

    select case(instrumentId)
    case(EPN)
      flag = ibclr(flag, NOT_IN_PN_EXP_BIT)
      flag = ibclr(flag, NOT_IN_PN_FOV_BIT)
    case(EMOS1)
      flag = ibclr(flag, NOT_IN_M1_EXP_BIT)
      flag = ibclr(flag, NOT_IN_M1_FOV_BIT)
    case(EMOS2)
      flag = ibclr(flag, NOT_IN_M2_EXP_BIT)
      flag = ibclr(flag, NOT_IN_M2_FOV_BIT)
    case default
call error('badInstrument','')
    end select
if (testFlag) print *, 'eee'

!    !...................................................................
!    ! Open the template table in the source set and get the source
!    ! spectrum information:
!
!    templateTab = table(sourceSet, 'LOGNLOGS')
!    photonSpectralIndex = real64Attribute(templateTab, 'SPECINDX')
!    call release(templateTab)
!if (testFlag) print *, 'fff'

    !...................................................................
    ! Now convert S_DENS_0 values to count rates within the supplied
    ! rate band.

!    convertToFlux = fluxDensity0ToFlux(photonSpectralIndex&
!      , rateEeVRanges, 1.0)
    allocate(rate(numSources))
    rate = fluxDensity0 * normalizedFlux * ecf * 1.0e11
if (testFlag) print *, 'ggg'

    !...................................................................
    ! Convert source Ra/Dec to both image coordinates and TELCOORD coordinates:

    allocate(srcTanX(numSources), srcTanY(numSources))
    call raDecToTan(raDeg*PiOn180, decDeg*PiOn180&
     , wcs%x%refWorld*PiOn180, wcs%y%refWorld*PiOn180&
     , srcTanX, srcTanY) ! in ssclib/coordinate
if (testFlag) print *, 'hhh'

    allocate(srcXpixels(numSources), srcYpixels(numSources))
    call tanToPixels(wcs, srcTanX, srcTanY, srcXpixels, srcYpixels)
    ! in ssclib/wcs_aux
    deallocate(srcTanX, srcTanY)
if (testFlag) print *, 'iii'

    allocate(srcThetaArcsec(numSources), srcPhi(numSources))
    call raDecToTel(raDeg*PiOn180, decDeg*PiOn180, scAttitude&
     , expStartTimeStamp, srcThetaArcsec, srcPhi)
    ! in ssclib/coordinate
if (testFlag) print *, 'jjj'

    call release(raDegCol)
    call release(decDegCol)

    call getGaussianPsfPatch(0.0, 0.0, sigmaPixels, sigmaPixels&
      , gaussianPatch)
if (testFlag) call dumpImageToFits(gaussianPatch, 'junk_gauss_patch.ds')
    gaussianPatchWidthXi = size(gaussianPatch, 1)
    gaussianPatchWidthYi = size(gaussianPatch, 2)
    gaussianPatchHalfWidthXi = floor(size(gaussianPatch, 1) / 2.0)
    gaussianPatchHalfWidthYi = floor(size(gaussianPatch, 2) / 2.0)

    !...................................................................
    ! Now work through the list of sources and add a correctly-normalized
    ! PSF image for each one that falls on a non-zero pixel of the
    ! exposure map:

    allocate(image(xSize, ySize))
    image = 0.0 ! default

if (testFlag) print *, 'kkk'
    do src = 1, numSources
if (testFlag) print *, ''
if (testFlag) print *, '  0 src=', src
!      call printEvery(src, numSources, fraction=0.05)
      errStr = sprintEvery(src, numSources, fraction=0.05) ! in ssclib/ssc_misc.
      if (errStr .ne. '') then
        errStr = '     '//errStr
        call message(appMsg, verbose)
      end if

      xi = nint(srcXpixels(src))
      yi = nint(srcYpixels(src))
    if (xi < 1 .or. xi > xSize .or. yi < 1 .or. yi > ySize) cycle
    if (.not. paddedMask(xi, yi)) cycle
if (testFlag) print *, '  1'

      if (expMapImage(xi, yi) <= 0.0) then
        select case(instrumentId)
        case(EPN)
          flag(src-1) = ibset(flag(src-1), NOT_IN_PN_EXP_BIT)
        case(EMOS1)
          flag(src-1) = ibset(flag(src-1), NOT_IN_M1_EXP_BIT)
        case(EMOS2)
          flag(src-1) = ibset(flag(src-1), NOT_IN_M2_EXP_BIT)
        end select
      end if
if (testFlag) print *, '  2'

      if (.not. btest(flag(src-1), COARSE_PSF_BIT)) then
      ! use high-precision model:
if (testFlag) print *, '  3-hi'

        call getCalPsfPatch(xSize, ySize, wcs, psfEnergyFraction&
          , psfEeV, srcThetaArcsec(src), srcPhi(src), patch, thetaGtFovRadius)
if (testFlag) print *, '  4'

        if (thetaGtFovRadius) then
          select case(instrumentId)
          case(EPN)
            flag(src-1) = ibset(flag(src-1), NOT_IN_PN_FOV_BIT)
          case(EMOS1)
            flag(src-1) = ibset(flag(src-1), NOT_IN_M1_FOV_BIT)
          case(EMOS2)
            flag(src-1) = ibset(flag(src-1), NOT_IN_M2_FOV_BIT)
          end select
        end if
if (testFlag) print *, '  5'

      else ! can use low-precision PSF model:
if (testFlag) print *, '  3-lo'

!        call getGaussianPsfPatch(srcXpixels(src), srcYpixels(src)&
!          , sigmaXpixels, sigmaYpixels, patch)

xStart = nint(srcXpixels(src)) - gaussianPatchHalfWidthXi
xFinis = xStart + gaussianPatchWidthXi - 1
yStart = nint(srcYpixels(src)) - gaussianPatchHalfWidthYi
yFinis = yStart + gaussianPatchWidthYi - 1

!print *, 'size(gaussianPatch, 1)', size(gaussianPatch, 1)
!print *, 'size(gaussianPatch, 2)', size(gaussianPatch, 2)
!print *, 'gaussianPatchHalfWidthXi', gaussianPatchHalfWidthXi
!print *, 'gaussianPatchHalfWidthYi', gaussianPatchHalfWidthYi
!print *, 'xStart,xFinis', xStart,xFinis
!print *, 'yStart,yFinis', yStart,yFinis

allocate(patch(xStart:xFinis,yStart:yFinis))
patch = gaussianPatch

if (testFlag) print *, '  4-lo'
      end if
if (testFlag) print *, '  6'

      call addPatch(image, patch, rate(src))
if (testFlag) print *, '  7'
      deallocate(patch)
    end do ! src
if (testFlag) print *, 'lll'
    deallocate(expMapImage, paddedMask)
    deallocate(srcXpixels, srcYpixels)

    !...................................................................
    ! Save the output:

    call saveImage(image, expMapSetName, rateImageSetName, tempImageName)
    ! in ssclib/save_image.
    ! Copies over all keywords (eg wcs, INSTRUM) in the expMap.
if (testFlag) print *, 'mmm'

    imageSet = dataSet(rateImageSetName, MODIFY)
    imageArray = array(imageSet, 0)
    call setAttribute(imageArray, 'PSFENERG', psfEeV, 'eV'&
      , 'The energy at which the cal PSF was obtained.')
    call setAttribute(imageArray, 'E_FRAC',  psfEnergyFraction, ''&
      , 'The energy fraction contained in the (cal) PSF array.')
    call setAttribute(block(imageArray), fluxCutoffAttr)
    call setAttribute(imageArray, 'ECF',  ecf, '10^-11 counts cm^2 / erg'&
      , comment="Ratio in the given band between count rate and flux.")
    call setAttribute(imageArray, 'S0_TO_S', normalizedFlux, 'eV'&
      , comment="Flux in the given band of a source with flux density = 1 at 1 keV.")

    numRanges = size(rateEeVRanges)
if (testFlag) print *, 'nnn'

    call setAttribute(imageArray, 'N_ERANGS'&
        , numRanges, '', 'Number of energy ranges.')
    do rangeNum = 1, numRanges
      iStr = intToStr(rangeNum) ! in ssclib/ssc_misc.
      call setAttribute(imageArray, 'ELO'//trim(iStr)&
        , rateEeVRanges(rangeNum)%lower%value, 'eV')
      call setAttribute(imageArray, 'EHI'//trim(iStr)&
        , rateEeVRanges(rangeNum)%upper%value, 'eV')
    end do

    call setAttribute(imageArray, 'ELO', rateEeVRanges(1)%lower%value, 'eV')
    call setAttribute(imageArray, 'EHI', rateEeVRanges(numRanges)%upper%value&
      , 'eV')
    ! I am still writing these because eboxdetect, emldetect etc don't
    ! recognize bands composed of a union of sub-bands.

    call copyBlock(imageSet, block(sourceSet, 'LOGNLOGS'))

    call release(imageSet)
    call release(sourceSet)

    deallocate(gaussianPatch)

if (testFlag) print *, 'zzz eimsim'
  end subroutine eimsim

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    expMapSetName = stringParameter('expmapset')
    sourceListSetName = stringParameter('srclistset')
    paddedMaskSetName = stringParameter('maskset')
    rateImageSetName = stringParameter('outset')
    tempImageName = stringParameter('tempset')
    obsId = stringParameter('obsid')
    ecf = realParameter('ecf')
    normalizedFlux = realParameter('normedflux')
    psfEnergyFraction = realParameter('psfefrac')

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine getGaussianPsfPatch(srcXpixels, srcYpixels&
  , sigmaXpixels, sigmaYpixels, patch)
  !---------------------------------------------------------------------
  !
    implicit none

    real(single), intent(in) ::&
      srcXpixels, srcYpixels, sigmaXpixels, sigmaYpixels
    real(single), pointer :: patch(:,:)

    real(single), parameter :: patchWidthInSigmas = 7.0
    integer(int16) :: xStart, xFinis, yStart, yFinis, i
    real(single), allocatable :: x(:), y(:), xExp(:), yExp(:)
    integer(int16), pointer :: xi(:), yi(:)

    !-------------------------------------------------------------------
    ! Start

    xStart = floor(  srcXpixels - patchWidthInSigmas * sigmaXpixels)
    xFinis = ceiling(srcXpixels + patchWidthInSigmas * sigmaXpixels)
    yStart = floor(  srcYpixels - patchWidthInSigmas * sigmaYpixels)
    yFinis = ceiling(srcYpixels + patchWidthInSigmas * sigmaYpixels)

    allocate(patch(xStart:xFinis,yStart:yFinis))
    allocate(x(xStart:xFinis), xExp(xStart:xFinis))
    allocate(y(yStart:yFinis), yExp(yStart:yFinis))
    allocate(xi(xStart:xFinis), yi(yStart:yFinis))

    xi = (/(i,i=xStart, xFinis)/)
    x = (real(xi, single) - srcXpixels) / RootTwo / sigmaXpixels

    yi = (/(i,i=yStart, yFinis)/)
    y = (real(yi, single) - srcYpixels) / RootTwo / sigmaYpixels
    xExp = exp(-(x*x))
    yExp = exp(-(y*y))

    patch = spread(xExp, 2, (1 + yFinis - yStart))&
          * spread(yExp, 1, (1 + xFinis - xStart))

!    call vignetPatch(patch)

    deallocate(xi, x, xExp, yi, y, yExp)

  end subroutine getGaussianPsfPatch

  !---------------------------------------------------------------------
  subroutine addPatch(image, patch, rate)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(inout) :: image(:,:)
    real(single), pointer :: patch(:,:)
    real(single), intent(in) :: rate

    integer :: xStart, xFinis, yStart, yFinis, xSize, ySize

    !-------------------------------------------------------------------
    ! Start

    xStart = lbound(patch, 1)
    xFinis = ubound(patch, 1)
    yStart = lbound(patch, 2)
    yFinis = ubound(patch, 2)
    xSize = size(image, 1)
    ySize = size(image, 2)

    if (xStart > xSize .or. xFinis < 1 .or. yStart > ySize .or. yFinis < 1) then
  return
    end if

    xStart = max(1, xStart)
    xFinis = min(xSize, xFinis)
    yStart = max(1, yStart)
    yFinis = min(ySize, yFinis)

!    image(xStart:xFinis, yStart:yFinis) = image(xStart:xFinis, yStart:yFinis)&
!      + patch(xStart:xFinis, yStart:yFinis) * rate / sum(patch)
    image(xStart:xFinis, yStart:yFinis) = image(xStart:xFinis, yStart:yFinis)&
      + patch(xStart:xFinis, yStart:yFinis) * rate

  end subroutine addPatch

  !---------------------------------------------------------------------
  subroutine getCalPsfPatch(xSize, ySize, wcs, psfEnergyFraction&
  , psfEeV, srcThetaArcsec, srcPhi, patch, thetaGTFOVRadius)
  !---------------------------------------------------------------------
  ! Requires cal to have been set up for instrumentId and accuracyLevel.

    implicit none

    integer(int16), intent(in)  :: xSize, ySize
    type(WcsT),     intent(in)  :: wcs
    real(double),   intent(in)  :: psfEnergyFraction, psfEeV,&
                                   srcThetaArcsec, srcPhi
    real(single),   pointer     :: patch(:,:)
    logical,        intent(out) :: thetaGTFOVRadius

    real(double) :: effectiveSrcThetaArcsec, psfRadius
    type(PsfValidityRangesT) :: validPsfRanges
    type(PsfT) :: psfHandle
    type(PsfWindowT) :: psfWindow
    type(PsfBinSizeT) :: psfPixelSizeMm
    real(single), pointer :: psfImage(:,:)
    integer(int16) :: startXi, finisXi, startYi, finisYi
    type(Point2dT), allocatable ::&
      psfImagePixelCorners(:,:)
    type(Point2dDbleT), allocatable ::&
      psfImgPixCrnrTel(:,:), psfImgPixCrnrTan(:,:), psfImgPixCrnrDet(:,:)
      ! types defined in coordinate.f90.
    integer :: foreignStatus
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start
if (testFlag) print *, '    aaa getCalPsfPatch'

    thetaGTFOVRadius = .false.

    validPsfRanges = CAL_psfValidityRanges()

    if  (psfEeV < validPsfRanges%eMin&
    .or. psfEeV > validPsfRanges%eMax)&
call error('energyOutOfRange',"")

    if (srcThetaArcsec < validPsfRanges%thetaMin)&
call error('thetaOutOfRange',"")

    if (srcPhi < validPsfRanges%phiMin&
    .or. srcPhi > validPsfRanges%phiMax)&
call error('phiOutOfRange',"")

    if (srcThetaArcsec > validPsfRanges%thetaMax) then
      thetaGTFOVRadius = .true.
      effectiveSrcThetaArcsec = validPsfRanges%thetaMax&
        * (1.0d0 - epsilon(srcThetaArcsec))
    else
      effectiveSrcThetaArcsec = srcThetaArcsec
    end if

    psfHandle = CAL_getPSF(psfEeV, effectiveSrcThetaArcsec, srcPhi)
    psfPixelSizeMm = CAL_psfBinSize(psfHandle)
    psfRadius = CAL_psfEncircledEnergy(psfHandle, fraction=psfEnergyFraction)
!    psfWindow = CAL_psfEnboxedEnergy(psfHandle, fraction=psfEnergyFraction)
    psfWindow%halfX = psfRadius * (1.0 + vignettedSkirtFraction)
    psfWindow%halfY = psfRadius * (1.0 + vignettedSkirtFraction)

    call CAL_psfImage(psfHandle, psfWindow, psfImage)

    ! Normalize, then vignet the psf image edges:

!    psfImage = psfImage / (sum(psfImage) / psfEnergyFraction)
!    call vignetPatch(psfImage)
    psfImage = psfImage * psfPixelSizeMm%binX * psfPixelSizeMm%binY
    call vignetPatch(vignettedSkirtFraction, psfImage)

    allocate(psfImgPixCrnrTel(size(psfImage,1)+1, size(psfImage,2)+1))
    allocate(psfImgPixCrnrDet(size(psfImage,1)+1, size(psfImage,2)+1))
    allocate(psfImgPixCrnrTan(size(psfImage,1)+1, size(psfImage,2)+1))
    allocate(psfImagePixelCorners(size(psfImage,1)+1, size(psfImage,2)+1))

    call getPsfImagePixelCorners(srcThetaArcsec, srcPhi&
      , psfPixelSizeMm, psfImage, psfImgPixCrnrTel)
    ! in ssclib/coordinate.
    !
    ! The pairs of numbers (type(Point2dT)) in this array represent the
    ! coordinates of the corners of the psfImage pixels, expressed in
    ! the exposure-map image (ie, sky) coordinate system.

    call psfToDet(psfImgPixCrnrTel%x, psfImgPixCrnrTel%y&
      , psfImgPixCrnrDet%x, psfImgPixCrnrDet%y)
    ! in ssclib/coordinate.

    call detToTan(psfImgPixCrnrDet%x, psfImgPixCrnrDet%y&
      , wcs%x%refWorld*PiOn180, wcs%y%refWorld*PiOn180&
      , scAttitude, expStartTimeStamp, psfImgPixCrnrTan%x&
      , psfImgPixCrnrTan%y)
    ! in ssclib/coordinate.

    call tanToPixels(wcs, psfImgPixCrnrTan%x, psfImgPixCrnrTan%y&
      , psfImagePixelCorners%x, psfImagePixelCorners%y)
    ! in ssclib/wcs_aux.

    startXi = max(1,     nint(minval(psfImagePixelCorners%x)))
    finisXi = min(xSize, nint(maxval(psfImagePixelCorners%x), int16))
    startYi = max(1,     nint(minval(psfImagePixelCorners%y)))
    finisYi = min(ySize, nint(maxval(psfImagePixelCorners%y), int16))
    allocate(patch(startXi:finisXi,startYi:finisYi))

    psfImagePixelCorners%x = psfImagePixelCorners%x - real(startXi - 1)
    psfImagePixelCorners%y = psfImagePixelCorners%y - real(startYi - 1)

    call regridCartesian(psfImagePixelCorners, psfImage&
      , patch, status=foreignStatus) ! in ssclib/regridders.

if (foreignStatus < 0) call error('','')
    if (foreignStatus > 0) then
      print *, 'regrid exit status:', foreignStatus
    end if

    call CAL_releaseMemory(psfImage)
if (testFlag) print *, '    zzz getCalPsfPatch'

  end subroutine getCalPsfPatch

  !---------------------------------------------------------------------
  subroutine vignetPatch(vigSkirtFraction, patch)
  !---------------------------------------------------------------------
  ! This is meant to take an array containing a source PSF image and
  ! multiply it by a local vignetting function, so as to avoid the otherwise
  ! sharp drop-off at the edge of the array.
  !
  ! This has now been a bit modified to take account of another problem.
  ! emldetect assumes of course that the psf of the source it is measuring
  ! is of normal shape - at least out as far as the energy-fraction
  ! cutoff specified via the emldetect parameter --ecut. Hence it is no
  ! good, when making these simulated images, to multiply the whole PSF
  ! by some vignetting function, because that will distort its shape. So
  ! what I do now is retain the pristine PSF out to the ecut edge, but
  ! multiply by a vignetting or feathering function outboard of this.

    implicit none

    real(single), intent(in)    :: vigSkirtFraction
    real(single), intent(inout) :: patch(:,:)

    integer(int16) :: xSize, ySize, i, xPristineHalfSize, yPristineHalfSize
    real(single) ::&
      xHalfSize, yHalfSize,&
      xVector(size(patch, 1)), yVector(size(patch, 2)),&
      xVigFunction(size(patch, 1)), yVigFunction(size(patch, 2))

    !-------------------------------------------------------------------
    ! Start

    xSize = size(patch, 1)
    ySize = size(patch, 2)
    xHalfSize = xSize / 2.0 - 0.5
    yHalfSize = ySize / 2.0 - 0.5
    xPristineHalfSize = xHalfSize / (1.0 + vigSkirtFraction)
    yPristineHalfSize = yHalfSize / (1.0 + vigSkirtFraction)

    xVector = (/(abs(i - (xHalfSize + 1.0)), i=1, xSize)/)
    ! The +1 prevents patch being 0 all around the rim, which would be
    ! a waste of pixels.

    where (xVector > xPristineHalfSize)
      xVigFunction = (xVector - xPristineHalfSize)&
      / ((xHalfSize + 1.0) - xPristineHalfSize)
      xVigFunction = 1.0 - xVigFunction * xVigFunction
    elsewhere
      xVigFunction = 1.0
    end where

    yVector = (/(abs(i - (yHalfSize + 1.0)), i=1, ySize)/)
    ! The +1 prevents patch being 0 all around the rim, which would be
    ! a waste of pixels.

    where (yVector > yPristineHalfSize)
      yVigFunction = (yVector - yPristineHalfSize)&
      / ((yHalfSize + 1.0) - yPristineHalfSize)
      yVigFunction = 1.0 - yVigFunction * yVigFunction
    elsewhere
      yVigFunction = 1.0
    end where

    patch = patch * spread(xVigFunction, 2, ySize)&
                  * spread(yVigFunction, 1, xSize)

  end subroutine vignetPatch

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module eimsim_mod
