!
! Module: wcs_aux
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Routines to facilitate handling of wcs values.
! 
! Comments: TODO:
!	As of 2004-01-30 the library generally supports the WCS system outlined in Greisen and Calabretta A&A 395, 1061-1075 (2002). However, a couple of caveats: (i) only the RA---TAN/DEC--TAN projection is supported; (ii) if CD1_1 is present, all CDi_j kwds are required to be present; (iii) PCi_j kwds are not yet supported.
!
! $Id: wcs_aux.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module l_wcs_aux

  use dal
  use errorHandling	! used in dumpWcs and the pixels<->tan routines.

  use constants
  use dal_aux	! contains subs getRealKwd(), getRealKwd(), getStringKwd().
  use ssc_misc	! for approxEqual
  use array_utils	! for vectorCross

  implicit none 

  private 

  public initializeWcsT
  public initializeWcsDualT
  private checkWcsDualType
  private getDefaultXWorldUnit
  private getDefaultYWorldUnit
  public hasWcsDual
  private hasWcsDualGeneric
  public hasWcs
  private hasWcsGeneric
  public readWcsDual
  public readWcs
  public writeWcsDual
  public writeWcs
  public writeNextWcs
  public deleteWcs
  public dumpWcs

  public easifyWcsType
  public wcsAreEquivalent
  public wcsPixelAreaWorld
  public checkImageEquivalence
  public convertOldToNew
  public convertNewToOld

  public pixelsToTan
  public tanToPixels
  public tanRadiusToPixelsEllipse

!public readWcsSourceList
!interface readWcsSourceList
!  module procedure readWcsSourceListSet, readWcsSourceListName
!end interface
!!*** this is necessary until src lists conform to ogip.


  interface hasWcsDual
    module procedure hasWcsDualColumnsName, hasWcsDualColumns
    module procedure hasWcsDualArray
    module procedure hasWcsDualSet, hasWcsDualSetName
    module procedure hasWcsDualTable
  end interface

  interface hasWcs
    module procedure hasWcsColumnName, hasWcsColumn
    module procedure hasWcsArray
    module procedure hasWcsSet
    module procedure hasWcsTable
  end interface

  interface readWcsDual
    module procedure readWcsDualColumnsName, readWcsDualColumns
    module procedure readWcsDualSetArray,    readWcsDualArray
    module procedure readWcsDualSet,         readWcsDualSetName
    module procedure readWcsDualSetArrayName
    module procedure readWcsDualTable,       readWcsDualSetArrayNumber
  end interface

  interface readWcs
    module procedure readWcsColumnName, readWcsColumn
    module procedure readWcsArray
    module procedure readWcsSet
    module procedure readWcsTable
  end interface

  interface writeWcsDual
    module procedure writeWcsDualColumns, writeWcsDualArray
  end interface

  interface writeWcs
    module procedure writeWcsColumn, writeWcsArray
  end interface

  interface writeNextWcs
    module procedure writeNextWcsArray
  end interface

  interface dumpWcs
    module procedure dumpWcsDual, dumpWcsVec, dumpWcsSingle
  end interface

  interface easifyWcsType
    module procedure easifyWcsTypeScalar
    module procedure easifyWcsTypeVector
  end interface

  interface tanToPixels
    module procedure tanToPixelsScalar, tanToPixelsVector, tanToPixelsArray
  end interface

  interface tanRadiusToPixelsEllipse
    module procedure tanRadToPixEllipseScalar, tanRadToPixEllipseVector
!, tanRadToPixEllipseArray
  end interface

  interface pixelsToTan
    module procedure pixelsToTanScalar, pixelsToTanVector, pixelsToTanArray
  end interface


  type, public :: WcsT
    character(len=8) ::&
      ctype,&		! CTYPE
      worldUnit		! CUNIT
    real(double) ::&
      refPixels,&	! CRPIX
      refWorld,&	! CRVAL
      matrixOnAxis,&	! CD1_
      matrixOffAxis,&	! CD2_
      pixelSizeWorld,&	! CDELT (old style - replaced by CD1_)
!      rotAngle		! CROTA
    logical ::&
      crotFound,&
      styleIsNew
  end type WcsT

  type, public :: WcsDualT
    type(WcsT) :: x, y
    real(double) ::&
      rotAngle		! CROTA
    logical ::&
      crotFound,&
      styleIsNew
  end type WcsDualT

logical :: useNewWcsStyle = .false.

contains 

  !---------------------------------------------------------------------
  subroutine initializeWcsDualT(wcsDual)
  !---------------------------------------------------------------------

    implicit none

    type(WcsDualT), intent(out) :: wcsDual

    !-------------------------------------------------------------------
    ! Start

    call initializeWcsT(wcsDual%x)
    call initializeWcsT(wcsDual%y)
    wcsDual%crotFound = .false.
    wcsDual%styleIsNew = .false.
    wcsDual%rotAngle = 0.0d0

  end subroutine initializeWcsDualT

  !---------------------------------------------------------------------
  subroutine initializeWcsT(wcs)
  !---------------------------------------------------------------------

    implicit none

    type(WcsT), intent(out) :: wcs

    !-------------------------------------------------------------------
    ! Start

    wcs%ctype          = ''
    wcs%worldUnit      = ''
    wcs%matrixOnAxis   = 1.0d0
    wcs%matrixOffAxis  = 0.0d0
    wcs%refPixels      = 0.0d0
    wcs%refWorld       = 0.0d0
    wcs%pixelSizeWorld = 0.0d0
!    wcs%rotAngle       = 0.0d0
    wcs%crotFound      = .false.
    wcs%styleIsNew     = .false.

  end subroutine initializeWcsT

  !---------------------------------------------------------------------
  subroutine checkWcsDualType(wcsDual)
  !---------------------------------------------------------------------

    implicit none

    type(WcsDualT), intent(in) :: wcsDual

    !-------------------------------------------------------------------
    ! Start:

    select case(wcsDual%x%ctype)
    case('RA---TAN')
    case default
call error('notSupported',"ssclib:checkWcsDualType: This value of CTYPE1 is not yet supported.")
    end select

    select case(wcsDual%y%ctype)
    case('DEC--TAN')
    case default
call error('notSupported',"ssclib:checkWcsDualType: This value of CTYPE2 is not yet supported.")
    end select

  end subroutine checkWcsDualType

  !---------------------------------------------------------------------
  subroutine getDefaultXWorldUnit(wcsDual)
  !---------------------------------------------------------------------

    implicit none

    type(WcsDualT), intent(inout) :: wcsDual

    !-------------------------------------------------------------------
    ! Start:

    select case(wcsDual%x%ctype)
    case('RA---TAN')
      wcsDual%x%worldUnit = 'deg'
    case default
call error('notSupported',"ssclib:getDefaultXWorldUnit: This value of CTYPE1 is not yet supported.")
    end select

  end subroutine getDefaultXWorldUnit

  !---------------------------------------------------------------------
  subroutine getDefaultYWorldUnit(wcsDual)
  !---------------------------------------------------------------------

    implicit none

    type(WcsDualT), intent(inout) :: wcsDual

    !-------------------------------------------------------------------
    ! Start:

    select case(wcsDual%y%ctype)
    case('DEC--TAN')
      wcsDual%y%worldUnit = 'deg'
    case default
call error('notSupported',"ssclib:getDefaultYWorldUnit: This value of CTYPE2 is not yet supported.")
    end select

  end subroutine getDefaultYWorldUnit


  !---------------------------------------------------------------------
  logical function hasWcsDualColumnsName(tab, xColName, yColName)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in)  :: tab
    character(*), intent(in)  :: xColName, yColName

    type(ColumnT) :: xCol, yCol

    !-------------------------------------------------------------------
    ! Start:

    xCol = column(tab, xColName)
    yCol = column(tab, yColName)

    hasWcsDualColumnsName = hasWcsDualColumns(xCol, yCol)

    call release(xCol)
    call release(yCol)

  end function hasWcsDualColumnsName

  !---------------------------------------------------------------------
  logical function hasWcsDualColumns(xCol, yCol)
  !---------------------------------------------------------------------

    implicit none

    type(ColumnT), intent(in) :: xCol, yCol

    !-------------------------------------------------------------------
    ! Start:

    hasWcsDualColumns = .false. ! default

    if (hasWcsColumn(xCol) .and. hasWcsColumn(yCol))&
      hasWcsDualColumns = .true.

  end function hasWcsDualColumns

  !---------------------------------------------------------------------
  logical function hasWcsDualColumns_20060317(xCol, yCol)
  !---------------------------------------------------------------------

    implicit none

    type(ColumnT), intent(in) :: xCol, yCol

    !-------------------------------------------------------------------
    ! Start:

    hasWcsDualColumns_20060317 = .false. ! default

    if (    hasAttribute(xCol, 'TCRPX')&
    .and.   hasAttribute(yCol, 'TCRPX')&
    .and.   hasAttribute(xCol, 'TCRVL')&
    .and.   hasAttribute(yCol, 'TCRVL')&
    .and.   hasAttribute(xCol, 'TCTYP')&
    .and.   hasAttribute(yCol, 'TCTYP')&
    .and. ((hasAttribute(xCol, 'TCDLT')&
    .and.   hasAttribute(yCol, 'TCDLT'))&
    .or.   (hasAttribute(xCol, 'TCD1_')&
    .and.   hasAttribute(yCol, 'TCD1_')&
    .and.   hasAttribute(xCol, 'TCD2_')&
    .and.   hasAttribute(yCol, 'TCD2_'))))&
      hasWcsDualColumns_20060317 = .true.

  end function hasWcsDualColumns_20060317

  !---------------------------------------------------------------------
  logical function hasWcsDualSetName(setName)
  !---------------------------------------------------------------------
  ! To be used with arrays in the primary HDU where it is not certain
  ! whether the wcs keywords are in the primary header or the 'array' header.

    implicit none

    character(*), intent(in)  :: setName

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    hasWcsDualSetName = .false. ! default

    set = dataSet(setName, READ)
    inArray = array(set, 0, READ)
    if (hasWcsDualArray(inArray)) then
      hasWcsDualSetName = .true.
      if (hasWcsDualSet(set)) call warning('setAlsoHasWcs',"ssclib:hasWcsDualSetName: &
        &Both the dataset header and the array header contain WCS keywords. &
        &The correct place for them is in the array header only.")

    else if (hasWcsDualSet(set)) then
      hasWcsDualSetName = .true.
      call warning('onlySetHadWcs',"ssclib:hasWcsDualSetName: The dataset header &
        &contains WCS keywords but not the array header. It should be the other &
        &way around.")
    end if

    call release(set)

  end function hasWcsDualSetName

  !---------------------------------------------------------------------
  logical function hasWcsDualSet(set)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in) :: set

    type(AttributableT) :: generic

    !-------------------------------------------------------------------
    ! Start

    generic = attributable(set)
    hasWcsDualSet = hasWcsDualGeneric(generic)

  end function hasWcsDualSet

  !---------------------------------------------------------------------
  logical function hasWcsDualArray(inArray)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in) :: inArray

    type(AttributableT) :: generic

    !-------------------------------------------------------------------
    ! Start

    generic = attributable(inArray)
    hasWcsDualArray = hasWcsDualGeneric(generic)

  end function hasWcsDualArray

  !---------------------------------------------------------------------
  logical function hasWcsDualTable(tab)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab

    type(AttributableT) :: generic

    !-------------------------------------------------------------------
    ! Start

    generic = attributable(tab)
    hasWcsDualTable = hasWcsDualGeneric(generic)

  end function hasWcsDualTable

  !---------------------------------------------------------------------
  logical function hasWcsDualGeneric(generic)
  !---------------------------------------------------------------------

    implicit none

    type(AttributableT), intent(in)  :: generic

    !-------------------------------------------------------------------
    ! Start:

    hasWcsDualGeneric = .false. ! default

    if (hasWcsGeneric(generic,1) .and. hasWcsGeneric(generic,2))&
      hasWcsDualGeneric = .true.

  end function hasWcsDualGeneric

  !---------------------------------------------------------------------
  logical function hasWcsDualGeneric_20060317(generic)
  !---------------------------------------------------------------------

    implicit none

    type(AttributableT), intent(in)  :: generic

    !-------------------------------------------------------------------
    ! Start:

    hasWcsDualGeneric_20060317 = .false. ! default

    if (    hasAttribute(generic, 'CRPIX1')&
    .and.   hasAttribute(generic, 'CRPIX2')&
    .and.   hasAttribute(generic, 'CRVAL1')&
    .and.   hasAttribute(generic, 'CRVAL2')&
    .and.   hasAttribute(generic, 'CTYPE1')&
    .and.   hasAttribute(generic, 'CTYPE2')&
    .and. ((hasAttribute(generic, 'CDELT1')&
    .and.   hasAttribute(generic, 'CDELT2'))&
    .or.   (hasAttribute(generic, 'CD1_1')&
    .and.   hasAttribute(generic, 'CD1_2')&
    .and.   hasAttribute(generic, 'CD2_1')&
    .and.   hasAttribute(generic, 'CD2_2'))))&
      hasWcsDualGeneric_20060317 = .true.

  end function hasWcsDualGeneric_20060317

  !---------------------------------------------------------------------
  logical function hasWcsColumnName(tab, colName)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in)  :: tab
    character(*), intent(in)  :: colName

    type(ColumnT) :: col

    !-------------------------------------------------------------------
    ! Start:

    col = column(tab, colName)
    hasWcsColumnName = hasWcsColumn(col)
    call release(col)

  end function hasWcsColumnName

  !---------------------------------------------------------------------
  logical function hasWcsColumn(col)
  !---------------------------------------------------------------------

    implicit none

    type(ColumnT), intent(in) :: col

    !-------------------------------------------------------------------
    ! Start:

    hasWcsColumn = .false. ! default

    if (    hasAttribute(col, 'TCRPX')&
    .and.   hasAttribute(col, 'TCRVL')&
    .and.   hasAttribute(col, 'TCTYP')&
    .and. ((hasAttribute(col, 'TCDLT'))&
    .or.   (hasAttribute(col, 'TCD1_')&
    .and.   hasAttribute(col, 'TCD2_'))))&
      hasWcsColumn = .true.

  end function hasWcsColumn

  !---------------------------------------------------------------------
  logical function hasWcsSet(set, i)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in) :: set
    integer,        intent(in) :: i

    type(AttributableT) :: generic

    !-------------------------------------------------------------------
    ! Start

    generic = attributable(set)
    hasWcsSet = hasWcsGeneric(generic, i)

  end function hasWcsSet

  !---------------------------------------------------------------------
  logical function hasWcsArray(inArray, i)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in) :: inArray
    integer,      intent(in) :: i

    type(AttributableT) :: generic

    !-------------------------------------------------------------------
    ! Start

    generic = attributable(inArray)
    hasWcsArray = hasWcsGeneric(generic, i)

  end function hasWcsArray

  !---------------------------------------------------------------------
  logical function hasWcsTable(tab, i)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab
    integer,      intent(in) :: i

    type(AttributableT) :: generic

    !-------------------------------------------------------------------
    ! Start

    generic = attributable(tab)
    hasWcsTable = hasWcsGeneric(generic, i)

  end function hasWcsTable

  !---------------------------------------------------------------------
  logical function hasWcsGeneric(generic, i)
  !---------------------------------------------------------------------

    implicit none

    type(AttributableT), intent(in) :: generic
    integer,             intent(in) :: i

    character(6) :: ctypeKwd, crpixKwd, crvalKwd, cdeltKwd, crotaKwd, cunitKwd
    character(5) :: cd1Kwd, cd2Kwd

    !-------------------------------------------------------------------
    ! Start

if (i>9) call error('',"")

    write(ctypeKwd, '(a5, i1)') 'CTYPE', i
    write(crpixKwd, '(a5, i1)') 'CRPIX', i
    write(crvalKwd, '(a5, i1)') 'CRVAL', i
    write(cd1Kwd,   '(a4, i1)') 'CD1_', i
    write(cd2Kwd,   '(a4, i1)') 'CD2_', i
    write(cdeltKwd, '(a5, i1)') 'CDELT', i
    write(crotaKwd, '(a5, i1)') 'CROTA', i
    write(cunitKwd, '(a5, i1)') 'CUNIT', i

    hasWcsGeneric = .false. ! default

    if (    hasAttribute(generic, ctypeKwd)&
    .and.   hasAttribute(generic, crpixKwd)&
    .and.   hasAttribute(generic, crvalKwd)&
    .and. ((hasAttribute(generic, cdeltKwd))&
    .or.   (hasAttribute(generic, cd1Kwd)&
    .and.   hasAttribute(generic, cd2Kwd))))&
      hasWcsGeneric = .true.

  end function hasWcsGeneric


  !---------------------------------------------------------------------
  subroutine readWcsDualColumnsName(tab, xColName, yColName, wcsDual)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in)  :: tab
    character(*), intent(in)  :: xColName, yColName
    type(WcsDualT),   intent(out) :: wcsDual

    type(ColumnT) :: xCol, yCol

    !-------------------------------------------------------------------
    ! Start:

    xCol = column(tab, xColName)
    yCol = column(tab, yColName)

!    call readWcsColumns(xCol, yCol, wcsDual)
    call readWcsColumn(xCol, wcsDual%x)
    call readWcsColumn(yCol, wcsDual%y)

    call release(xCol)
    call release(yCol)

  end subroutine readWcsDualColumnsName

  !---------------------------------------------------------------------
  subroutine readWcsDualColumns(xCol, yCol, wcsDual)
  !---------------------------------------------------------------------

!******** change this so it calls readWcsColumn for x and y cols, then does the old->new thing afterwards if necessary??

    implicit none

    type(ColumnT), intent(in) :: xCol, yCol
    type(WcsDualT),   intent(out) :: wcsDual

    real(double) :: rotAngle, xPixelSizeWorld, yPixelSizeWorld

    !-------------------------------------------------------------------
    ! Start:

    wcsDual%x%ctype          = stringAttribute(xCol, 'TCTYP')
    wcsDual%y%ctype          = stringAttribute(yCol, 'TCTYP')

    wcsDual%x%refPixels      = real64Attribute(xCol, 'TCRPX')
    wcsDual%y%refPixels      = real64Attribute(yCol, 'TCRPX')
    wcsDual%x%refWorld       = real64Attribute(xCol, 'TCRVL')
    wcsDual%y%refWorld       = real64Attribute(yCol, 'TCRVL')

    wcsDual%crotFound = .false. ! default
    if (hasAttribute(xCol, "TCD1_")) then
      wcsDual%x%matrixOnAxis  = real64Attribute(xCol, 'TCD1_')
      wcsDual%x%matrixOffAxis = real64Attribute(xCol, 'TCD2_')
      wcsDual%y%matrixOffAxis = real64Attribute(yCol, 'TCD1_')
      wcsDual%y%matrixOnAxis  = real64Attribute(yCol, 'TCD2_')

    else
      xPixelSizeWorld = real64Attribute(xCol, 'TCDLT')
      yPixelSizeWorld = real64Attribute(yCol, 'TCDLT')
      if (hasAttribute(yCol, "TCROT")) then
        rotAngle = real64Attribute(yCol, 'TCROT')
        wcsDual%crotFound = .true.

      else
        rotAngle = 0.0d0

! See comments in writeWcsColumns.
if (hasAttribute(parent(xCol), 'TMPTCROT')) then
  call warning('tcrotSubstituteFound',"ssclib:readWcsColumns: The DAL &
    &interface does not yet enable one to write the WCS roll angle keyword &
    &as a column attribute; until this is fixed, I'm writing this value to &
    &the TMPTCROT keyword in the table. Such a keyword has been detected and &
    &the roll angle will be read from it.")
  rotAngle = real64Attribute(parent(xCol), 'TMPTCROT') * PiOn180
end if
      end if

      call convertOldToNew(xPixelSizeWorld, yPixelSizeWorld, rotAngle, wcsDual)
    end if

    if (hasAttribute(xCol, "TCUNI")) then
      wcsDual%x%worldUnit = stringAttribute(xCol, 'TCUNI')
    else
      call getDefaultXWorldUnit(wcsDual)
    end if

    if (hasAttribute(yCol, "TCUNI")) then
      wcsDual%y%worldUnit = stringAttribute(yCol, 'TCUNI')
    else
      call getDefaultYWorldUnit(wcsDual)
    end if

  end subroutine readWcsDualColumns

  !---------------------------------------------------------------------
  subroutine readWcsDualSetName(setName, wcsDual)
  !---------------------------------------------------------------------
  ! To be used with arrays in the primary HDU where it is not certain
  ! whether the wcs keywords are in the primary header or the 'array' header.

    implicit none

    character(*), intent(in)  :: setName
    type(WcsDualT),   intent(out) :: wcsDual

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(setName, READ)
    inArray = array(set, 0, READ)
    call readWcsDualSetArray(set, inArray, wcsDual)

    call release(set)

  end subroutine readWcsDualSetName

  !---------------------------------------------------------------------
  subroutine readWcsDualSetArrayName(setName, arrayName, wcsDual)
  !---------------------------------------------------------------------
  ! To be used with arrays in the primary HDU where it is not certain
  ! whether the wcs keywords are in the primary header or the 'array' header.

    implicit none

    character(*), intent(in)  :: setName, arrayName
    type(WcsDualT),   intent(out) :: wcsDual

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(setName, READ)
    inArray = array(set, arrayName, READ)
    call readWcsDualSetArray(set, inArray, wcsDual)

    call release(set)

  end subroutine readWcsDualSetArrayName

  !---------------------------------------------------------------------
  subroutine readWcsDualSetArrayNumber(setName, arrayNumber, wcsDual)
  !---------------------------------------------------------------------
  ! To be used with arrays in the primary HDU where it is not certain
  ! whether the wcs keywords are in the primary header or the 'array' header.

    implicit none

    character(*),   intent(in)  :: setName
    integer(int32), intent(in)  :: arrayNumber
    type(WcsDualT),     intent(out) :: wcsDual

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(setName, READ)
    inArray = array(set, arrayNumber, READ)
    call readWcsDualSetArray(set, inArray, wcsDual)

    call release(set)

  end subroutine readWcsDualSetArrayNumber

  !---------------------------------------------------------------------
  subroutine readWcsDualSetArray(set, inArray, wcsDual)
  !---------------------------------------------------------------------
  ! To be used with arrays in the primary HDU where it is not certain
  ! whether the wcs keywords are in the primary header or the 'array' header.

    implicit none

    type(DataSetT), intent(in)  :: set
    type(ArrayT),   intent(in)  :: inArray
    type(WcsDualT),     intent(out) :: wcsDual

    type(WcsDualT) :: setWcsDual

    !-------------------------------------------------------------------
    ! Start:

    if (hasWcsDualSet(set)) then ! set contains the wcs keywords.
      call readWcsDualSet(set, setWcsDual)
      ! Now check to see if the array block also has a set:

      if (hasWcsDualArray(inArray)) then ! array also contains the wcs keywords.
        if (hasWcsDualSet(set)) call warning('setAlsoHasWcs',"ssclib:readWcsDualSetArray: &
          &Both the dataset header and the array header contain WCS keywords. &
          &The correct place for them is in the array header only.")
        call readWcsDualArray(inArray, wcsDual)

      else ! array doesn't contains any wcs keywords.
        call warning('onlySetHadWcs',"ssclib:readWcsDualSetArray: The dataset header &
          &contains WCS keywords but not the array header. It should be the other &
          &way around.")
        wcsDual = setWcsDual
      end if
    else ! set header doesn't contain any wcs keywords. Check the array:
      call readWcsDualArray(inArray, wcsDual)
    end if

  end subroutine readWcsDualSetArray

  !---------------------------------------------------------------------
  subroutine readWcsDualArray(inArray, wcsDual)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT),   intent(in)  :: inArray
    type(WcsDualT), intent(out) :: wcsDual

    type(AttributableT) :: generic

    !-------------------------------------------------------------------
    ! Start:

    generic = attributable(inArray)
    call readWcsDualGeneric(generic, wcsDual)

  end subroutine readWcsDualArray

  !---------------------------------------------------------------------
  subroutine readWcsDualSet(set, wcsDual)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in)  :: set
    type(WcsDualT),     intent(out) :: wcsDual

    type(AttributableT) :: generic

    !-------------------------------------------------------------------
    ! Start:

    generic = attributable(set)
    call readWcsDualGeneric(generic, wcsDual)

  end subroutine readWcsDualSet

  !---------------------------------------------------------------------
  subroutine readWcsDualTable(tab, wcsDual)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in)  :: tab
    type(WcsDualT),   intent(out) :: wcsDual

    type(AttributableT) :: generic

    !-------------------------------------------------------------------
    ! Start:

    generic = attributable(tab)
    call readWcsDualGeneric(generic, wcsDual)

  end subroutine readWcsDualTable

  !---------------------------------------------------------------------
  subroutine readWcsDualGeneric(generic, wcsDual)
  !---------------------------------------------------------------------

!******** change this so it calls readWcsGeneric for 1 and 2, then does the old->new thing afterwards if necessary??

    implicit none

    type(AttributableT), intent(in)  :: generic
    type(WcsDualT),          intent(out) :: wcsDual

    real(double) :: rotAngle, xPixelSizeWorld, yPixelSizeWorld

    !-------------------------------------------------------------------
    ! Start:

    wcsDual%x%ctype     = stringAttribute(generic, 'CTYPE1')
    wcsDual%y%ctype     = stringAttribute(generic, 'CTYPE2')

    wcsDual%x%refPixels = real64Attribute(generic, 'CRPIX1')
    wcsDual%y%refPixels = real64Attribute(generic, 'CRPIX2')
    wcsDual%x%refWorld  = real64Attribute(generic, 'CRVAL1')
    wcsDual%y%refWorld  = real64Attribute(generic, 'CRVAL2')

    wcsDual%crotFound = .false. ! default
    if (hasAttribute(generic, "CD1_1")) then
      wcsDual%x%matrixOnAxis  = real64Attribute(generic, 'CD1_1')
      wcsDual%x%matrixOffAxis = real64Attribute(generic, 'CD2_1')
      wcsDual%y%matrixOffAxis = real64Attribute(generic, 'CD1_2')
      wcsDual%y%matrixOnAxis  = real64Attribute(generic, 'CD2_2')

    else
      xPixelSizeWorld = real64Attribute(generic, 'CDELT1')
      yPixelSizeWorld = real64Attribute(generic, 'CDELT2')
      if (hasAttribute(generic, "CROTA2")) then
        rotAngle = real64Attribute(generic, 'CROTA2')
        wcsDual%crotFound = .true.
      else
        rotAngle = 0.0d0
      end if

      call convertOldToNew(xPixelSizeWorld, yPixelSizeWorld, rotAngle, wcsDual)
    end if

    if (hasAttribute(generic, "CUNIT1")) then
      wcsDual%x%worldUnit = stringAttribute(generic, 'CUNIT1')
    else
      call getDefaultXWorldUnit(wcsDual)
    end if

    if (hasAttribute(generic, "CUNIT2")) then
      wcsDual%y%worldUnit = stringAttribute(generic, 'CUNIT2')
    else
      call getDefaultYWorldUnit(wcsDual)
    end if

  end subroutine readWcsDualGeneric

  !---------------------------------------------------------------------
  subroutine readWcsDualSourceListSet(set, wcsDual)
  !---------------------------------------------------------------------
!*** These kwds should surely be stored as proper column kwds?

    implicit none

    type(DataSetT), intent(in)  :: set
    type(WcsDualT),     intent(out) :: wcsDual

    real(double) :: rotAngle, xPixelSizeWorld, yPixelSizeWorld

    !-------------------------------------------------------------------
    ! Start:

    wcsDual%x%ctype     = stringAttribute(set, 'TCTYPE1')
    wcsDual%y%ctype     = stringAttribute(set, 'TCTYPE2')

!    call checkWcsDualType(wcsDual)

    wcsDual%x%refPixels = real64Attribute(set, 'TCRPIX1')
    wcsDual%y%refPixels = real64Attribute(set, 'TCRPIX2')
    wcsDual%x%refWorld  = real64Attribute(set, 'TCRVAL1')
    wcsDual%y%refWorld  = real64Attribute(set, 'TCRVAL2')

    wcsDual%crotFound = .false. ! default
    if (hasAttribute(set, "TCD1_1")) then
      wcsDual%x%matrixOnAxis  = real64Attribute(set, 'TCD1_1')
      wcsDual%x%matrixOffAxis = real64Attribute(set, 'TCD2_1')
!      wcsDual%y%matrixOnAxis  = real64Attribute(set, 'TCD1_2')
!      wcsDual%y%matrixOffAxis = real64Attribute(set, 'TCD2_2')
wcsDual%y%matrixOffAxis = real64Attribute(set, 'TCD1_2')
wcsDual%y%matrixOnAxis  = real64Attribute(set, 'TCD2_2')

    else
      xPixelSizeWorld = real64Attribute(set, 'TCDELT1')
      yPixelSizeWorld = real64Attribute(set, 'TCDELT2')
      if (hasAttribute(set, "TCROTA2")) then
        rotAngle = real64Attribute(set, 'TCROTA2')
        wcsDual%crotFound = .true.
      else
        rotAngle = 0.0d0
      end if

      call convertOldToNew(xPixelSizeWorld, yPixelSizeWorld, rotAngle, wcsDual)
    end if

    if (hasAttribute(set, "TCUNI1")) then
      wcsDual%x%worldUnit = stringAttribute(set, 'TCUNI1')
    else
      call getDefaultXWorldUnit(wcsDual)
    end if

    if (hasAttribute(set, "TCUNI2")) then
      wcsDual%y%worldUnit = stringAttribute(set, 'TCUNI2')
    else
      call getDefaultYWorldUnit(wcsDual)
    end if

  end subroutine readWcsDualSourceListSet

  !---------------------------------------------------------------------
  subroutine readWcsDualSourceListName(setName, wcsDual)
  !---------------------------------------------------------------------
!*** These kwds should surely be stored as proper column kwds?

    implicit none

    character(*), intent(in) :: setName
    type(WcsDualT), intent(out) :: wcsDual

    type(DataSetT) :: set

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(setName, READ)
    call readWcsDualSourceListSet(set, wcsDual)
    call release(set)

  end subroutine readWcsDualSourceListName

  !---------------------------------------------------------------------
  subroutine readWcsColumnName(tab, colName, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in)  :: tab
    character(*), intent(in)  :: colName
    type(WcsT),   intent(out) :: wcs

    type(ColumnT) :: col

    !-------------------------------------------------------------------
    ! Start:

    col = column(tab, colName)
    call readWcsColumn(col, wcs)
    call release(col)

  end subroutine readWcsColumnName

  !---------------------------------------------------------------------
  subroutine readWcsColumn(col, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(ColumnT), intent(in) :: col
    type(WcsT),   intent(out) :: wcs

!    real(double) :: rotAngle, xPixelSizeWorld, yPixelSizeWorld

    !-------------------------------------------------------------------
    ! Start:

    wcs%ctype     = stringAttribute(col, 'TCTYP')
    wcs%refPixels = real64Attribute(col, 'TCRPX')
    wcs%refWorld  = real64Attribute(col, 'TCRVL')

    wcs%styleIsNew = .true.
    wcs%crotFound = .false.
    if (hasAttribute(col, "TCD1_")) then
      wcs%matrixOnAxis  = real64Attribute(col, 'TCD1_')
      wcs%matrixOffAxis = real64Attribute(col, 'TCD2_')

    else
      wcs%styleIsNew = .false.
      wcs%pixelSizeWorld = real64Attribute(col, 'TCDLT')
      if (hasAttribute(col, "TCROT")) then
        wcs%rotAngle = real64Attribute(col, 'TCROT')
        wcs%crotFound = .true.

      else
        wcs%rotAngle = 0.0d0

!! See comments in writeWcsColumns.
!if (hasAttribute(parent(xCol), 'TMPTCROT')) then
!  call warning('tcrotSubstituteFound',"ssclib:readWcsColumns: The DAL &
!    &interface does not yet enable one to write the WCS roll angle keyword &
!    &as a column attribute; until this is fixed, I'm writing this value to &
!    &the TMPTCROT keyword in the table. Such a keyword has been detected and &
!    &the roll angle will be read from it.")
!  rotAngle = real64Attribute(parent(xCol), 'TMPTCROT') * PiOn180
!end if
      end if

!      call convertOldToNew(xPixelSizeWorld, yPixelSizeWorld, rotAngle, wcs)
    end if

    if (hasAttribute(col, "TCUNI")) then
      wcs%worldUnit = stringAttribute(col, 'TCUNI')
    end if

  end subroutine readWcsColumn

  !---------------------------------------------------------------------
  subroutine readWcsArray(inArray, i, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in)  :: inArray
    integer,      intent(in)  :: i
    type(WcsT),   intent(out) :: wcs

    type(AttributableT) :: generic

    !-------------------------------------------------------------------
    ! Start:

    generic = attributable(inArray)
    call readWcsGeneric(generic, i, wcs)

  end subroutine readWcsArray

  !---------------------------------------------------------------------
  subroutine readWcsSet(set, i, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in)  :: set
    integer,        intent(in)  :: i
    type(WcsT),     intent(out) :: wcs

    type(AttributableT) :: generic

    !-------------------------------------------------------------------
    ! Start:

    generic = attributable(set)
    call readWcsGeneric(generic, i, wcs)

  end subroutine readWcsSet

  !---------------------------------------------------------------------
  subroutine readWcsTable(tab, i, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in)  :: tab
    integer,      intent(in)  :: i
    type(WcsT),   intent(out) :: wcs

    type(AttributableT) :: generic

    !-------------------------------------------------------------------
    ! Start:

    generic = attributable(tab)
    call readWcsGeneric(generic, i, wcs)

  end subroutine readWcsTable

  !---------------------------------------------------------------------
  subroutine readWcsGeneric(generic, i, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(AttributableT), intent(in)  :: generic
    integer,             intent(in)  :: i
    type(WcsT),          intent(out) :: wcs

!    real(double) :: rotAngle, xPixelSizeWorld, yPixelSizeWorld
    character(6) :: ctypeKwd, crpixKwd, crvalKwd, cdeltKwd, crotaKwd, cunitKwd
    character(5) :: cd1Kwd, cd2Kwd

    !-------------------------------------------------------------------
    ! Start:

if (i>9) call error('',"")

    write(ctypeKwd, '(a5, i1)') 'CTYPE', i
    write(crpixKwd, '(a5, i1)') 'CRPIX', i
    write(crvalKwd, '(a5, i1)') 'CRVAL', i
    write(cd1Kwd,   '(a4, i1)') 'CD1_', i
    write(cd2Kwd,   '(a4, i1)') 'CD2_', i
    write(cdeltKwd, '(a5, i1)') 'CDELT', i
    write(crotaKwd, '(a5, i1)') 'CROTA', i
    write(cunitKwd, '(a5, i1)') 'CUNIT', i

    wcs%ctype     = stringAttribute(generic, ctypeKwd)
    wcs%refPixels = real64Attribute(generic, crpixKwd)
    wcs%refWorld  = real64Attribute(generic, crvalKwd)

    wcs%styleIsNew = .true.
    wcs%crotFound = .false.
    if (hasAttribute(generic, cd1Kwd)) then
      wcs%matrixOnAxis  = real64Attribute(generic, cd1Kwd)
      wcs%matrixOffAxis = real64Attribute(generic, cd2Kwd)

    else
      wcs%styleIsNew = .false.
      wcs%pixelSizeWorld = real64Attribute(generic, cdeltKwd)
      if (hasAttribute(generic, crotaKwd)) then
        wcs%rotAngle = real64Attribute(generic, crotaKwd)
        wcs%crotFound = .true.

      else
        wcs%rotAngle = 0.0d0

!! See comments in writeWcsColumns.
!if (hasAttribute(parent(xCol), 'TMPTCROT')) then
!  call warning('tcrotSubstituteFound',"ssclib:readWcsColumns: The DAL &
!    &interface does not yet enable one to write the WCS roll angle keyword &
!    &as a column attribute; until this is fixed, I'm writing this value to &
!    &the TMPTCROT keyword in the table. Such a keyword has been detected and &
!    &the roll angle will be read from it.")
!  rotAngle = real64Attribute(parent(xCol), 'TMPTCROT') * PiOn180
!end if
      end if

!      call convertOldToNew(xPixelSizeWorld, yPixelSizeWorld, rotAngle, wcs)
    end if

    if (hasAttribute(generic, cunitKwd)) then
      wcs%worldUnit = stringAttribute(generic, cunitKwd)
    end if

  end subroutine readWcsGeneric


  !---------------------------------------------------------------------
  subroutine writeWcsDualColumns(xCol, yCol, wcsDual)
  !---------------------------------------------------------------------

    implicit none

    type(ColumnT), intent(in) :: xCol, yCol
    type(WcsDualT),    intent(in) :: wcsDual

    real(double) :: xPixelSizeWorld, yPixelSizeWorld, rotAngle

    !-------------------------------------------------------------------
    ! Start:

    call setAttribute(xCol, "TCTYP", wcsDual%X%ctype)
    call setAttribute(xCol, "TCRPX", wcsDual%X%refPixels)
    call setAttribute(xCol, "TCRVL", wcsDual%X%refWorld)
    call setAttribute(xCol, 'TCUNI', wcsDual%X%worldUnit)

    call setAttribute(yCol, "TCTYP", wcsDual%Y%ctype)
    call setAttribute(yCol, "TCRPX", wcsDual%Y%refPixels)
    call setAttribute(yCol, "TCRVL", wcsDual%Y%refWorld)
    call setAttribute(yCol, 'TCUNI', wcsDual%Y%worldUnit)

    if (useNewWcsStyle) then
      call setAttribute(xCol, 'TCD1_',  wcsDual%X%matrixOnAxis)
      call setAttribute(xCol, 'TCD2_',  wcsDual%X%matrixOffAxis)

!      call setAttribute(yCol, 'TCD1_',  wcsDual%Y%matrixOnAxis)
!      call setAttribute(yCol, 'TCD2_',  wcsDual%Y%matrixOffAxis)
call setAttribute(yCol, 'TCD1_',  wcsDual%Y%matrixOffAxis)
call setAttribute(yCol, 'TCD2_',  wcsDual%Y%matrixOnAxis)

    else ! use the old style.
      call convertNewToOld(wcsDual, xPixelSizeWorld, yPixelSizeWorld, rotAngle)

      call setAttribute(xCol, "TCDLT", xPixelSizeWorld)
      call setAttribute(yCol, "TCDLT", yPixelSizeWorld)
!!!      call setAttribute(yCol, "TCROT", rotAngle)

!*** apparently the dal doesn't like the TCROT keyword. I've removed it for the time being to get rid of the dal warning which always used to occur when writeWcsColumns was invoked. However I've SPRed the dal (SOC SPR 2523) and if this gets accepted you can decomment the setAttribute line and get rid of this workaround.

if (rotAngle .ne. 0.0) then
  call warning('unableToWriteTcrot',"ssclib:writeWcsDualColumns: The DAL &
    &interface does not yet enable one to write the WCS roll angle keyword &
    &as a column attribute. I'm writing this value to the TMPTCROT keyword &
    &in the table.")

  call setAttribute(parent(xCol), 'TMPTCROT', rotAngle/PiOn180, 'deg'&
    , 'Temporary storage of column-wcs roll angle.')

end if
    end if

  end subroutine writeWcsDualColumns

  !---------------------------------------------------------------------
  subroutine writeWcsDualArray(inArray, wcsDual)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in) :: inArray
    type(WcsDualT),   intent(in) :: wcsDual

    real(double) :: xPixelSizeWorld, yPixelSizeWorld, rotAngle

    !-------------------------------------------------------------------
    ! Start:

    call setAttribute(inArray, 'CRPIX1', wcsDual%X%refPixels)
    call setAttribute(inArray, 'CRPIX2', wcsDual%Y%refPixels)
    call setAttribute(inArray, 'CRVAL1', wcsDual%X%refWorld)
    call setAttribute(inArray, 'CRVAL2', wcsDual%Y%refWorld)
    call setAttribute(inArray, 'CTYPE1', wcsDual%X%ctype)
    call setAttribute(inArray, 'CTYPE2', wcsDual%Y%ctype)

    call setAttribute(inArray, 'CUNIT1', wcsDual%X%worldUnit)
    call setAttribute(inArray, 'CUNIT2', wcsDual%Y%worldUnit)

    if (useNewWcsStyle) then
      call setAttribute(inArray, 'CD1_1',  wcsDual%X%matrixOnAxis)
      call setAttribute(inArray, 'CD2_1',  wcsDual%X%matrixOffAxis)
!      call setAttribute(inArray, 'CD1_2',  wcsDual%Y%matrixOnAxis)
!      call setAttribute(inArray, 'CD2_2',  wcsDual%Y%matrixOffAxis)
call setAttribute(inArray, 'CD1_2',  wcsDual%Y%matrixOffAxis)
call setAttribute(inArray, 'CD2_2',  wcsDual%Y%matrixOnAxis)

    else ! use the old style.
      call convertNewToOld(wcsDual, xPixelSizeWorld, yPixelSizeWorld, rotAngle)

      call setAttribute(inArray, 'CROTA2', rotAngle)
      call setAttribute(inArray, 'CDELT1', xPixelSizeWorld)
      call setAttribute(inArray, 'CDELT2', yPixelSizeWorld)
    end if

  end subroutine writeWcsDualArray

  !---------------------------------------------------------------------
  subroutine writeWcsColumn(col, wcs)
  !---------------------------------------------------------------------

    implicit none

    type(ColumnT), intent(in) :: col
    type(WcsT),    intent(in) :: wcs

!    real(double) :: xPixelSizeWorld, yPixelSizeWorld, rotAngle
!    character(6) :: ctypeKwd, crpixKwd, crvalKwd, cdeltKwd, crotaKwd, cunitKwd
!    character(5) :: cd1Kwd, cd2Kwd

    !-------------------------------------------------------------------
    ! Start:

    call setAttribute(col, "TCTYP", wcs%ctype)
    call setAttribute(col, "TCRPX", wcs%refPixels)
    call setAttribute(col, "TCRVL", wcs%refWorld)
    call setAttribute(col, 'TCUNI', wcs%worldUnit)

    if (useNewWcsStyle) then
      call setAttribute(col, 'TCD1_',  wcs%matrixOnAxis)
      call setAttribute(col, 'TCD2_',  wcs%matrixOffAxis)
!******** but for y axis 2 is on-axis and 1 off-axis! Got to unsnarl this can of worms some way or another

    else ! use the old style.
!      call convertNewToOld(wcs, xPixelSizeWorld, yPixelSizeWorld, rotAngle)

      call setAttribute(col, "TCDLT", wcs%pixelSizeWorld)
!!!      call setAttribute(col, "TCROT", wcs%rotAngle/PiOn180)

!*** apparently the dal doesn't like the TCROT keyword. I've removed it for the time being to get rid of the dal warning which always used to occur when writeWcsColumns was invoked. However I've SPRed the dal (SOC SPR 2523) and if this gets accepted you can decomment the setAttribute line and get rid of this workaround.

if (wcs%rotAngle .ne. 0.0) then
  call warning('unableToWriteTcrot',"ssclib:writeWcsColumn: The DAL &
    &interface does not yet enable one to write the WCS roll angle keyword &
    &as a column attribute. I'm writing this value to the TMPTCROT keyword &
    &in the table.")

  call setAttribute(parent(col), 'TMPTCROT', wcs%rotAngle/PiOn180, 'deg'&
    , 'Temporary storage of column-wcs roll angle.')

end if
    end if

  end subroutine writeWcsColumn

  !---------------------------------------------------------------------
  subroutine writeWcsArray(inArray, wcs, i, force)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in)           :: inArray
    type(WcsT),   intent(in)           :: wcs
    integer,      intent(in)           :: i
    logical,      intent(in), optional :: force

!    real(double) :: xPixelSizeWorld, yPixelSizeWorld, rotAngle
    character(6) :: ctypeKwd, crpixKwd, crvalKwd, cdeltKwd, crotaKwd, cunitKwd
    character(5) :: cd1Kwd, cd2Kwd
    logical :: localForce

    !-------------------------------------------------------------------
    ! Start:

if (i>9) call error('',"")

    if (present(force)) then
      localForce = force
    else
      localForce = .false.
    end if

    if (hasWcsArray(inArray, i)) then
      if (localForce) then
        call warning('wcsAlreadyExists',"A set of WCS keywords of this &
          &number already exists in the array. They have been overwritten.")
      else
        call warning('wcsAlreadyExists',"A set of WCS keywords of this &
          &number already exists in the array. They have not been overwritten.")
  return
      end if
    end if

    write(ctypeKwd, '(a5, i1)') 'CTYPE', i
    write(crpixKwd, '(a5, i1)') 'CRPIX', i
    write(crvalKwd, '(a5, i1)') 'CRVAL', i
    write(cd1Kwd,   '(a4, i1)') 'CD1_', i
    write(cd2Kwd,   '(a4, i1)') 'CD2_', i
    write(cdeltKwd, '(a5, i1)') 'CDELT', i
    write(crotaKwd, '(a5, i1)') 'CROTA', i
    write(cunitKwd, '(a5, i1)') 'CUNIT', i

    call setAttribute(inArray, crpixKwd, wcs%refPixels)
    call setAttribute(inArray, crvalKwd, wcs%refWorld)
    call setAttribute(inArray, ctypeKwd, wcs%ctype)
    call setAttribute(inArray, cunitKwd, wcs%worldUnit)

    if (useNewWcsStyle) then
      call setAttribute(inArray, cd1Kwd,  wcs%matrixOnAxis)
      call setAttribute(inArray, cd2Kwd,  wcs%matrixOffAxis)
!******** but for y axis 2 is on-axis and 1 off-axis! Got to unsnarl this can of worms some way or another

    else ! use the old style.
!      call convertNewToOld(wcs, xPixelSizeWorld, yPixelSizeWorld, rotAngle)

      call setAttribute(inArray, crotaKwd, wcs%rotAngle/PiOn180)
      call setAttribute(inArray, cdeltKwd, wcs%pixelSizeWorld)
    end if

  end subroutine writeWcsArray

  !---------------------------------------------------------------------
  function writeNextWcsArray(inArray, wcs) result(i)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in) :: inArray
    type(WcsT),   intent(in) :: wcs
    integer                  :: i

    integer :: j

    !-------------------------------------------------------------------
    ! Start:

    i = 10
    do j = 1, 9
      if (.not.hasWcsArray(inArray, j)) then
        i = j
    exit
      end if
    end do

if (i>9) call error('',"")

    call writeWcsArray(inArray, wcs, i)

  end function writeNextWcsArray


  !---------------------------------------------------------------------
  subroutine deleteWcs(set)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in) :: set

    integer, parameter :: numRootStrs = 6, maxNumDims = 3
    character(5) :: rootStrs(numRootStrs)
    integer :: i, j
    character(6) :: str1
    character(5) :: str2

    !-------------------------------------------------------------------
    ! Start

    rootStrs = (/'CRPIX', 'CRVAL', 'CTYPE', 'CDELT', 'CROTA', 'CUNIT'/)

    do i = 1, maxNumDims
      do j = 1, numRootStrs
        write(str1, '(a5, i1)') rootStrs(j), i
        if (hasAttribute(set, str1)) call deleteAttribute(set, str1)
      end do
      do j = 1, maxNumDims
        write(str2, '(a2, i1, a1, i1)') 'CD', i, '_', j
        if (hasAttribute(set, str2)) call deleteAttribute(set, str2)
        write(str2, '(a2, i1, a1, i1)') 'PC', i, '_', j
        if (hasAttribute(set, str2)) call deleteAttribute(set, str2)
      end do
    end do

  end subroutine deleteWcs

  !---------------------------------------------------------------------
  subroutine easifyWcsTypeScalar(ctypeX, ctypeY, wcsType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: ctypeX, ctypeY
    character(8), intent(out) :: wcsType

    !-------------------------------------------------------------------
    ! Start:

    if (  trim(ctypeX) .eq. 'DETX' .and. trim(ctypeY) .eq. 'DETY') then
      wcsType = 'DETXY'

    else&
    if (trim(ctypeX).eq.'RA---TAN' .and. trim(ctypeY) .eq. 'DEC--TAN') then
      wcsType = 'XY'
    end if

  end subroutine easifyWcsTypeScalar

  !---------------------------------------------------------------------
  subroutine easifyWcsTypeVector(ctype, wcsType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: ctype(2)
    character(8), intent(out) :: wcsType

    !-------------------------------------------------------------------
    ! Start:

    if (  trim(ctype(1)) .eq. 'DETX' .and. trim(ctype(2)) .eq. 'DETY') then
      wcsType = 'DETXY'

    else&
    if (trim(ctype(1)).eq.'RA---TAN' .and. trim(ctype(2)) .eq. 'DEC--TAN') then
      wcsType = 'XY'
    end if

  end subroutine easifyWcsTypeVector

  !---------------------------------------------------------------------
  subroutine dumpWcsDual(wcsDual, verbosityArg)
  !---------------------------------------------------------------------

    implicit none

    type(WcsDualT), intent(in) :: wcsDual
    integer(int32), intent(in), optional :: verbosityArg

    integer(int32) :: verbosity

    !-------------------------------------------------------------------
    ! Start

    if (present(verbosityArg)) then
      verbosity = verbosityArg
    else
      verbosity = VERBOSE
    end if

    call message(AppMsg, verbosity, 'WCS dump:')
    call message(AppMsg, verbosity, '  X axis:')
    call dumpWcsSingle(wcsDual%x, verbosity)
    call message(AppMsg, verbosity, '  Y axis:')
    call dumpWcsSingle(wcsDual%y, verbosity)

  end subroutine dumpWcsDual

  !---------------------------------------------------------------------
  subroutine dumpWcsVec(wcsVec, verbosityArg)
  !---------------------------------------------------------------------

    implicit none

    type(WcsT), intent(in) :: wcsVec(:)
    integer(int32), intent(in), optional :: verbosityArg

    integer(int32) :: verbosity, numWcs, i

    !-------------------------------------------------------------------
    ! Start

    if (present(verbosityArg)) then
      verbosity = verbosityArg
    else
      verbosity = VERBOSE
    end if

    numWcs = size(wcsVec)
if (numWcs > 9) call error('',"")
    call message(AppMsg, verbosity, 'WCS dump:')
    do i = 1, numWcs
      write(errStr, '(a7, i1)') '  Axis ', i
      call message(AppMsg, verbosity)
      call dumpWcsSingle(wcsVec(i), verbosity)
    end do

  end subroutine dumpWcsVec

  !---------------------------------------------------------------------
  subroutine dumpWcsSingle(wcs, verbosityArg)
  !---------------------------------------------------------------------

    implicit none

    type(WcsT),     intent(in) :: wcs
    integer(int32), intent(in), optional :: verbosityArg

    integer(int32) :: verbosity

    !-------------------------------------------------------------------
    ! Start

    if (present(verbosityArg)) then
      verbosity = verbosityArg
    else
      verbosity = VERBOSE
    end if

    call message(AppMsg, verbosity, '    CTYPE: '//wcs%ctype)
    call message(AppMsg, verbosity, '    CUNIT: '//wcs%worldUnit)
    write(errstr, '(a, es23.16)')   '    CRVAL:', wcs%refWorld
    call message(AppMsg, verbosity)
    write(errstr, '(a, es23.16)')   '    CRPIX:', wcs%refPixels
    call message(AppMsg, verbosity)
    write(errstr, '(a, es23.16)')   '    CD1_1:', wcs%matrixOnAxis
    call message(AppMsg, verbosity)
    write(errstr, '(a, es23.16)')   '    CD2_1:', wcs%matrixOffAxis
    call message(AppMsg, verbosity)
    write(errstr, '(a, es23.16)')   '    CDELT:', wcs%pixelSizeWorld
    call message(AppMsg, verbosity)

  end subroutine dumpWcsSingle


  !---------------------------------------------------------------------
  subroutine pixelsToTanScalar(xPixels, yPixels, wcsDual, tanX, tanY)
  !---------------------------------------------------------------------
  ! See pixelsToTanVector for a description.

    implicit none

    real(single), intent(in)  :: xPixels, yPixels
    type(WcsDualT),   intent(in)  :: wcsDual
    real(double), intent(out) :: tanX, tanY

    real(single) :: xPixelsVec(1), yPixelsVec(1)
    real(double) :: tanXVec(1), tanYVec(1)

    !-------------------------------------------------------------------
    ! Start

!    tanX = (xPixels - wcs%x%refPixels) * abs(wcs%x%pixelSizeWorld) * PiOn180
!    tanY = (yPixels - wcs%y%refPixels) * abs(wcs%y%pixelSizeWorld) * PiOn180

    xPixelsVec(1) = xPixels
    yPixelsVec(1) = yPixels

    call pixelsToTanVector(xPixelsVec, yPixelsVec, wcsDual, tanXVec, tanYVec)

    tanX = tanXVec(1)
    tanY = tanYVec(1)

  end subroutine pixelsToTanScalar

  !---------------------------------------------------------------------
  subroutine pixelsToTanVector(xPixels, yPixels, wcsDual, tanX, tanY)
  !---------------------------------------------------------------------
  ! This transform is only valid for CTYPE1='RA---TAN' and CTYPE2='DEC--TAN'.
  !
  ! The tan system coordinates are aligned with the ra/dec world coordinate
  ! system, but are expressed in natural units rather than in whatever units
  ! are specified for the world coordinates. In addition, the tanX axis
  ! points to the right (in the direction of decreasing ra) and the tanY
  ! axis points upwards (in the direction of increasing dec) when the view
  ! is from the centre of the projection sphere outwards.
  !
  ! Numerical examples:
  !
  !	wcs%x%ctype = 'RA---TAN'
  !	wcs%y%ctype = 'DEC--TAN'
  !	wcs%x%refWorld = 25 degrees ra
  !	wcs%y%refWorld = 0 degrees dec
  !
  !	In this case, a position with ra = 30 deg and dec = 0 would give
  !	tanX = -tan(30-25) = -0.0875, tanY = 0; whereas a position with
  !	ra = 0 deg and dec = 7 would give tanX = 0, tanY = tan(7) = 0.123.
  !
  ! As described in Greisen & Calabretta A&A 395, 1061 (2002) equation (3),
  ! the relation between pixel coordinates p_i and intermediate world
  ! coordinates x_i is
  !
  !	      _ N
  !	x_i = >_  CDi_j (p_j - CRPIXj).
  !	     j = 1
  !
  ! This is implemented in the present subroutine, then the result x_i
  ! is converted to my tan system.
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)  :: xPixels(:), yPixels(size(xPixels))
    type(WcsDualT),   intent(in)  :: wcsDual
    real(double), intent(out) :: tanX(size(xPixels)), tanY(size(xPixels))

    real(double) ::&
      intermediateWorldXToTanX, intermediateWorldYToTanY
    real(double), allocatable ::&
      intermediateWorldX(:), intermediateWorldY(:)
    real(single), allocatable :: relPixelsX(:), relPixelsY(:)
    character(3) :: xProjTypeStr, yProjTypeStr

    !-------------------------------------------------------------------
    ! Start

    allocate(intermediateWorldX(size(xPixels)))
    allocate(intermediateWorldY(size(xPixels)))
    allocate(relPixelsX(size(xPixels)), relPixelsY(size(xPixels)))

!    call checkWcsType(wcs)
    xProjTypeStr = getProjectionType(wcsDual%x)
    yProjTypeStr = getProjectionType(wcsDual%y)
    if (xProjTypeStr .ne. 'TAN' .and. yProjTypeStr .ne. 'TAN')&
call error('projectionNotYetSupported',"ssclib:pixelsToTanVector: Non-TAN &
&projections are not yet supported.")

    select case(wcsDual%x%worldUnit)
    case('deg')
      intermediateWorldXToTanX = -PiOn180 ! note the negation
    case default
call error('unitNotYetSupported',"ssclib:pixelsToTanVector: CUNIT1 values other &
&than 'deg' are not yet supported.")
    end select

    select case(wcsDual%y%worldUnit)
    case('deg')
      intermediateWorldYToTanY = PiOn180
    case default
call error('unitNotYetSupported',"ssclib:pixelsToTanVector: CUNIT2 values other &
&than 'deg' are not yet supported.")
    end select

    relPixelsX = xPixels - wcsDual%x%refPixels
    relPixelsY = yPixels - wcsDual%y%refPixels

    intermediateWorldX = wcsDual%x%matrixOnAxis  * relPixelsX&
                       + wcsDual%y%matrixOffAxis * relPixelsY
    intermediateWorldY = wcsDual%y%matrixOnAxis  * relPixelsY&
                       + wcsDual%x%matrixOffAxis * relPixelsX

    tanX = intermediateWorldX * intermediateWorldXToTanX
    tanY = intermediateWorldY * intermediateWorldYToTanY

  end subroutine pixelsToTanVector

  !---------------------------------------------------------------------
  subroutine pixelsToTanArray(xPixels, yPixels, wcsDual, tanX, tanY)
  !---------------------------------------------------------------------
  ! See tanToPixelsVector for a description.

    implicit none

    real(single), intent(in)  :: xPixels(:,:),&
                                 yPixels(size(xPixels,1),size(xPixels,2))
    type(WcsDualT),   intent(in)  :: wcsDual
    real(double), intent(out) :: tanX(size(xPixels,1),size(xPixels,2)),&
                                 tanY(size(xPixels,1),size(xPixels,2))

    real(single), allocatable :: xPixelsVec(:), yPixelsVec(:)
    real(double), allocatable :: tanXVec(:), tanYVec(:)
    integer(int32) :: xSize, ySize, vSize

    !-------------------------------------------------------------------
    ! Start

    xSize = size(xPixels, 1)
    ySize = size(xPixels, 2)
    vSize = xSize * ySize
    allocate(xPixelsVec(vSize), yPixelsVec(vSize))
    allocate(tanXVec(vSize), tanYVec(vSize))

    xPixelsVec = reshape(xPixels, (/vSize/))
    yPixelsVec = reshape(yPixels, (/vSize/))

    call pixelsToTanVector(xPixelsVec, yPixelsVec, wcsDual, tanXVec, tanYVec)

    tanX = reshape(tanXVec, (/xSize, ySize/))
    tanY = reshape(tanYVec, (/xSize, ySize/))

  end subroutine pixelsToTanArray

  !---------------------------------------------------------------------
  subroutine tanToPixelsScalar(wcsDual, tanX, tanY, xPixels, yPixels)
  !---------------------------------------------------------------------
  ! See tanToPixelsVector for a description.

    implicit none

    real(double), intent(in)  :: tanX, tanY
    type(WcsDualT),   intent(in)  :: wcsDual
    real(single), intent(out) :: xPixels, yPixels

    real(single) :: xPixelsVec(1), yPixelsVec(1)
    real(double) :: tanXVec(1), tanYVec(1)

    !-------------------------------------------------------------------
    ! Start

    tanXVec(1) = tanX
    tanYVec(1) = tanY

    call tanToPixelsVector(wcsDual, tanXVec, tanYVec, xPixelsVec, yPixelsVec)

    xPixels = xPixelsVec(1)
    yPixels = yPixelsVec(1)

  end subroutine tanToPixelsScalar

  !---------------------------------------------------------------------
  subroutine tanToPixelsVector(wcsDual, tanX, tanY, xPixels, yPixels)
  !---------------------------------------------------------------------
  ! This transform is only valid for CTYPE1='RA---TAN' and CTYPE2='DEC--TAN'.
  !
  ! The tan system coordinates are aligned with the ra/dec world coordinate
  ! system, but are expressed in natural units rather than in whatever units
  ! are specified for the world coordinates. In addition, the tanX axis
  ! points to the right (in the direction of decreasing ra) and the tanY
  ! axis points upwards (in the direction of increasing dec) when the view
  ! is from the centre of the projection sphere outwards.
  !
  ! Numerical examples:
  !
  !	wcs%x%ctype = 'RA---TAN'
  !	wcs%y%ctype = 'DEC--TAN'
  !	wcs%x%refWorld = 25 degrees ra
  !	wcs%y%refWorld = 0 degrees dec
  !
  !	In this case, a position with ra = 30 deg and dec = 0 would give
  !	tanX = -tan(30-25) = -0.0875, tanY = 0; whereas a position with
  !	ra = 0 deg and dec = 7 would give tanX = 0, tanY = tan(7) = 0.123.
  !
  ! As described in Greisen & Calabretta A&A 395, 1061 (2002) equation (3),
  ! the relation between pixel coordinates p_i and intermediate world
  ! coordinates x_i is
  !
  !	      _ N
  !	x_i = >_  CDi_j (p_j - CRPIXj).
  !	     j = 1
  !
  ! This is inverted in the present subroutine, after x_i is first converted
  ! from my tan system.
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: tanX(:), tanY(size(tanX))
    type(WcsDualT),   intent(in)  :: wcsDual
    real(single), intent(out) :: xPixels(size(tanX)), yPixels(size(tanX))

    real(double), allocatable ::&
      intermediateWorldX(:), intermediateWorldY(:)
    real(double) ::&
      intermediateWorldXToTanX, intermediateWorldYToTanY,&
      det, invMatrix(2,2)
    real(single), allocatable :: relPixelsX(:), relPixelsY(:)

    ! The above changes of sized arrays to allocatable was done in order
    ! to work around a compiler bug. If local arrays are declared to the
    ! size of an input array, this seems to work fine 99% of the time,
    ! but once in a while results in a seg fault at the instant the
    ! subroutine is called. On these occasions no executable statements
    ! of the subroutine are executed before the seg fault. Previously I
    ! have experienced this trouble just with user-defined structure
    ! arrays, now I get it with simple real kind arrays. IMS 2004/04/27.

    character(3) :: xProjTypeStr, yProjTypeStr
    logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '      aaa tanToPixelsVector'

    allocate(intermediateWorldX(size(tanX)), intermediateWorldY(size(tanX)))
    allocate(relPixelsX(size(tanX)), relPixelsY(size(tanX)))

!    call checkWcsType(wcs)
    xProjTypeStr = getProjectionType(wcsDual%x)
    yProjTypeStr = getProjectionType(wcsDual%y)
    if (xProjTypeStr .ne. 'TAN' .and. yProjTypeStr .ne. 'TAN')&
call error('projectionNotYetSupported',"ssclib:tanToPixelsVector: Non-TAN &
&projections are not yet supported.")

if (testFlag) print *, '      bbb'

    select case(wcsDual%x%worldUnit)
    case('deg')
      intermediateWorldXToTanX = -PiOn180 ! note the negation
    case default
call error('unitNotYetSupported',"ssclib:pixelsToTanVector: CUNIT1 values other &
&than 'deg' are not yet supported.")
    end select
if (testFlag) print *, '      ccc'

    select case(wcsDual%y%worldUnit)
    case('deg')
      intermediateWorldYToTanY = PiOn180
    case default
call error('unitNotYetSupported',"ssclib:pixelsToTanVector: CUNIT2 values other &
&than 'deg' are not yet supported.")
    end select
if (testFlag) print *, '      ddd'

    intermediateWorldX = tanX / intermediateWorldXToTanX
    intermediateWorldY = tanY / intermediateWorldYToTanY
if (testFlag) print *, '      eee'

    det = wcsDual%x%matrixOnAxis  * wcsDual%y%matrixOnAxis&
        - wcsDual%x%matrixOffAxis * wcsDual%y%matrixOffAxis
    if (det == 0.0)&
call error('singularWcsRotMatrix',"ssclib:pixelsToTanVector: Can't invert the WCS rotation matrix")
if (testFlag) print *, '      fff'

    invMatrix(1,1) =  wcsDual%y%matrixOnAxis  / det		! 1st index row,
    invMatrix(1,2) = -wcsDual%y%matrixOffAxis / det		! 2nd index col.
    invMatrix(2,1) = -wcsDual%x%matrixOffAxis / det
    invMatrix(2,2) =  wcsDual%x%matrixOnAxis  / det
if (testFlag) print *, '      ggg'

    relPixelsX = invMatrix(1,1) * intermediateWorldX&
               + invMatrix(1,2) * intermediateWorldY
    relPixelsY = invMatrix(2,1) * intermediateWorldX&
               + invMatrix(2,2) * intermediateWorldY
if (testFlag) print *, '      hhh'

    xPixels = relPixelsX + wcsDual%x%refPixels
    yPixels = relPixelsY + wcsDual%y%refPixels
if (testFlag) print *, '      zzz tanToPixelsVector'

  end subroutine tanToPixelsVector

  !---------------------------------------------------------------------
  subroutine tanToPixelsArray(wcsDual, tanX, tanY, xPixels, yPixels)
  !---------------------------------------------------------------------
  ! See tanToPixelsVector for a description.

    implicit none

    real(double), intent(in)  :: tanX(:,:),&
                                 tanY(size(tanX,1),size(tanX,2))
    type(WcsDualT),   intent(in)  :: wcsDual
    real(single), intent(out) :: xPixels(size(tanX,1),size(tanX,2)),&
                                 yPixels(size(tanX,1),size(tanX,2))

    real(single), allocatable :: xPixelsVec(:), yPixelsVec(:)
    real(double), allocatable :: tanXVec(:), tanYVec(:)
    integer(int32) :: xSize, ySize, vSize

    !-------------------------------------------------------------------
    ! Start

    xSize = size(tanX, 1)
    ySize = size(tanX, 2)
    vSize = xSize * ySize
    allocate(xPixelsVec(vSize), yPixelsVec(vSize))
    allocate(tanXVec(vSize), tanYVec(vSize))

    tanXVec = reshape(tanX, (/vSize/))
    tanYVec = reshape(tanY, (/vSize/))

    call tanToPixelsVector(wcsDual, tanXVec, tanYVec, xPixelsVec, yPixelsVec)

    xPixels = reshape(xPixelsVec, (/xSize, ySize/))
    yPixels = reshape(yPixelsVec, (/xSize, ySize/))

  end subroutine tanToPixelsArray

  !---------------------------------------------------------------------
  subroutine tanRadToPixEllipseScalar(radiusTan, wcsDual, shortSemiAxisPixels&
  , longSemiAxisPixels, rotatedAngle)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: radiusTan
    type(WcsDualT),   intent(in)  :: wcsDual
    real(single), intent(out) :: shortSemiAxisPixels,&
                                 longSemiAxisPixels,&
                                 rotatedAngle

    real(single) ::&
      shortSemiAxisPixelsVec(1), longSemiAxisPixelsVec(1), rotatedAngleVec(1)

    !-------------------------------------------------------------------
    ! Start

    call tanRadToPixEllipseVector((/radiusTan/), wcsDual, shortSemiAxisPixelsVec&
      , longSemiAxisPixelsVec, rotatedAngleVec)

    shortSemiAxisPixels = shortSemiAxisPixelsVec(1)
    longSemiAxisPixels  = longSemiAxisPixelsVec(1)
    rotatedAngle        = rotatedAngleVec(1)

  end subroutine tanRadToPixEllipseScalar

  !---------------------------------------------------------------------
  subroutine tanRadToPixEllipseVector(radiusTan, wcsDual, shortSemiAxisPixels&
  , longSemiAxisPixels, rotatedAngle)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: radiusTan(:)
    type(WcsDualT),   intent(in)  :: wcsDual
    real(single), intent(out) :: shortSemiAxisPixels(size(radiusTan)),&
                                 longSemiAxisPixels(size(radiusTan)),&
                                 rotatedAngle(size(radiusTan))

    character(3) :: xProjTypeStr, yProjTypeStr
    real(double) :: intermediateWorldXToTanX, xPixelSizeWorld, yPixelSizeWorld, rotAngleDble
!, intermediateWorldYToTanY
    real(double), allocatable :: radiusWorld(:)

    !-------------------------------------------------------------------
    ! Start

    xProjTypeStr = getProjectionType(wcsDual%x)
    yProjTypeStr = getProjectionType(wcsDual%y)
    if (xProjTypeStr .ne. 'TAN' .and. yProjTypeStr .ne. 'TAN')&
call error('projectionNotYetSupported',"ssclib:tanRadToPixEllipseVector: Non-TAN &
&projections are not yet supported.")

    select case(wcsDual%x%worldUnit)
    case('deg')
      intermediateWorldXToTanX = -PiOn180 ! note the negation
    case default
call error('unitNotYetSupported',"ssclib:pixelsToTanVector: CUNIT1 values other &
&than 'deg' are not yet supported.")
    end select

!    select case(wcs%y%worldUnit)
!    case('deg')
!      intermediateWorldYToTanY = PiOn180
!    case default
!call error('unitNotYetSupported',"ssclib:pixelsToTanVector: CUNIT2 values other &
!&than 'deg' are not yet supported.")
!    end select

    allocate(radiusWorld(size(radiusTan)))
!, rotAngleDble(size(radiusTan)))
!    allocate(xPixelSizeWorld(size(radiusTan)), yPixelSizeWorld(size(radiusTan)))

    radiusWorld = radiusTan / intermediateWorldXToTanX
    call convertNewToOld(wcsDual, xPixelSizeWorld, yPixelSizeWorld, rotAngleDble)
    rotatedAngle = real(rotAngleDble, single)
    shortSemiAxisPixels = radiusWorld / xPixelSizeWorld
    longSemiAxisPixels  = radiusWorld / yPixelSizeWorld

!***** need to check this routine and also fix up the tan->world bit. I wrote it in a hurry and my head is a bit dull today.

  end subroutine tanRadToPixEllipseVector

  !---------------------------------------------------------------------
  logical(bool) function wcsAreEquivalent(wcs1, wcs2)
  !---------------------------------------------------------------------

    implicit none

    type(WcsDualT), intent(in) :: wcs1, wcs2

    !-------------------------------------------------------------------
    ! Start

    wcsAreEquivalent = .false. ! default

  if (wcs1%x%ctype          .ne. wcs2%x%ctype         ) return
  if (.not. approxEqual(wcs1%x%refPixels,      wcs2%x%refPixels)     ) return
  if (.not. approxEqual(wcs1%x%refWorld,       wcs2%x%refWorld)      ) return

  if (wcs1%y%ctype          .ne. wcs2%y%ctype         ) return
  if (.not. approxEqual(wcs1%y%refPixels,      wcs2%y%refPixels)     ) return
  if (.not. approxEqual(wcs1%y%refWorld,       wcs2%y%refWorld)      ) return

!*** fix up

    wcsAreEquivalent = .true.

  end function wcsAreEquivalent

  !---------------------------------------------------------------------
  subroutine convertOldToNew(xPixelSizeWorld, yPixelSizeWorld, rotAngle, wcsDual)
  !---------------------------------------------------------------------
  ! Makes use of equation (189) from Calabretta & Greisen A&A 395 1077
  ! (2002), ie
  !
  !	(CD1_1 CD1_2)   (CDELT1*cos[CROTA2]  -CDELT2*sin[CROTA2])
  !	(           ) = (                                       ).
  !	(CD2_1 CD2_2)   (CDELT1*sin[CROTA2]   CDELT2*cos[CROTA2])
  !
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)    :: xPixelSizeWorld, yPixelSizeWorld, rotAngle
    type(WcsDualT),   intent(inout) :: wcsDual

    !-------------------------------------------------------------------
    ! Start:

    wcsDual%x%matrixOnAxis   =  xPixelSizeWorld * cos(rotAngle)
    wcsDual%x%matrixOffAxis  =  xPixelSizeWorld * sin(rotAngle)
    wcsDual%y%matrixOffAxis  = -yPixelSizeWorld * sin(rotAngle)
    wcsDual%y%matrixOnAxis   =  yPixelSizeWorld * cos(rotAngle)

  end subroutine convertOldToNew

  !---------------------------------------------------------------------
  subroutine convertNewToOld(wcsDual, xPixelSizeWorld, yPixelSizeWorld, rotAngle)
  !---------------------------------------------------------------------
  ! Makes use of equation (189) from Calabretta & Greisen A&A 395 1077
  ! (2002), ie
  !
  !	(CD1_1 CD1_2)   (CDELT1*cos[CROTA2]  -CDELT2*sin[CROTA2])
  !	(           ) = (                                       ).
  !	(CD2_1 CD2_2)   (CDELT1*sin[CROTA2]   CDELT2*cos[CROTA2])
  !
  !---------------------------------------------------------------------

    implicit none

    type(WcsDualT),   intent(in) :: wcsDual
    real(double), intent(out) :: xPixelSizeWorld, yPixelSizeWorld, rotAngle

    real(double) :: rotAngleAlt, rotAngle2

    !-------------------------------------------------------------------
    ! Start:

    xPixelSizeWorld = sqrt(wcsDual%X%matrixOnAxis * wcsDual%X%matrixOnAxis&
      + wcsDual%X%matrixOffAxis * wcsDual%X%matrixOffAxis)

    yPixelSizeWorld = sqrt(wcsDual%Y%matrixOnAxis * wcsDual%Y%matrixOnAxis&
      + wcsDual%Y%matrixOffAxis * wcsDual%Y%matrixOffAxis)

    rotAngle    = atan2(-wcsDual%Y%matrixOffAxis, wcsDual%Y%matrixOnAxis)
    rotAngleAlt = atan2( wcsDual%X%matrixOffAxis, wcsDual%X%matrixOnAxis)

    if (abs(rotAngleAlt - rotAngle) > PiOn2) then ! one of the pixel
    ! dimensions should be negative. It is immaterial which, but I'll
    ! pick the x-axis dimension as recommended by C&G:
      xPixelSizeWorld = -xPixelSizeWorld

      ! Just as a check:

      rotAngle2 = atan2(-wcsDual%X%matrixOffAxis, -wcsDual%X%matrixOnAxis)
      if (abs(rotAngle2 - rotAngle) > PiOn2)&
call error('wcs_aux_bug0',"ssclib:convertNewToOld: This should not be able &
&to happen. Please contact the code developer.")
    end if

  end subroutine convertNewToOld

  !---------------------------------------------------------------------
  function wcsPixelAreaWorld(wcsDual) result(area)
  !---------------------------------------------------------------------
  ! Calculates the area in world coordinates of an image pixel.

    implicit none

    type(WcsDualT), intent(in) :: wcsDual
    real(double)           :: area

    real(single) :: squareXPixels(2,2), squareYPixels(2,2)
    real(double) ::&
      relPixelsX(2,2), relPixelsY(2,2),&
      parlogramXWorld(2,2), parlogramYWorld(2,2),&
      vectorA(3), vectorB(3), crossProd(3)

    !-------------------------------------------------------------------
    ! Start

    squareXPixels(:,2) = (/0.0, 1.0/)
    squareXPixels(:,1) = (/0.0, 1.0/)

    squareYPixels(:,2) = (/1.0, 1.0/)
    squareYPixels(:,1) = (/0.0, 0.0/)

    relPixelsX = squareXPixels - wcsDual%x%refPixels
    relPixelsY = squareYPixels - wcsDual%y%refPixels

    parlogramXWorld = wcsDual%x%matrixOnAxis  * relPixelsX&
                    + wcsDual%y%matrixOffAxis * relPixelsY
    parlogramYWorld = wcsDual%y%matrixOnAxis  * relPixelsY&
                    + wcsDual%x%matrixOffAxis * relPixelsX

    ! The transformed square has now (in general) been sheared into a
    ! parallelogram, then rotated and offset. We'll now load the sides
    ! of the transformed square into two vectors. Note that we need a z
    ! component in order to be able to do a cross product; but of course
    ! since all the above operations take place in the plane, z components
    ! are set to zero.

    vectorA = (/parlogramXWorld(2,1) - parlogramXWorld(1,1)&
              , parlogramYWorld(2,1) - parlogramYWorld(1,1), 0.0d0/)

    vectorB = (/parlogramXWorld(1,2) - parlogramXWorld(1,1)&
              , parlogramYWorld(1,2) - parlogramYWorld(1,1), 0.0d0/)

    ! Area of a parallelogram defined by two vectors is just the magnitude
    ! of their cross product:

    crossProd = vectorCross(vectorA, vectorB)
    area = sqrt(sum(crossProd * crossProd))

  end function wcsPixelAreaWorld

  !---------------------------------------------------------------------
  subroutine checkImageEquivalence(imageSetNames, equivalent, xSize, ySize, wcsDual)
  !---------------------------------------------------------------------

    implicit none

    character(*),   intent(in)  :: imageSetNames(:)
    logical,        intent(out) :: equivalent
    integer(int16), intent(out) :: xSize, ySize
    type(WcsDualT),     intent(out) :: wcsDual

    integer(int16) :: numImages, imageNum, tempXSize, tempYSize

    !-------------------------------------------------------------------
    ! Start

    equivalent = .true. ! default

    xSize = 0 ! Don't like to leave
    ySize = 0 ! them unset.

    call initializeWcsDualT(wcsDual)

    numImages = size(imageSetNames)
  if (numImages <= 0) return

call readWcsDual(imageSetNames(1), wcsDual)
!*** should also read the wcs from each of the other arrays and test them for equivalence. But there are still problems with the wcs equiv test.

    call getArraySize(imageSetNames(1), xSize, ySize) ! in ssclib/dal_aux
    do imageNum = 2, numImages
      call getArraySize(imageSetNames(imageNum), tempXSize, tempYSize)
      if (tempXSize .ne. xSize .or. tempYSize .ne. ySize) then
        equivalent = .false.
  return
      end if
    end do

  end subroutine checkImageEquivalence

  !---------------------------------------------------------------------
  function getProjectionType(wcs) result(projectionTypeStr)
  !---------------------------------------------------------------------

    implicit none

    type(WcsT), intent(in) :: wcs
    character(3)           :: projectionTypeStr

    !-------------------------------------------------------------------
    ! Start

    projectionTypeStr = wcs%ctype(6:8)

  end function getProjectionType

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module l_wcs_aux
