!+
! Module: padmask_mod
!
! Authors:  Ian Stewart - JBO
!          
! Description: Contains several tasks to help make simulated x-ray images.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module padmask_mod

  use dal
  use param  
  use errorHandling

  use array_utils
  use dal_aux
  use save_image
  use wcs_aux

  implicit none 

  private 

  ! Command-line parameters:
  character(256) ::&
    expMapSetName='',&
    paddedMaskSetName='',&
    tempSetName=''
  real(single) ::&
    padSizeArcsec=0.0

  public padmask

contains 

  !---------------------------------------------------------------------
  subroutine padmask
  !---------------------------------------------------------------------
  ! The output from this sub is used to select which sources from the list
  ! (which covers in general a larger area of the sky) will be added to the
  ! rate image. The padded mask is formed by first creating a mask from an
  ! exposure map and then adding a border of padSizeArcsec width. This
  ! ensures that sources which have centres which fall outside the exposure
  ! map but nevertheless would be expected to make some contribution to the
  ! image through their wings are included on the rate image.

    implicit none

    integer(int16) :: xSize, ySize, xiPad, yiPad,&
      xi,&
      yi, startXi, finisXi, startYi, finisYi
    type(WcsT) :: wcs
    logical(bool), pointer :: mask(:,:), edgeMask(:,:)
    logical(bool), allocatable :: padArray(:,:)
    real(single) :: x, y
    real(single), pointer :: expMapImage(:,:)
    type(DataSetT) :: maskSet
    type(ArrayT) :: maskArray
    real(double) :: rotAngle, xPixelSizeWorld, yPixelSizeWorld

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    call readWcs(expMapSetName, wcs) ! in ssclib/wcs_aux.
    call readArrayData(expMapSetName, expMapImage) ! in ssclib/dal_aux.
    xSize = size(expMapImage, 1)
    ySize = size(expMapImage, 2)
    allocate(mask(xSize, ySize))
    mask = (expMapImage > 0.0)
    deallocate(expMapImage)

    ! Now add the pad:

    call convertNewToOld(wcs, xPixelSizeWorld, yPixelSizeWorld, rotAngle)
    ! in ssclib/wcs_aux.

    select case(wcs%x%worldUnit)
    case('deg')
      xiPad = ceiling(padSizeArcsec / 3600.0 / abs(xPixelSizeWorld))
    case default
call error('notSupported',"At present, WCS units (CUNITn keywords) must be 'deg'.")
    end select

    select case(wcs%y%worldUnit)
    case('deg')
      yiPad = ceiling(padSizeArcsec / 3600.0 / abs(yPixelSizeWorld))
    case default
call error('notSupported',"At present, WCS units (CUNITn keywords) must be 'deg'.")
    end select

    ! First make elliptical pad array (elliptical rather than circular
    ! because the image pixels may not be square):

    allocate(padArray(-xiPad:xiPad,-yiPad:yiPad))
    padArray = .false. ! default
    do xi = -xiPad, xiPad
      x = xi / real(xiPad)
      do yi = -yiPad, yiPad
        y = yi / real(yiPad)
        if (x*x+y*y .lt. 1.0) padArray(xi, yi) = .true.
      end do
    end do

    allocate(edgeMask(xSize, ySize))
    call findEdges(mask, edgeMask) ! in ssclib/array_utils.

    ! For each of the edge points of mask, add a copy of the pad array.
    ! (Another way to think of this is as a convolution of the edge mask
    ! with the pad array.) The result is added to mask.

    do xi = 1, xSize
    if (all(.not. edgeMask(xi, :))) cycle
      do yi = 1, ySize
      if (.not. edgeMask(xi, yi)) cycle
        startXi = max(1_int16, xi - xiPad)
        finisXi = min(xSize,   xi + xiPad)
        startYi = max(1_int16, yi - yiPad)
        finisYi = min(ySize,   yi + yiPad)

        mask(startXi:finisXi, startYi:finisYi)&
        = mask(startXi:finisXi, startYi:finisYi)&
        .or. padArray(startXi-xi:finisXi-xi,startYi-yi:finisYi-yi)
      end do
    end do
    deallocate(edgeMask, padArray)

    call saveImage(mask, expMapSetName, paddedMaskSetName, tempSetName)
    ! in ssclib/save_image.
    deallocate(mask)

    maskSet = dataSet(paddedMaskSetName, MODIFY)
    maskArray = array(maskSet, 0, MODIFY)

    call setAttribute(maskArray, 'PADSIZE', padSizeArcsec, 'arcsec')

    call release(maskSet)

  end subroutine padmask

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    expMapSetName = stringParameter('expmapset')
    paddedMaskSetName = stringParameter('maskset')
    tempSetName = stringParameter('tempset')
    padSizeArcsec = realParameter('padsize')

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module padmask_mod
