!+
! Module: srccompare_mod
!
! Authors:  Ian Stewart - JBO
!          
! Description: Compares detected with the original simulated sources.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module srccompare_mod

  use dal
  use param  
  use errorHandling

  use power_law_utils

  use array_utils
!use l_array_utils_2
  use constants
  use coordinate
  use dal_aux
!use temp_ssclib

  implicit none 

  private 

  type srcCoordType
    real(double) :: xArcsec, yArcsec, l, flux
!    real(double) :: tanX, tanY, l, flux
  ! The 'l' values represent fluxes which have been subjected to some transform to enable them to be better used in the matching process. If --fluxstyle=simple, l=flux; if --fluxstyle=root, l=sqrt(flux) [but err%l=err%flux/2/sqrt(coord%flux)]; if --fluxstyle==rand, 'l' values are given by the flux subjected to the reverse of the logN-logS transform applied in srclistsim, which converted an ensemble of random numbers of even distribution to random fluxes having the desired logN-logS distribution.
  end type srcCoordType
  type detSrcType
    type(srcCoordType) :: coord, err, simCoord
    integer(int32) :: srcId, indexOfNearestSim, srcIdOfNearestSim, flag
    real(single) :: scaledR, simSrcDensity, nullProb, inverseSensyAtSim
    logical :: ignore
  end type detSrcType
  type simSrcType
    type(srcCoordType) :: coord
    integer(int32) :: srcId, indexOfBestDet, srcIdOfBestDet
    real(single) :: inverseSensy
    logical :: ignore
!, flagMask
  end type simSrcType

  ! Command-line parameters:
  character(256) ::&
    srcSpecSetName='',&
    simSrcListSetName='',&
    detSrcListSetTabName=''
  character(20) ::&
    detSrcIdColName='',&
!    randColName='',&
!    randErrColName='',&
!    simFluxColName='',&
!    simFluxErrColName='',&
    detFluxColName='',&
    detFluxErrColName='',&
    detRandColName='',&
    detRandErrColName='',&
    detRaColName='',&
    detDecColName='',&
    detRaDecErrColName='',&
    fluxStyle=''

  public srccompare

contains 

  !---------------------------------------------------------------------
  subroutine srccompare
  !---------------------------------------------------------------------
  ! Given FLUXRAND, X_IMA and Y_IMA values for two source lists, one simulated
  ! and the other detected, this subroutine performs two functions: (i) it
  ! attempts to deduce which of the simulated sources corresponds with each
  ! detected source; (ii) it calculates the probability of the chance
  ! occurrence of a source 'closer' than this to the detected source.
  !
  ! The first part of this task, namely the search for the 'best' match
  ! between each detection and the simulated source list, is carried out by
  ! finding the nearest simulated neighbour to each detection in an
  ! abstract space. This space is defined by the three axes
  ! (X_IMA/X_IMA_ERR), (Y_IMA/Y_IMA_ERR) and (FLUXRAND/FLXRNERR).
  !.....................................................................

    implicit none

    type(DataSetT) :: srcSpecSet
    type(TableT) :: srcSpecTab
    real(double) :: coneRaDeg, coneDecDeg
    real(single), pointer :: modelFluxes(:), modelCumSkyDensity(:)
    real(single) :: hiPower, sourcesPerSteradian
    real(single), allocatable :: modelPowers(:), kVals(:)
    type(detSrcType), pointer :: detSrcData(:)
    type(simSrcType), pointer :: simSrcData(:)
    integer :: numDetSrcs, detSrc, numModelKnees, i, j
    real(double), allocatable :: tanR2(:)
    logical, allocatable :: mask(:)

character(12) :: style = 'ellipse' !*** don't try halfellipse just yet. Need to think about the correct volume shape.

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    srcSpecSet = dataSet(srcSpecSetName, READ)
    srcSpecTab = table(srcSpecSet, 'SRCSPECS')
    coneRaDeg  = real64Attribute(srcSpecTab, 'CONE_RA')
    coneDecDeg = real64Attribute(srcSpecTab, 'CONE_DEC')
    call getLogNlogS(srcSpecTab, modelFluxes, modelCumSkyDensity&
      , hiPower) ! in power_law_utils

    numModelKnees = size(modelFluxes)
if (numModelKnees<1) call error('',"")
    allocate(modelPowers(numModelKnees), kVals(numModelKnees))
    call calcLogNLogSPowers(modelFluxes, modelCumSkyDensity, hiPower&
      , modelPowers) ! in power_law_utils
    kVals = modelCumSkyDensity * (1.0 + modelPowers)&
      / (modelFluxes**(1.0+modelPowers))

    sourcesPerSteradian = modelCumSkyDensity(0) / (PiOn180)**2.0
!print *, sourcesPerSteradian, modelCumSkyDensity(0)
!    deallocate(modelFluxes, modelCumSkyDensity)
    deallocate(modelCumSkyDensity)

    call release(srcSpecTab)
    call release(srcSpecSet)

    ! Load the relevant det and sim source columns:
    !
    call readSimSourceList(coneRaDeg, coneDecDeg, simSrcData)
    call readDetSourceList(coneRaDeg, coneDecDeg, detSrcData)
    numDetSrcs = size(detSrcData)
    allocate(mask(numDetSrcs))

    ! Calculate detSrcData%simSrcDensity from sourcesPerSteradian and tanR:
    !
    allocate(tanR2(numDetSrcs))
!    tanR2 = detSrcData%coord%tanX * detSrcData%coord%tanX&
!      + detSrcData%coord%tanY * detSrcData%coord%tanY
tanR2 = detSrcData%coord%xArcsec * detSrcData%coord%xArcsec&
  + detSrcData%coord%yArcsec * detSrcData%coord%yArcsec
tanR2 = tanR2 * PiOn180 * PiOn180 / 3600.0d0 / 3600.0d0
    detSrcData%simSrcDensity = sourcesPerSteradian * (1.0 + tanR2)**1.5
!***** I _think_ this is correct.

    !...................................................................
    ! Main loop over detected sources:

    do detSrc = 1, numDetSrcs
      call findNearestSimSource(detSrcData(detSrc), simSrcData, style)
!, DETECTED_BIT, indexOfNearestSim(detSrc), devOverStdDev(detSrc))


!******** ONLY CONSIDERS THOSE SOURCES ON THE EXPOSURE MAP. Must have a way of flagging these.

!****** Equally, we assume that det srcs not lying on the exp map have already been weeded out.

      call probOfOneOrMoreByChance(kVals, modelFluxes, modelPowers&
        , detSrcData(detSrc))
    end do

deallocate(modelFluxes)

    ! It can happen that 2 or more det sources match with a single sim source. In this case we should try to weed out all but 1 det source as the match. The best way to do this is probably to retain just that det source which has the smallest nullProb value.
    !
    detSrcData%flag = 0
    do i = minval(detSrcData%indexOfNearestSim), maxval(detSrcData%indexOfNearestSim)
      mask = (detSrcData%indexOfNearestSim == i)
    if (count(mask)<1) cycle

!      j = minloc(detSrcData%nullProb, mask)
      j = minloc1d(detSrcData%nullProb, mask) ! in ssclib/array_utils.
      mask(j) = .false.
      where(mask) detSrcData%flag = ibset(detSrcData%flag, 0)
      simSrcData(i)%indexOfBestDet = j
      simSrcData(i)%srcIdOfBestDet = detSrcData(j)%srcId
    end do

!!!    call flagAlsoRans(indexOfNearestSim, devOverStdDev, alsoRan, status)
!!!    if (status .lt. 0) then
!!!      print *, 'det src', detSrc, ': flagAlsoRans status return'&
!!!      , status
!!!  return
!!!    end if

    !...................................................................
    ! Wrap up:

    call saveFinalDetSrcList(detSrcData)
    call modifySimSrcList(simSrcData)

!    simSourceSet = dataSet(simSrcSetName, MODIFY)
!    simSourceTab = table(simSourceSet, 'SRCLIST')
!    flag => int16Data(column(simSourceTab, 'FLAG_'//obsId))
!    flag = simSrcFlag
!    call release(simSourceSet)


    deallocate(detSrcData, simSrcData)

  end subroutine srccompare

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    srcSpecSetName       = stringParameter('srcspecset')
    simSrcListSetName    = stringParameter('simsrclistset')
    detSrcListSetTabName = stringParameter('detsrclisttab')
    detSrcIdColName      = stringParameter('srcidcol')
!    simFluxColName       = stringParameter('simfluxcol')
!    simFluxErrColName    = stringParameter('simfluxerrcol')
    detFluxColName       = stringParameter('fluxcol')
    detFluxErrColName    = stringParameter('fluxerrcol')
    detRaColName         = stringParameter('racol')
    detDecColName        = stringParameter('deccol')
    detRaDecErrColName   = stringParameter('radecerrcol')
    fluxStyle            = stringParameter('fluxstyle')

    select case(fluxStyle)
    case('simple')
    case('rand')
      detRandColName    = stringParameter('randcol')
      detRandErrColName = stringParameter('randerrcol')
    case('root')
call error('notYetAvailable',"")
    case default
call error('badFluxStyle',"")
    end select

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine readSimSourceList(coneRaDeg, coneDecDeg, simSrcData)
  !---------------------------------------------------------------------

    implicit none

    real(double),     intent(in) :: coneRaDeg, coneDecDeg
    type(simSrcType), pointer :: simSrcData(:)

    type(DataSetT) :: set
    type(TableT) :: tab
    integer :: numSources, i
!, flagBit, numValidSources
    type(ColumnT) :: raDegCol, decDegCol
    real(double), pointer :: raDeg(:), decDeg(:)
!    integer(int32), allocatable :: indices(:), flagMask(:)
    real(double), allocatable :: ra(:), dec(:), tanX(:), tanY(:)
!, rands(:)
!    logical, allocatable :: srcIsOnMap(:)
    type(ColumnT) :: isyCol
    real(single), pointer :: isyData(:)

    !-------------------------------------------------------------------
    ! Start

    set = dataSet(simSrcListSetName, READ)
    tab = table(set, 'SRCLIST')
    numSources = numberOfRows(tab)
if (numSources <= 0) call error('noSimSources',"")

    raDegCol  = column(tab, 'RA')
    decDegCol = column(tab, 'DEC')
    raDeg  => real64Data(raDegCol)
    decDeg => real64Data(decDegCol)

!    if (hasAttribute(tab, 'BIT_EMAP')) then
!      flagBit = int32Attribute(tab, 'BIT_EMAP')
!
!      allocate(indices(numSources), flagMask(numSources), rands(numSources))
!      allocate(srcIsOnMap(numSources))
!
!      call readColDataToFixed(tab, 'INDEX',    indices)
!      call readColDataToFixed(tab, 'FLAGMASK', flagMask)
!      call readColDataToFixed(tab, 'FLUXRAND', rands)
!
!      srcIsOnMap = btest(flagMask, flagBit)
!      numValidSources = count(srcIsOnMap)
!if (numValidSources <= 0) call error('noValidSimSources',"")
!
!      allocate(simSrcData(numValidSources))
!      allocate(ra(numValidSources), dec(numValidSources))
!
!      simSrcData%index      = pack(indices,  srcIsOnMap)
!      simSrcData%flagMask   = pack(flagMask, srcIsOnMap)
!      simSrcData%coord%flux = pack(rands,    srcIsOnMap)
!      ra          = PiOn180 * pack(raDeg,    srcIsOnMap)
!      dec         = PiOn180 * pack(decDeg,   srcIsOnMap)
!
!    else
!**** I'm not setting these flags now, 1stly its too much trouble, 2ndly I decided it doesn't matter if one tries to match with sources off the exp map, they're not likely to match anyway.


      allocate(simSrcData(numSources))

!      call readColDataToFixed(tab, 'INDEX',    simSrcData%index)
       call readColDataToFixed(tab, 'INDEX',    simSrcData%srcId)
!      call readColDataToFixed(tab, 'FLAGMASK', simSrcData%flagMask)
!simSrcData%flagMask = 0
!      call readColDataToFixed(tab, 'INV_SENSY', simSrcData%inverseSensy)
isyCol = column(tab, 'INV_SENSY')
isyData => real32Data(isyCol)
simSrcData%inverseSensy = 0.0 ! default
do i = 1, numSources
  if (.not.isNull(isyCol, i-1)) simSrcData(i)%inverseSensy = isyData(i-1)
end do
call release(isyCol)
!*** at present this is just either 0 or 1 depending on whether the source lies on the detection mask or not. IMS 2006-09-19.

      call readColDataToFixed(tab, 'FLUX',      simSrcData%coord%flux)

      select case(fluxStyle)
      case('rand')
        call readColDataToFixed(tab, 'FLUXRAND', simSrcData%coord%l)
      case('simple')
        simSrcData%coord%l = simSrcData%coord%flux
      case('root')
        simSrcData%coord%l = sqrt(simSrcData%coord%flux)
      case default
call error('badFluxStyle',"")
      end select

      allocate(ra(numSources), dec(numSources))
      ra  = PiOn180 * raDeg
      dec = PiOn180 * decDeg
!    end if

!    call raDecToTan(ra, dec, coneRaDeg * PiOn180&
!      , coneDecDeg * PiOn180, simSrcData%coord%tanX, simSrcData%coord%tanY)
allocate(tanX(numSources), tanY(numSources))
call raDecToTan(ra, dec, coneRaDeg * PiOn180, coneDecDeg * PiOn180, tanX, tanY)
    ! in ssclib/coordinate.
simSrcData%coord%xArcsec = tanX * 3600.0 / PiOn180
simSrcData%coord%yArcsec = tanY * 3600.0 / PiOn180

    simSrcData%indexOfBestDet = 0 ! default - 0 indicates no match.
    simSrcData%srcIdOfBestDet = 0 ! default - 0 indicates no match.

    call release(decDegCol)
    call release(raDegCol)
    call release(tab)
    call release(set)

  end subroutine readSimSourceList

  !---------------------------------------------------------------------
  subroutine readDetSourceList(coneRaDeg, coneDecDeg, detSrcData)
  !---------------------------------------------------------------------

    implicit none

    real(double),     intent(in) :: coneRaDeg, coneDecDeg
    type(detSrcType), pointer    :: detSrcData(:)

    character(256) :: detSrcListSetName
    character(20) :: detSrcListTabName
    type(DataSetT) :: set
    type(TableT) :: tab
    integer :: numSources
    real(double), allocatable :: tanX(:), tanY(:)
!raDecErr(:), 
    type(ColumnT) :: raDegCol, decDegCol, raDecErrCol
    real(double), pointer :: raDeg(:), decDeg(:)
    real(single), pointer :: raDecErrArcsec(:)

    !-------------------------------------------------------------------
    ! Start

    call splitSetTabName(detSrcListSetTabName, detSrcListSetName, detSrcListTabName)
    ! in ssclib/dal_aux.
    set = dataSet(detSrcListSetName, READ)
    tab = table(set, detSrcListTabName)
    numSources = numberOfRows(tab)
if (numSources <= 0) call error('noDetSources','')
    allocate(detSrcData(numSources))
!, raDecErr(numSources))

    call readColDataToFixed(tab, detSrcIdColName,   detSrcData%srcId)
    call readColDataToFixed(tab, detFluxColName,    detSrcData%coord%flux)
    call readColDataToFixed(tab, detFluxErrColName, detSrcData%err%flux)

    select case(fluxStyle)
    case('rand')
      call readColDataToFixed(tab, detRandColName,    detSrcData%coord%l)
      call readColDataToFixed(tab, detRandErrColName, detSrcData%err%l)
    case('simple')
      detSrcData%coord%l = detSrcData%coord%flux
      detSrcData%err%l   = detSrcData%err%flux
    case('root')
      where(detSrcData%coord%flux > 0.0)
        detSrcData%coord%l = sqrt(detSrcData%coord%flux)
        detSrcData%err%l = detSrcData%err%flux / 2.0&
          / sqrt(detSrcData%coord%flux)
      end where
    case default
call error('badFluxStyle',"")
    end select

    raDegCol    = column(tab, detRaColName)
    decDegCol   = column(tab, detDecColName)
    raDecErrCol = column(tab, detRaDecErrColName)

    raDeg    => real64Data(raDegCol)
    decDeg   => real64Data(decDegCol)
    raDecErrArcsec => real32Data(raDecErrCol) ! assume it is arcsec???

!    raDecErr = raDecErrArcsec * PiOn180 / 3600.0d0

allocate(tanX(numSources), tanY(numSources))
!    call raDecToTan(raDeg * PiOn180, decDeg * PiOn180, coneRaDeg * PiOn180&
!      , coneDecDeg * PiOn180, detSrcData%coord%tanX, detSrcData%coord%tanY)
call raDecToTan(raDeg * PiOn180, decDeg * PiOn180, coneRaDeg * PiOn180&
  , coneDecDeg * PiOn180, tanX, tanY)
    ! in ssclib/coordinate.
detSrcData%coord%xArcsec = tanX * 3600.0 / PiOn180
detSrcData%coord%yArcsec = tanY * 3600.0 / PiOn180
    ! in ssclib/coordinate.

!    detSrcData%err%tanX = raDecErr * (1.0 + detSrcData%coord%tanX&
!      * detSrcData%coord%tanX)
!    detSrcData%err%tanY = raDecErr * (1.0 + detSrcData%coord%tanY&
!      * detSrcData%coord%tanY)
detSrcData%err%xArcsec = raDecErrArcsec * (1.0 + tanX * tanX)
detSrcData%err%yArcsec = raDecErrArcsec * (1.0 + tanY * tanY)

    call release(raDecErrCol)
    call release(decDegCol)
    call release(raDegCol)
    call release(tab)
    call release(set)

  end subroutine readDetSourceList

  !---------------------------------------------------------------------
  subroutine findNearestSimSource(detSrcData, simSrcData, style)
  !---------------------------------------------------------------------
  ! Style = 'ellipse':
  ! ------------------
  !
  ! Given the position (detX and detY) and brightness (detL) of
  ! a detected source, plus the uncertainties in these values, and given
  ! a list of the simulated sources, the present subroutine selects that
  ! simulated source which is 'closest', that is, which gives the smallest
  ! value of the quantity
  !
  !	      (simX-detX)^2     (simY-detY)^2     (simL-detL)^2
  !	r2 = --------------- + --------------- + ---------------.
  !	       detXuncert^2      detYuncert^2      detLuncert^2
  !
  ! Style = 'halfellipse':
  ! ------------------
  !
  ! For this style, the 'closest' simulated source is defined as that one
  ! which gives the smallest value of the quantity
  !
  !	      (simX-detX)^2     (simY-detY)^2     (simL-detL)^2
  !	r2 = --------------- + --------------- + ---------------
  !	       detXuncert^2      detYuncert^2      detLuncert^2
  !
  ! (same as above) for simL <= detL, and
  !
  !	      (simX-detX)^2     (simY-detY)^2
  !	z2 = --------------- + ---------------
  !	       detXuncert^2      detYuncert^2
  !
  ! for simL > detL.
  !.....................................................................

    implicit none

    type(detSrcType), intent(inout) :: detSrcData
    type(simSrcType), intent(in)    :: simSrcData(:)
    character(*),     intent(in)    :: style

    integer :: numSources, rLocVec(1), zLocVec(1), iSpatiallyNearestBrighterSim
    real(double), allocatable :: dX(:), dY(:), dL(:), r2(:), z2(:)

    !-------------------------------------------------------------------
    ! Start

    detSrcData%indexOfNearestSim = 0 ! nonsensical default
    detSrcData%scaledR = 0.0
    numSources = size(simSrcData)
    if (numSources <= 0)&
call error('bug0',"Please contact the code developer.")

!    if (detSrcData%err%tanX <= 0.0&
!    .or. detSrcData%err%tanY <= 0.0&
    if (detSrcData%err%xArcsec <= 0.0&
    .or. detSrcData%err%yArcsec <= 0.0&
    .or. detSrcData%err%l <= 0.0)&
call error('negativeUncerts',"Non-positive values of position &
&or rate uncertainties were found in the list of detected sources.")

    if (numSources == 1) then ! trivial
      detSrcData%indexOfNearestSim = 1
      detSrcData%srcIdOfNearestSim = simSrcData(1)%srcId
!!!      simSrcFlag(1) = ibset(simSrcFlag(1), bitToFlagDetection)
      call warning('onlyOneSimSource',"There is only one &
        &simulated source to match to.")
  return
    end if

    allocate(dX(numSources), dY(numSources), dL(numSources))
    allocate(z2(numSources), r2(numSources))

!    dX = simSrcData%coord%tanX - detSrcData%coord%tanX
!    dY = simSrcData%coord%tanY - detSrcData%coord%tanY
dX = (simSrcData%coord%xArcsec - detSrcData%coord%xArcsec) * PiOn180 / 3600.0d0
dY = (simSrcData%coord%yArcsec - detSrcData%coord%yArcsec) * PiOn180 / 3600.0d0
    dL = simSrcData%coord%l    - detSrcData%coord%l

!    z2 = dX * dX / detSrcData%err%tanX / detSrcData%err%tanX&
!       + dY * dY / detSrcData%err%tanY / detSrcData%err%tanY
z2 = (dX * dX / detSrcData%err%xArcsec / detSrcData%err%xArcsec&
   +  dY * dY / detSrcData%err%yArcsec / detSrcData%err%yArcsec)&
   * 3600.0d0 * 3600.0d0 / PiOn180 / PiOn180

    r2 = z2 + dL * dL / detSrcData%err%l / detSrcData%err%l

    select case(style)
    case('ellipse')
      rLocVec = minloc(r2)
      detSrcData%indexOfNearestSim = rLocVec(1)
      detSrcData%srcIdOfNearestSim = simSrcData(rLocVec(1))%srcId
!!!      simSrcFlag(rLocVec(1)) = ibset(simSrcFlag(rLocVec(1)), bitToFlagDetection)
!      detSrcData%scaledR = sqrt(r2(detSrcData%indexOfNearestSim))
detSrcData%scaledR = sqrt(r2(rLocVec(1)))
detSrcData%simCoord = simSrcData(rLocVec(1))%coord

    case('halfellipse')
      if (any(dL > 0.0)) then
        zLocVec = minloc(z2, (dL > 0.0)) ! closest source (in a spatial
        ! sense) which is brighter than the detection.

        iSpatiallyNearestBrighterSim = zLocVec(1)

        if (any(dL <= 0.0)) then
          rLocVec = minloc(r2, (dL <= 0.0)) ! closest source (in the absolute
          ! sense) which is fainter than the detection.

          detSrcData%indexOfNearestSim = rLocVec(1)
          detSrcData%srcIdOfNearestSim = simSrcData(rLocVec(1))%srcId
!!!          simSrcFlag(rLocVec(1)) = ibset(simSrcFlag(rLocVec(1))&
!!!            , bitToFlagDetection)

!          if (r2(indexOfNearestSim) < z2(iSpatiallyNearestBrighterSim))&
!          then
          if (r2(rLocVec(1)) < z2(iSpatiallyNearestBrighterSim)) then
!            devOverStdDev = sqrt(r2(indexOfNearestSim))
detSrcData%scaledR = sqrt(r2(rLocVec(1)))
detSrcData%simCoord = simSrcData(rLocVec(1))%coord

          else
!            detSrcData%indexOfNearestSim = simSrcData(iSpatiallyNearestBrighterSim)%index
            detSrcData%indexOfNearestSim = iSpatiallyNearestBrighterSim
            detSrcData%srcIdOfNearestSim = simSrcData(iSpatiallyNearestBrighterSim)%srcId
!!!            simSrcFlag(iSpatiallyNearestBrighterSim)&
!!!              = ibset(simSrcFlag(iSpatiallyNearestBrighterSim)&
!!!              , bitToFlagDetection)

!            devOverStdDev = sqrt(z2(indexOfNearestSim))
detSrcData%scaledR = sqrt(z2(iSpatiallyNearestBrighterSim))
detSrcData%simCoord = simSrcData(iSpatiallyNearestBrighterSim)%coord
!*** flag the detection?
          end if

        else ! there are no sim sources fainter than detection.
          detSrcData%indexOfNearestSim = iSpatiallyNearestBrighterSim
          detSrcData%srcIdOfNearestSim = simSrcData(iSpatiallyNearestBrighterSim)%srcId
!!!          simSrcFlag(iSpatiallyNearestBrighterSim)&
!!!            = ibset(simSrcFlag(iSpatiallyNearestBrighterSim)&
!!!            , bitToFlagDetection)

!          devOverStdDev = sqrt(z2(indexOfNearestSim))
detSrcData%scaledR = sqrt(z2(iSpatiallyNearestBrighterSim))
detSrcData%simCoord = simSrcData(iSpatiallyNearestBrighterSim)%coord
!*** flag the detection?
        end if

      else ! there are no sim sources brighter than the detection.
        rLocVec = minloc(r2)
        detSrcData%indexOfNearestSim = rLocVec(1)
        detSrcData%srcIdOfNearestSim = simSrcData(rLocVec(1))%srcId
!!!         simSrcFlag(rLocVec(1)) = ibset(simSrcFlag(rLocVec(1))&
!!!           , bitToFlagDetection)

!        devOverStdDev = sqrt(r2(indexOfNearestSim))
detSrcData%scaledR = sqrt(r2(rLocVec(1)))
detSrcData%simCoord = simSrcData(rLocVec(1))%coord
      end if

    case default
call error('badStyle',"eimsim_utils: Source matching style not recognized.")
    end select

    detSrcData%inverseSensyAtSim&
      = simSrcData(detSrcData%indexOfNearestSim)%inverseSensy

  end subroutine findNearestSimSource

  !---------------------------------------------------------------------
  subroutine probOfOneOrMoreByChance(kIs, Lis, powerIs, detSrcData)
  !---------------------------------------------------------------------
  ! Where sources are randomly distributed in 3 dimensions r_ = (X,Y,L)
  ! with a probability density q(r_), the probability P that 1 or more
  ! sources will lie within a given volume V is given by
  !
  !	P = 1 - exp(-N_V)						(1)
  !
  ! where N_V is the number of sources that lies (on average) inside V; or,
  ! in other words,
  !
  !	               V
  !	           (  /           )
  !	P = 1 - exp( -| dr_ q[r_] ).
  !	           (  /           )
  !
  ! In cases (such as the present) where q is (to good approximation)
  !a function of L only, N_V can be expressed as
  !
  !	     L_V_max
  !	       /
  !	N_V =  | dL A_V(L) q(L).
  !	       /
  !	    L_V_min
  !
  ! where L_V_min and L_V_max are respectively the smallest and largest
  ! values of L enclosed within V, and A_V(L) is the area of the cross-
  ! section through V normal to the L axis at the specified value of L.
  !
  ! In the present subroutine, V is assumed to be an ellipsoid centred on
  ! detX, detY and detL, with axes axisX, axisY and axisL obtained by
  ! multiplying detSrcData%scaledR (this is set within subroutine
  ! findNearestSimSource) times respectively detXuncert, detYuncert and
  ! detLuncert. The cross-section function A_V(L) is therefore given by
  !
  !	A_V(L) = Pi axisX axisY (1 - [L - detL]^2 / axisL^2).
  !
  ! The functional form of q(L) depends on the value of the 'fluxStyle'
  ! variable as follows:
  !
  !	fluxStyle == 'simple':
  !	  In this case L is simply proportional to source flux, and q(L)
  !	  is the same piecewise power law defined in the source
  !	  specification dataset which should be supplied to the command-
  !	  line parameter --srcspecset; ie
  !
  !	  q(L) = k_i L**power_i for L_i <= L < L_(i+1).
  !
  !	fluxStyle == 'root':
  !	  Here L is the square root of flux, and q(L) should be modified
  !	  accordingly.
  !
  !	fluxStyle == 'rand':
  !	  In this case the 'brightness' value L has been subjected to a
  !	  coordinate transform so as to make q(L) constant at
  !
  !	  q(L) = detSrcData%simSrcDensity for 0.0 <= L < 1.0
  !	       = 0 elsewhere.
  !
  !.....................................................................

    implicit none

    real(single),     intent(in)    :: kIs(:),&
                                       Lis(size(kIs)),&
                                       powerIs(size(kIs))
    type(detSrcType), intent(inout) :: detSrcData

    real(single) :: axisX, axisY, axisL, numV

    !-------------------------------------------------------------------
    ! Start

    detSrcData%nullProb = 0.0 ! default

    if (detSrcData%err%xArcsec <= 0.0&
    .or. detSrcData%err%yArcsec <= 0.0&
    .or. detSrcData%err%l <= 0.0)&
call error('negativeUncerts',"Non-positive values of position &
&or rate uncertainties were found in the list of detected sources.")

    axisL = detSrcData%err%l * detSrcData%scaledR

    ! Transform the arcsec positions to tangents in the process:
    axisX = detSrcData%err%xArcsec * detSrcData%scaledR * PiOn180 / 3600.0
    axisY = detSrcData%err%yArcsec * detSrcData%scaledR * PiOn180 / 3600.0

    select case(fluxStyle)
!    case('simple')
!      numV = calcNumSrcsInVolumeSimple(detSrcData%coord%l, axisX, axisY&
!        , axisL, kIs, Lis, powerIs)
    case('rand')
      numV = calcNumSrcsInVolumeRand(detSrcData%coord%l, axisX, axisY&
        , axisL, detSrcData%simSrcDensity)
!    case('root')
    case default
call error('badFluxStyle',"")
    end select

    if (numV <= 0.0) then
      detSrcData%nullProb = 0.0
      call warning('negativeProbability', 'This ought not to &
        &happen! Best contact the code developer.')

    else if (numV >= -log(tiny(numV))) then
      detSrcData%nullProb = 1.0 ! saves the possibility of an underflow on the exp().

    else if (log(numV * numV / 2.0) < (1 - precision(numV))) then
      detSrcData%nullProb = numV
!      call warning('probabilityUnderflow', 'No cause for concern. This &
!      &is just to inform you that the probability has been calculated &
!      &by use of the approximation prob = N rather than the exact &
!      &prob = 1 - exp(-N).')

    else
      detSrcData%nullProb = 1.0 - exp(-numV)
    end if

  end subroutine probOfOneOrMoreByChance

  !---------------------------------------------------------------------
  subroutine probOfOneOrMoreByChance_old(kVals, modelFluxes, modelPowers, detSrcData)
  !---------------------------------------------------------------------

    implicit none

    real(single),     intent(in)    :: kVals(:),&
                                       modelFluxes(size(kVals)),&
                                       modelPowers(size(kVals))
    type(detSrcType), intent(inout) :: detSrcData

!    real(single), intent(in)  :: srcL,&
!                                 srcLuncert,&
!                                 devOverStdDev,&
!                                 srcPerPixel,&
!                                 srcXuncert,&
!                                 srcYuncert
!    real(single), intent(out) :: probability

    real(single) :: deltaX, deltaY, deltaL, numV

    !-------------------------------------------------------------------
    ! Start

    detSrcData%nullProb = 0.0 ! default

!    if (detSrcData%err%tanX <= 0.0&
!    .or. detSrcData%err%tanY <= 0.0&
    if (detSrcData%err%xArcsec <= 0.0&
    .or. detSrcData%err%yArcsec <= 0.0&
    .or. detSrcData%err%l <= 0.0)&
call error('negativeUncerts',"Non-positive values of position &
&or rate uncertainties were found in the list of detected sources.")

!    deltaX = detSrcData%err%tanX * detSrcData%scaledR
!    deltaY = detSrcData%err%tanY * detSrcData%scaledR
deltaX = detSrcData%err%xArcsec * detSrcData%scaledR * PiOn180 / 3600.0
deltaY = detSrcData%err%yArcsec * detSrcData%scaledR * PiOn180 / 3600.0
    deltaL = detSrcData%err%l * detSrcData%scaledR

    select case(fluxStyle)
!    case('simple')
!      numV = calcNumSrcsInVolumeSimple(detSrcData%coord%l, deltaX, deltaY&
!        , deltaL, kVals, modelFluxes, modelPowers)
    case('rand')
      numV = calcNumSrcsInVolumeRand(detSrcData%coord%l, deltaX, deltaY&
        , deltaL, detSrcData%simSrcDensity)
!print *, detSrcData%simSrcDensity, numV
!    case('root')
    case default
call error('badFluxStyle',"")
    end select

    if (numV <= 0.0) then
      detSrcData%nullProb = 0.0
      call warning('negativeProbability', 'This ought not to &
        &happen! Best contact the code developer.')

    else if (numV .ge. -log(tiny(numV))) then
      detSrcData%nullProb = 1.0 ! saves the possibility of an underflow on the exp().

    else if (log(numV * numV / 2.0) .lt. (1 - precision(numV))) then
      detSrcData%nullProb = numV
!      call warning('probabilityUnderflow', 'No cause for concern. This &
!      &is just to inform you that the probability has been calculated &
!      &by use of the approximation prob = N rather than the exact &
!      &prob = 1 - exp(-N).')

    else
      detSrcData%nullProb = 1.0 - exp(-numV)
    end if

  end subroutine probOfOneOrMoreByChance_old

  !---------------------------------------------------------------------
  function calcNumSrcsInVolumeRand(detL, axisX, axisY, axisL, srcPerPixel)&
  result(numV)
  !---------------------------------------------------------------------
  ! Calculates the integral
  !
  !	     detL+axisL
  !	       /
  !	N_V =  | dL A_V(L) q(L).
  !	       /
  !	    detL-axisL
  !
  ! where
  !
  !	A_V(L) = Pi axisX axisY (1 - [L - detL]^2 / axisL^2)
  !
  ! and q(L) is a constant = srcPerPixel (the result of transforming the
  ! flux coordinate).

    implicit none

    real(double), intent(in) :: detL
    real(single), intent(in) :: axisX, axisY, axisL, srcPerPixel
    real(single)             :: numV

    real(double) :: L_V_max, L_V_min, axisLsquared, L_start, L_finis

    !-------------------------------------------------------------------
    ! Start

    numV = 0.0 ! default

    L_V_max = real(detL, double) + real(axisL, double)
    L_V_min = real(detL, double) - real(axisL, double)

    if (L_V_max <= 0.0d0) then
      call warning('srcTooFaint',"The brightness of the &
        &detected source, plus its brightness uncertainty, was fainter &
        &than any of the simulated sources.")
  return
    end if

    if (L_V_min >= 1.0d0) then
call error('bug2',"If this occurs, please contact the code &
&developer.")
    end if

    axisLsquared = axisL * axisL
    L_start = max(0.0d0, L_V_min)
    L_finis = min(1.0d0, L_V_max)

    numV = ellipsoidIntegral(L_start, L_finis, axisLsquared&
          , 0.0d0, real(srcPerPixel, double), real(detL, double))
    
    numV = numV * Pi * axisX * axisY

  end function calcNumSrcsInVolumeRand

  !---------------------------------------------------------------------
  function calcNumSrcsInVolumeRand_old(srcL, deltaX, deltaY, deltaL, srcPerPixel)&
  result(numV)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: srcL
    real(single), intent(in) :: deltaX, deltaY, deltaL, srcPerPixel
    real(single)             :: numV

    real(double) :: L_V_max, L_V_min, deltaLsquared, L_start, L_finis

    !-------------------------------------------------------------------
    ! Start

    numV = 0.0 ! default

    L_V_max = real(srcL, double) + real(deltaL, double)
    L_V_min = real(srcL, double) - real(deltaL, double)

    if (L_V_max .le. 0.0d0) then
      call warning('srcTooFaint',"The brightness of the &
        &detected source, plus its brightness uncertainty, was fainter &
        &than any of the simulated sources.")
  return
    end if

    if (L_V_min .ge. 1.0d0) then
call error('bug2',"If this occurs, please contact the code &
&developer.")
    end if

    deltaLsquared = deltaL * deltaL
    L_start = max(0.0d0, L_V_min)
    L_finis = min(1.0d0, L_V_max)

    numV = ellipsoidIntegral(L_start, L_finis, deltaLsquared&
          , 0.0d0, real(srcPerPixel, double), real(srcL, double))
    
    numV = numV * Pi * deltaX * deltaY

  end function calcNumSrcsInVolumeRand_old

  !---------------------------------------------------------------------
  function calcNumSrcsInVolumeSimple(detL, axisX, axisY, axisL&
  , kIs, Lis, powerIs) result(numV)
  !---------------------------------------------------------------------
  ! Calculates the integral
  !
  !	     detL+axisL
  !	       /
  !	N_V =  | dL A_V(L) q(L).
  !	       /
  !	    detL-axisL
  !
  ! where
  !
  !	A_V(L) = Pi axisX axisY (1 - [L - detL]^2 / axisL^2)
  !
  ! and q(L) is a sequential power law, ie
  !
  !	q(L) = k_i L**power_i for L_i <= L < L_(i+1).
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: detL
    real(single), intent(in) :: axisX, axisY, axisL,&
                                kIs(:),&
                                Lis(size(kIs)),&
                                powerIs(size(kIs))
    real(single)             :: numV

    integer :: numKnees, i
    real(double) :: L_V_max, L_V_min, axisLsquared, L_start, L_finis

    !-------------------------------------------------------------------
    ! Start

    numV = 0.0 ! default

    numKnees = size(kIs)

    L_V_max = real(detL, double) + real(axisL, double)
    L_V_min = real(detL, double) - real(axisL, double)

    axisLsquared = axisL * axisL

    do i = 1, numKnees-1
    if (Lis(i) > L_V_max .or. Lis(i+1) < L_V_min) cycle
      L_start = max(real(Lis(i  ), double), L_V_min)
      L_finis = min(real(Lis(i+1), double), L_V_max)

      numV = numV + ellipsoidIntegral(L_start, L_finis, axisLsquared&
        , real(powerIs(i), double), real(kIs(i), double), real(detL, double))
    end do

    L_start = max(real(Lis(numKnees), double), L_V_min)
    L_finis = L_V_max

    numV = numV + ellipsoidIntegral(L_start, L_finis, axisLsquared&
      , real(powerIs(numKnees), double), real(kIs(numKnees), double)&
      , real(detL, double))

    numV = numV * Pi * axisX * axisY

  end function calcNumSrcsInVolumeSimple

  !---------------------------------------------------------------------
  function calcNumSrcsInVolumeSimple_old(srcL, deltaX, deltaY, deltaL&
  , kVals, modelFluxes, modelPowers) result(numV)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: srcL
    real(single), intent(in) :: deltaX, deltaY, deltaL,&
                                kVals(:),&
                                modelFluxes(size(kVals)),&
                                modelPowers(size(kVals))
    real(single)             :: numV

    integer :: numModelKnees, i
    real(double) :: L_V_max, L_V_min, deltaLsquared, L_start, L_finis

    !-------------------------------------------------------------------
    ! Start

    numV = 0.0 ! default

    numModelKnees = size(kVals)

    L_V_max = real(srcL, double) + real(deltaL, double)
    L_V_min = real(srcL, double) - real(deltaL, double)

    deltaLsquared = deltaL * deltaL

    do i = 1, numModelKnees-1
    if (modelFluxes(i) > L_V_max .or. modelFluxes(i+1) < L_V_min) cycle
      L_start = max(real(modelFluxes(i  ), double), L_V_min)
      L_finis = min(real(modelFluxes(i+1), double), L_V_max)

      numV = numV + ellipsoidIntegral(L_start, L_finis, deltaLsquared&
        , real(modelPowers(i), double), real(kVals(i), double), real(srcL, double))
    end do

    L_start = max(real(modelFluxes(numModelKnees), double), L_V_min)
    L_finis = L_V_max

    numV = numV + ellipsoidIntegral(L_start, L_finis, deltaLsquared&
      , real(modelPowers(numModelKnees), double), real(kVals(numModelKnees), double)&
      , real(srcL, double))

    numV = numV * Pi * deltaX * deltaY

  end function calcNumSrcsInVolumeSimple_old

  !---------------------------------------------------------------------
  function ellipsoidIntegral(L_start, L_finis, deltaLsquared, power, amp&
  , srcL) result(term)
  !---------------------------------------------------------------------
  ! The present subroutine calculates the integral
  !
  !	   L_finis
  !	     /    (     [L - srcL]^2  )
  !	I =  | dL (1 - ---------------) L**power.
  !	     /    (     deltaLsquared )
  !	  L_start

    implicit none

    real(double), intent(in) :: L_start,&
                                L_finis,&
                                power,&
                                srcL,&
                                deltaLsquared,&
                                amp
    real(double)             :: term

    real(double) ::&
      intgrlTerm1, intgrlTerm2, intgrlTerm3,&
      pPlus1, pPlus2, pPlus3

    !-------------------------------------------------------------------
    ! Start

    pPlus1 = 1.0d0 + power
    pPlus2 = 2.0d0 + power
    pPlus3 = 3.0d0 + power

!    if (pPlus1 == 0.0 .or. pPlus2 == 0.0 .or. pPlus3 == 0.0)&
!call error('noIntegralPossible',"This integral cannot be done at &
!&the value of power supplied.")

    if (pPlus1 == 0.0) then
      intgrlTerm1 =  (log(L_finis) - log(L_start))&
        * (1.0d0 - srcL * srcL / deltaLsquared)
    else
      intgrlTerm1 =  (L_finis**pPlus1 - L_start**pPlus1) / pPlus1&
        * (1.0d0 - srcL * srcL / deltaLsquared)
    end if

    if (pPlus2 == 0.0) then
      intgrlTerm2 =  (log(L_finis) - log(L_start))&
        * 2.0d0 * srcL / deltaLsquared
    else
      intgrlTerm2 =  (L_finis**pPlus2 - L_start**pPlus2) / pPlus2&
        * 2.0d0 * srcL / deltaLsquared
    end if

    if (pPlus3 == 0.0) then
      intgrlTerm3 = -(log(L_finis) - log(L_start))&
        / deltaLsquared
    else
      intgrlTerm3 = -(L_finis**pPlus3 - L_start**pPlus3) / pPlus3&
        / deltaLsquared
    end if

    term = amp * (intgrlTerm1 + intgrlTerm2 + intgrlTerm3)

  end function ellipsoidIntegral

  !---------------------------------------------------------------------
  function ellipsoidIntegral_new(a, b, c, power, amp&
  , x0) result(integral)
  !---------------------------------------------------------------------
  ! The present subroutine calculates the integral
  !
  !	        b
  !	        /    (     [x - x0]^2 )
  !	I = amp | dx (1 - ------------) x**power.
  !	        /    (         c      )
  !	       a
  !
  ! This is broken into 3 sub-integrals as follows:
  !
  !	                      b
  !	  I                   /
  !	----- =  (1 - x0^2/c) | dx x**power
  !	 amp                  /
  !	                     a
  !	          b
  !	      2x0 /
  !	  +  ---- | dx x**(power+1)
  !	       c  /
  !	         a
  !	               b
  !	            1  /
  !	         - --- | dx x**(power+2).
  !	            c  /
  !	              a
  !
  ! The integrals for the 3 terms follow as given in the code.
  !
  ! NOTE that if you really ARE calculating the integral of a power
  ! function within part of an ellipsoidal volume, then it is up to the
  ! calling routine to make sure that b-x0 <= root(c) and x0-a <= root(c).
  ! Unless this condition is met, the integrand will be < 0 over part
  ! of its range, which is non-physical.
  !---------------------------------------------------------------------


    implicit none

    real(double), intent(in) :: a,&
                                b,&
                                power,&
                                x0,&
                                c,&
                                amp
    real(double)             :: integral

    real(double) ::&
      intgrlTerm1, intgrlTerm2, intgrlTerm3,&
      pPlus1, pPlus2, pPlus3

    !-------------------------------------------------------------------
    ! Start

    pPlus1 = 1.0d0 + power
    pPlus2 = 2.0d0 + power
    pPlus3 = 3.0d0 + power

    if (pPlus1 == 0.0) then
      intgrlTerm1 =  (log(b) - log(a)) * (1.0d0 - x0*x0/c)
    else
      intgrlTerm1 =  (b**pPlus1 - a**pPlus1) / pPlus1 * (1.0d0 - x0*x0/c)
    end if

    if (pPlus2 == 0.0) then
      intgrlTerm2 =  (log(b) - log(a)) * 2.0d0 * x0/c
    else
      intgrlTerm2 =  (b**pPlus2 - a**pPlus2) / pPlus2 * 2.0d0*x0/c
    end if

    if (pPlus3 == 0.0) then
      intgrlTerm3 = -(log(b) - log(a)) / c
    else
      intgrlTerm3 = -(b**pPlus3 - a**pPlus3) / pPlus3 / c
    end if

    integral = amp * (intgrlTerm1 + intgrlTerm2 + intgrlTerm3)

  end function ellipsoidIntegral_new

  !---------------------------------------------------------------------
  subroutine saveFinalDetSrcList(detSrcData)
  !---------------------------------------------------------------------

    implicit none

    type(detSrcType), intent(in) :: detSrcData(:)

    character(256) :: detSrcListSetName
    character(20) :: detSrcListTabName
    type(DataSetT) :: set
    type(TableT) :: tab
    integer :: numSources

    !-------------------------------------------------------------------
    ! Start

    call splitSetTabName(detSrcListSetTabName, detSrcListSetName, detSrcListTabName)
    ! in ssclib/dal_aux.
    set = dataSet(detSrcListSetName, MODIFY)
    tab = table(set, detSrcListTabName)
    numSources = numberOfRows(tab)
if (size(detSrcData) .ne. numSources) call error('bug10','Please contact the code developer')
!    allocate(detSrcData(numSources), raDecErr(numSources))

    call setAttribute(tab, 'COMPARED', .true., comment='Comparison with list of sim sources has been done.')

    call writeColData(tab, 'X',     detSrcData%coord%xArcsec, 'arcsec')
    call writeColData(tab, 'Y',     detSrcData%coord%yArcsec, 'arcsec')
    call writeColData(tab, 'X_ERR', real(detSrcData%err%xArcsec,single), 'arcsec')
    call writeColData(tab, 'Y_ERR', real(detSrcData%err%yArcsec,single), 'arcsec')
    call writeColData(tab, 'SIM_X', detSrcData%simCoord%xArcsec, 'arcsec')
    call writeColData(tab, 'SIM_Y', detSrcData%simCoord%yArcsec, 'arcsec')
    call writeColData(tab, 'SIM_FLUX', real(detSrcData%simCoord%flux,single))
!    call writeColData(tab, 'SIM_INDX', detSrcData%indexOfNearestSim&
    call writeColData(tab, 'SIM_INDX', detSrcData%srcIdOfNearestSim&
      , comment='INDEX value of matching sim source')

    call writeColData(tab, 'SIM_INV_SENSY', detSrcData%inverseSensyAtSim)
    call writeColData(tab, 'R_SIGMAS', detSrcData%scaledR&
      , comment='Distance to matching sim source in sigmas')
    call writeColData(tab, 'MATCH_PNULL', detSrcData%nullProb&
      , comment='Probability of match occuring by chance')
    call writeColData(tab, 'FLAG', detSrcData%flag)

    select case(fluxStyle)
    case('rand')
      call writeColData(tab, 'SIM_LINF', detSrcData%simCoord%l&
        , comment='Linearized flux (evenly distributed rands)')
    case('simple')
    case('root')
      call writeColData(tab, 'ROOTF',     detSrcData%coord%l&
        , comment='sqrt('//trim(detFluxColName)//')')
      call writeColData(tab, 'ROOTF_ERR', detSrcData%err%l&
        , comment='error in ROOTF')
      call writeColData(tab, 'SIM_ROOTF', detSrcData%simCoord%l&
        , comment='sqrt(SIM_FLUX)')
    case default
call error('badFluxStyle',"")
    end select

    call setAttribute(tab, 'FLUXSTYL', fluxStyle)

    ! writeColData is in ssclib/dal_aux.

    call release(tab)
    call release(set)

  end subroutine saveFinalDetSrcList

  !---------------------------------------------------------------------
  subroutine modifySimSrcList(simSrcData)
  !---------------------------------------------------------------------

    type(simSrcType), intent(in) :: simSrcData(:)

    type(DataSetT) :: set
    type(TableT) :: tab
    integer :: numSources
    type(ColumnT) :: detSrcIdCol
    integer(int32), pointer :: detSrcId(:)

    !-------------------------------------------------------------------
    ! Start

    set = dataSet(simSrcListSetName, MODIFY)
    tab = table(set, 'SRCLIST')
    numSources = numberOfRows(tab)
if (numSources .ne. size(simSrcData)) call error('bug',"")

    call setAttribute(tab, 'DETIDCOL', detSrcIdColName)

    detSrcIdCol = addOrOpenColumn(tab, 'DET_SRC_ID', INTEGER32&
      , comment='ID of the best matching det src. 0 => no match.')
    ! in ssclib/dal_aux.

    detSrcId => int32Data(detSrcIdCol) 
    detSrcId = simSrcData%srcIdOfBestDet

    call release(detSrcIdCol)
    call release(tab)
    call release(set)

  end subroutine modifySimSrcList

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module srccompare_mod
