!+
! Module: srcflux2histo_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Calculates a histogram of source flux values and compares to the model.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module srcflux2histo_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.
  use dal
  use param  
  use errorHandling

  use power_law_utils

  ! ssclib modules:
!  use dal_aux		! for readColDataToFixed
!  use detection_stats	! for chi2Prob
  use histogram_utils
!use l_histogram_utils
!  use pgplot_aux
use l_pgplot_aux
  use ssc_misc		! for fracVal

!use l_plot_utils

  implicit none 

  ! Command-line parameters:
  character(256) ::&
    sourceListName = '',&
    pgplotDevice = ''
  character(20) ::&
    style='',&
    binStyle=''
  real(single) ::&
    userHistoBinWidth = 0.0
  logical ::&
!    doReduced = .false.,&
    doDiff = .false.,&
    plotLogAxes = .false.
  integer ::&
    userNumBins = 0

  private 

  public srcflux2histo

contains 

  !---------------------------------------------------------------------
  subroutine srcflux2histo
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT) :: set
    type(TableT) :: tab
    integer ::&
!      degFree,&
      numSamples, numHistoBins,&
!      underflows, overflows,&
!      foreignStatus,&
      i, numKnees
    type(ColumnT) :: col
    real(single), pointer :: fluxes(:)
    real(single) ::&
      minFlux, maxFlux, deltaFlux, meanFlux, histoSpan, minHistoVal,&
      maxHistoVal, hiPower, skyArea,&
! histoBinWidth,&
      underflows, overflows, logMinFlux, logMaxFlux, deltaLogFlux
    real(single), allocatable ::&
      weights(:), binEdges(:), binWidths(:), binCentres(:), modelPowers(:),&
      kVals(:), offsets(:), multipliers(:), diffYs(:), diffYErrs(:),&
      diffYsTheory(:), intYs(:), intYsTheory(:), histo(:), histoErrs(:)
    character(50) :: xLabel, yLabel
    real(single), pointer :: modelFluxes(:), modelCumSkyDensity(:)
logical :: testFlag=.false.

    !-------------------------------------------------------------------
    ! Start:

if (testFlag) print *, 'aaa'

    call readParameters

    set = dataSet(sourceListName, READ)
    tab = table(set, 'SRCLIST')
!    degFree = int32Attribute(tab, 'DEG_FREE')
    numSamples = numberOfRows(tab)
!    if (numSamples < 2) then
!      call warning('notEnoughValues',"There must be more than 1 value &
!        &in the table for a plot to have any point!")
!      call release(tab)
!      call release(set)
!  return
!    end if
    if (numSamples < 2)&
call error('notEnoughValues',"There must be more than 1 value &
 &in the table for a plot to have any point!")

    col = column(tab, 'FLUX')
    fluxes => real32Data(col)
    minFlux = minval(fluxes)
    if (minFlux <= 0.0)&
call error('nonPositiveFlux', "One of the sources has a flux value which is <= 0.")
    maxFlux = maxval(fluxes)
    deltaFlux = maxFlux - minFlux
if (testFlag) print *, 'minFlux, maxFlux, deltaFlux:', minFlux, maxFlux, deltaFlux
!    if (deltaFlux <= 0.0) then
!      call warning('zeroFluxRange', "The flux values are all the &
!        &same. The result of such a plot would be a bit boring.")
!      call release(col)
!      call release(tab)
!      call release(set)
!  return
!    end if
    if (deltaFlux <= 0.0)&
call error('zeroFluxRange', "The flux values are all the &
  &same. The result of such a plot would be a bit boring.")

    logMinFlux = log10(minFlux)
    logMaxFlux = log10(maxFlux)
    deltaLogFlux = logMaxFlux - logMinFlux

    allocate(weights(numSamples))
    select case(style)
    case('flux')
      weights = fluxes
      yLabel = 'Flux density (cgs deg\u-2\d)'
    case('number')
      weights = 1
      yLabel = 'Number density (deg\u-2\d)'
    case default
call error('badStyle',"")
    end select

    ! Decide on the number of histogram bins:
    !
    select case(binStyle)
    case('binwidth')
!      histoBinWidth = userHistoBinWidth
      if (plotLogAxes) then
        numHistoBins = ceiling(deltaLogFlux / userHistoBinWidth) ! userHistoBinWidth
        ! is assumed to be in log space in this instance.
        if (numHistoBins < 2) then
          write(errStr, '(a, es9.2, a)') "The difference between &
            & log10(minimum) and log10(maximum) flux values is only "&
            , deltaLogFlux, ". At your &
            &chosen value of --binwidth, less than 2 histogram bins &
            &span this range! Maybe you want to reduce the bin width."
call error('notEnoughBins')
        end if

      else ! don't do log/log plot
        numHistoBins = ceiling(deltaFlux / userHistoBinWidth)
        if (numHistoBins < 2) then
          write(errStr, '(a, es9.2, a)') "The difference between minimum &
            &and maximum flux values is only ", deltaFlux, ". At your &
            &chosen value of --binwidth, less than 2 histogram bins &
            &span this range! Maybe you want to reduce the bin width."
call error('notEnoughBins')
        end if
      end if

!      if (numHistoBins < 2) then
!        write(errStr, '(a, es9.2, a)') "The difference between minimum &
!          &and maximum flux values is only ", deltaFlux, ". At your &
!          &chosen value of --binwidth, less than 2 histogram bins &
!          &span this range! Maybe you want to reduce the bin width."
!        call warning('notEnoughBins')
!        call release(col)
!        call release(tab)
!        call release(set)
!  return
!      end if

    case('nbins')
      numHistoBins = userNumBins
!      histoBinWidth = deltaFlux / real(numHistoBins)
!if (testFlag) print *, 'numHistoBins, histoBinWidth:', numHistoBins, histoBinWidth
    case default
call error('badBinStyle',"")
    end select

    allocate(binEdges(0:numHistoBins), binWidths(numHistoBins))
    allocate(binCentres(numHistoBins))

    ! Now calculate the edges and centres of the histogram bins:
    !
    select case(binStyle)
    case('binwidth')
      if (plotLogAxes) then
        meanFlux = (logMaxFlux + logMinFlux) / 2.0
        histoSpan = numHistoBins * userHistoBinWidth ! userHistoBinWidth is assumed
        ! to be in log space in this instance.
        minHistoVal = min(meanFlux - histoSpan / 2.0, logMinFlux)
        maxHistoVal = max(meanFlux + histoSpan / 2.0, logMaxFlux)
        binEdges = fracVal(minHistoVal, maxHistoVal&
          , (/(i/real(numHistoBins), i=0,numHistoBins)/)) ! in ssclib/ssc_misc.
        binCentres = (binEdges(1:) + binEdges(:numHistoBins-1)) / 2.0

        binEdges   = 10.0**binEdges
        binCentres = 10.0**binCentres

      else ! don't do log/log plot
        meanFlux = (maxFlux + minFlux) / 2.0
        histoSpan = numHistoBins * userHistoBinWidth
        minHistoVal = min(meanFlux - histoSpan / 2.0, minFlux)
        maxHistoVal = max(meanFlux + histoSpan / 2.0, maxFlux)
        binEdges = fracVal(minHistoVal, maxHistoVal&
          , (/(i/real(numHistoBins), i=0,numHistoBins)/)) ! in ssclib/ssc_misc.
        binCentres = (binEdges(1:) + binEdges(:numHistoBins-1)) / 2.0
      end if

    case('nbins')
      if (plotLogAxes) then
        binEdges = fracVal(logMinFlux, logMaxFlux&
          , (/(i/real(numHistoBins), i=0,numHistoBins)/)) ! in ssclib/ssc_misc.
        binCentres = (binEdges(1:) + binEdges(:numHistoBins-1)) / 2.0

        binEdges   = 10.0**binEdges
        binCentres = 10.0**binCentres

      else ! don't do log/log plot
        binEdges = fracVal(minFlux, maxFlux&
          , (/(i/real(numHistoBins), i=0,numHistoBins)/)) ! in ssclib/ssc_misc.
        binCentres = (binEdges(1:) + binEdges(:numHistoBins-1)) / 2.0
      end if
    case default
call error('badBinStyle',"")
    end select

    binWidths = binEdges(1:) - binEdges(:numHistoBins-1)

    allocate(histo(numHistoBins), histoErrs(numHistoBins))
    call makeHistogram(fluxes, histo, binEdges, weights&
      , underflows, overflows, histoErrs)!, foreignStatus) ! in ssclib/histogram_utils.
!    if (foreignStatus .ne. 0) then
!      print *, 'makeHistogram status return:', foreignStatus
!call error('badHistogramStatusReturn',"")
!    end if
if (testFlag) print *, 'minval(histo),maxval(histo):', minval(histo),maxval(histo)

if (testFlag) then
  do i = 1, numHistoBins
!    if (histo(i)>0) print *, i, binCentres(i), histo(i), sqrt(histo(i))
  end do
end if

    call release(col)
    call release(tab)

    tab = table(set, 'SRCSPECS')
    call getLogNlogS(tab, modelFluxes, modelCumSkyDensity&
      , hiPower) ! in power_law_utils.
    skyArea = real64Attribute(tab, 'SKY_AREA')
if (testFlag) print *, 'skyArea:', skyArea
    call release(tab)
    call release(set)
    numKnees = size(modelFluxes)
    allocate(modelPowers(numKnees), kVals(numKnees))
    call calcLogNLogSPowers(modelFluxes, modelCumSkyDensity, hiPower&
      , modelPowers) ! in power_law_utils
    kVals = -modelCumSkyDensity * (1.0 + modelPowers)&
      / (modelFluxes**(1.0+modelPowers))
    allocate(offsets(numKnees), multipliers(numKnees))
    call calcReverseAccumFluxCoeffs(modelFluxes, kVals, modelPowers&
      , offsets, multipliers)
!do i = 1, numKnees
!  print *, i, modelFluxes(i-1), modelCumSkyDensity(i-1), modelPowers(i), kVals(i), offsets(i), multipliers(i)
!end do

!    if (doReduced) then
!      xLabel='Reduced chi squared'
!    else
!      xLabel='Chi squared'
!    end if
xLabel='Flux (cgs)'

if (testFlag) print *, 'bbb'
    if (doDiff) then
      allocate(diffYs(numHistoBins), diffYErrs(numHistoBins))
      allocate(diffYsTheory(numHistoBins))

!      diffYs = histo / real(numSamples) / binWidths
!      diffYErrs = sqrt(real(histo)) / real(numSamples)&
!        / binWidths
      diffYs = histo / skyArea/ binWidths
!      diffYErrs = sqrt(histo) / skyArea / binWidths
      diffYErrs = histoErrs / skyArea / binWidths
if (testFlag) then
  do i = 1, numHistoBins
    if (histo(i)>0) print *, i, binCentres(i), diffYs(i), diffYErrs(i)
  end do
end if

      do i = 1, numHistoBins
        diffYsTheory(i) = calcNumDensity(binCentres(i), modelFluxes, kVals&
          , modelPowers)
        ! in power_law_utils.
      end do

      select case(style)
      case('flux')
        diffYsTheory = diffYsTheory * binCentres
      case('number')
      case default
call error('badStyle',"")
      end select

      call plotInit(pgplotDevice, xLabel, yLabel)
      if (plotLogAxes) then
        call setXLog
        call setYLog
      end if
      call plotAutoSizeX(binCentres)
      call plotAutoSizeY(diffYs, diffYErrs)
      call plotAutoSizeY(diffYsTheory)
if (testFlag) print *, 'ccc'
      call drawPointPlot(binCentres, diffYs, yErrs=diffYErrs, finished=.false.)
if (testFlag) print *, 'ddd'
      call drawLinePlot(binCentres, diffYsTheory)
      ! All these are in plotlib/plot_utils.

    else ! do integrated plot:
      yLabel = trim(yLabel)//' (cumulative)'

      allocate(intYs(numHistoBins), intYsTheory(numHistoBins))

      i = numHistoBins
      intYs(i) = overflows + histo(i)

      do i = numHistoBins-1, 1, -1
        intYs(i) = intYs(i+1) + histo(i)
      end do
!      intYs = intYs / real(numSamples)
      intYs = intYs / skyArea

      select case(style)
      case('flux')
        do i = 1, numHistoBins
          intYsTheory(i) = calcReverseCumFluxDensity(binCentres(i)&
            , modelFluxes, modelPowers, offsets, multipliers) ! in power_law_utils
!print *, i, binCentres(i), intYsTheory(i)
        end do

      case('number')
        do i = 1, numHistoBins
          intYsTheory(i) = calcReverseCumNumDensity(binCentres(i), modelFluxes&
            , kVals, modelPowers) ! in power_law_utils
        end do

      case default
call error('badStyle',"")
      end select

      call plotInit(pgplotDevice, xLabel, yLabel)
      if (plotLogAxes) then
        call setXLog
        call setYLog
      end if
      call plotAutoSizeX(binEdges)
      call plotAutoSizeY(intYs)
      call plotAutoSizeY(intYsTheory)
if (testFlag) print *, 'xxx'
!      call drawHistogram(binEdges, intYs, finished=.false.)
call drawLinePlot(binEdges(0:numHistoBins-1), intYs, finished=.false.)
if (testFlag) print *, 'yyy'
      call drawLinePlot(binCentres, intYsTheory)
      ! All these are in plotlib/plot_utils.
    end if ! doDiff

if (testFlag) print *, 'zzz'

  end subroutine srcflux2histo

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start:

    sourceListName = stringParameter('srclistset')
    pgplotDevice = stringParameter('pgdevice')

    binStyle = stringParameter('binstyle')
    select case(binStyle)
    case('binwidth')
      userHistoBinWidth = realParameter('binsize')
    case('nbins')
      userNumBins = intParameter('nbins')
    case default
call error('badBinStyle',"")
    end select

    doDiff = booleanParameter('dodiff')
    plotLogAxes = booleanParameter('dolog')
    style = stringParameter('style')

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module srcflux2histo_mod
