!+
! Module: srclistsim_mod
!
! Authors:  Ian Stewart - JBO
!          
! Description: Creates a list of sources with random positions and fluxes.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module srclistsim_mod

  use dal
  use param  
  use errorHandling

  use constants
  use coordinate
  use dal_aux
!use l_dal_aux_2
  use random_aux
!use l_random_aux

  use power_law_utils
!use temp_ssclib

  implicit none 

  private 

  ! Command-line parameters:
  character(256) ::&
    srcSpecSetName='',&
    sourceListName=''

  public srclistsim

contains 

  !---------------------------------------------------------------------
  subroutine srclistsim
  !---------------------------------------------------------------------
  ! This subroutine creates a list of simulated sources. Each source has
  ! a position in the sky (which must fall within the cone specified by
  ! the template set) and an amplitude.
  !
  ! Positions are random with a uniform distribution over the celestial
  ! sphere, and the amplitudes are random with a distribution according
  ! to the logN-logS stored in the template set.
  !
  ! The format of srcSpecSet is as follows:
  !
  !	- The logN-logS information is contained in a binary table called
  ! SRCSPECS. This table should contain at least 1 row, with the following
  ! real32 columns:
  !
  !		- FLUX: the source flux in ergs cm^-2 s^-1.
  !
  !		- DENSITY: the total sky density (in sources deg^-2) above
  ! that flux.
  !
  !	- The FLUX entries must increase monotonically with row number, and
  ! the DENSITY entries must decrease or remain constant with row number.
  !
  !	- This table should contain the following keywords:
  !
  !		- The circular area to contain the sources is specified by
  ! the header keywords CONE_RA, CONE_DEC and CONE_RAD, with units decimal
  ! degrees, decimal degrees and arcsec respectively. (The sky area of
  ! the cone is given by the keyword SKY_AREA in deg^2, but this is a
  ! convenience merely.)
  !
  !		- (optional) HI_SLOPE. This real32 keyword gives the exponent
  ! (= slope on a log-log graph) of the high-flux power-law extension. If it
  ! is not present, a value of -1.5 is assumed. If present, it must be <0.
  !
  ! It is assumed that the user will add additional keywords to indicate
  ! how 'FLUX' should be interpreted.
  !
  ! The table represents a graph of integrated source density in sources
  ! per square degree against the source flux. The
  ! FLUX/DENSITY pairs are taken to represent `knees' or vertices in a
  ! graph of integrated sky density N(>S) against flux S. The function
  ! is assumed to obey a power law in the gaps between vertices: thus
  ! in log-log space it is represented by a sequence of line segments.
  ! The pairs of values should occur in order of increasing flux, and
  ! non-physical situations (eg non-positive values, integrated densities
  ! which increase with flux) are naturally not permitted.
  !.....................................................................

    implicit none

    integer(int32) ::&
      numSources, numModelPoints
!, obsIdNum
    real(single), allocatable ::&
      srcFluxes(:), origRandoms(:), recalcRandoms(:),&
      theta(:), phi(:)
    real(single) ::&
      hiPower,&
      numSourcesReal
    real(single), pointer ::&
      modelFluxes(:),&
      modelCumSkyDensity(:),&
      plotFluxes(:),&
      plotCumSkyDensityDeg(:)
    real(double) ::&
      coneRaDeg, coneDecDeg, coneRadiusArcsec, coneRadius,&
      coneArea
    real(double), allocatable ::&
      srcRaDeg(:),&
      srcDecDeg(:),&
      srcRa(:), srcDec(:)
    type(DataSetT) :: srcSpecSet, srcSet
    type(TableT) :: srcSpecTab, srcTab, plotTab
    integer :: foreignStatus
!    type(ColumnT) :: flagCol
!    integer(int16), pointer :: flag(:)
    integer(int32), pointer :: indices(:)
integer(int32) :: i
logical :: testFlag=.false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, 'aaa'

    call readParameters

    ! Read the logN-logS distribution model from the template file:
    !
    srcSpecSet = dataSet(srcSpecSetName, READ)
    srcSpecTab = table(srcSpecSet, 'SRCSPECS')
if (testFlag) print *, 'bbb'
    call getLogNlogS(srcSpecTab, modelFluxes, modelCumSkyDensity&
      , hiPower) ! in power_law_utils
if (testFlag) print *, 'ccc'

    numModelPoints = size(modelFluxes)

    !...................................................................
    ! Calculate numSources by multiplying maxDensityDeg (the total density
    ! of sources per square degree) as got from logNlogSSet by the area
    ! of the cone in square degrees:

    coneRaDeg        = real64Attribute(srcSpecTab, 'CONE_RA')
    coneDecDeg       = real64Attribute(srcSpecTab, 'CONE_DEC')
    coneRadiusArcsec = real64Attribute(srcSpecTab, 'CONE_RAD')
    call release(srcSpecTab)

    coneRadius = coneRadiusArcsec * PiOn180 / 3600.0d0
    coneArea = TwoPi * (1.0d0 - cos(coneRadius))

    numSourcesReal = (coneArea / PiOn180 / PiOn180) * modelCumSkyDensity(0)
    ! Add some noise to the number:
    numSources = nint(numSourcesReal + gaussianRandom() * sqrt(numSourcesReal))
    ! gaussianRandom is in ssclib/random_aux.
    if (numSources <= 0)&
call error('noSourcesInSkyArea','power_law_utils: The sky area for simulated &
&sources times the density of sources is probably too small.')

    !...................................................................
    ! Get random amplitudes and positions for the sources:

if (testFlag) print *, 'eee'
    allocate(srcFluxes(numSources))
    allocate(srcRa(numSources), srcDec(numSources))
    allocate(srcRaDeg(numSources), srcDecDeg(numSources))
    allocate(theta(numSources), phi(numSources))
    allocate(origRandoms(numSources), recalcRandoms(numSources))

if (testFlag) print *, 'fff'
    call getSineRandoms(theta, 0.0, real(coneRadius, single))
!, foreignStatus)
    ! in ssclib/random_aux.
if (testFlag) print *, 'ggg'

!if (foreignStatus .ne. 0) then
!print *, 'getSineRandoms status:', foreignStatus
!call error('getSineRandomsError',"power_law_utils: The function getSineRandoms &
!&returned an error status.")
!end if
    call getRandomNumbers(phi) ! in ssclib/random_aux.
    phi = TwoPi * phi

if (testFlag) print *, 'hhh'
    call polarsToRaDec(real(theta, double)&
     , real(phi, double), coneRaDeg*PiOn180, coneDecDeg*PiOn180, srcRa, srcDec)
    ! in ssclib/coordinate
if (testFlag) print *, 'iii'

    srcRaDeg  = srcRa  / PiOn180
    srcDecDeg = srcDec / PiOn180

if (testFlag) print *, 'jjj'
    call getLogNlogSrandoms(modelFluxes, modelCumSkyDensity, hiPower&
      , srcFluxes, origRandoms, foreignStatus)
    ! in ssclib/random_aux.
if (testFlag) print *, 'kkk'

    if (foreignStatus < 0) then
print *, 'ssclib: Status return from getLogNlogSrandoms was', foreignStatus
call error('getLogNlogSrandoms',"power_law_utils: The function getLogNlogSrandoms &
&returned an error status.")
    end if

    !...................................................................
    ! Recover the original (evenly-distributed between 0 and 1) random
    ! numbers used to generate the flux values:

if (testFlag) print *, 'lll'
    call invertLogNlogSrandoms(modelFluxes, modelCumSkyDensity&
      , srcFluxes, hiPower, recalcRandoms, foreignStatus)
    ! in power_law_utils.
if (testFlag) print *, 'mmm'

!*** dump origRandoms and recalcRandoms and compare - they should be the same value in each case.

    if (foreignStatus < 0) then
print *, 'ssclib: Status return from invertLogNlogSrandoms was', foreignStatus
call error('invertLogNlogSrandoms','')
    end if

    !...................................................................
    ! Write source list to file:

    srcSet = dataSet(sourceListName, CREATE)
    srcTab = addTable(srcSet, 'SRCLIST', numSources)

!    call setAttribute(srcTab, 'SDY0_CUT',  psfCutoffFluxDensity0&
!      , 'erg / sec / cm^2 / eV', 'Flux density boundary between use &
!      &of coarse vs cal PSF.')

    allocate(indices(numSources))
    indices = (/(i,i=1,numSources)/)

    call writeColData(srcTab, 'INDEX',    indices) ! in ssclib/dal_aux.
    deallocate(indices)
    call writeColData(srcTab, 'FLUX',     srcFluxes)
    call writeColData(srcTab, 'RA',       srcRaDeg,  'deg')
    call writeColData(srcTab, 'DEC',      srcDecDeg, 'deg')
    call writeColData(srcTab, 'FLUXRAND', recalcRandoms, ''&
      , 'The original random numbers used to generate the FLUX values.')

!    ! The flags are mostly not set until the images are made:
!    do obsIdNum = 1, size(obsIdList)
!      flagCol = addColumn(srcTab, 'FLAG_'//obsIdList(obsIdNum), INTEGER16)
!      flag => int16Data(flagCol)
!      flag = 0
!      where (srcFluxDensity0 < psfCutoffFluxDensity0)&
!        flag = ibset(flag, COARSE_PSF_BIT)
!
!      call release(flagCol)
!    end do
    call release(srcTab)

    !...................................................................
    ! Add the template table to the source list dataset:

if (testFlag) print *, 'nnn'
    call copyBlock(srcSet, block(srcSpecSet, 'SRCSPECS'))
if (testFlag) print *, 'ooo'

    !...................................................................
    ! Generate a further table to assist in plotting the logN-logS model
    ! alongside the actual distribution of sources:

    plotTab = addtable(srcSet, 'MODEL_PL', numModelPoints + 1)
    plotFluxes => real32Data(addColumn(plotTab, 'FLUX', REAL32))
    plotCumSkyDensityDeg => real32Data(addColumn(plotTab, 'DENSITY', REAL32&
      , 'deg^-2'))

    plotFluxes(0:numModelPoints-1) = modelFluxes
    plotFluxes(numModelPoints) = 10.0 * maxval(srcFluxes)
    plotCumSkyDensityDeg(0:numModelPoints-1) = modelCumSkyDensity

if (testFlag) print *, 'ppp'
    plotCumSkyDensityDeg(numModelPoints)&
      = plotCumSkyDensityDeg(numModelPoints-1)&
      * (plotFluxes(numModelPoints)&
      / plotFluxes(numModelPoints-1))**(hiPower)

    !...................................................................
    ! Clean up:

    call release(srcSet)
    call release(srcSpecSet)
    deallocate(modelFluxes)
    deallocate(modelCumSkyDensity)
if (testFlag) print *, 'zzz'

  end subroutine srclistsim

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    srcSpecSetName = stringParameter('srcspecset')
    sourceListName = stringParameter('srclistset')
!    withAccuratePsfForBrights = booleanParameter('specialbrights')
!    if (withAccuratePsfForBrights)&
!      psfCutoffFluxDensity0 = realParameter('fluxcut')

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module srclistsim_mod
